// Copyright (c) 2015, Electric Cloud, Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   - Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//
//   - Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// (As an historical note, the above license instantiates
// the "BSD 2-Clause License" template that was presented
// by the Open Source Initiative at
//
//   http://www.opensource.org/licenses/bsd-license.php
//
// on 2012-02-23.)

// Summary:
//
// This header defines C++ wrappers for interfaces in:
//
//     "../interface/vrpsx_embedding.h"
//
// NOTE: This convenience header does not affect the
// definition of any any binary interface; see README.txt.
//
// In general, methods that end in "_nt" are "nothrow" variants
// that return error numbers instead of throwing exceptions.

#ifndef VRPSX_EMBEDDING_HPP
#define VRPSX_EMBEDDING_HPP 1

#include "../interface/vrpsx_embedding.h"
#include "vrpsx_system.hpp"


namespace vrpsx {
    class embedding_system_padding_t
    {
    private:
        char mPadding[offsetof(vrpsx_system_t, allocator)];
    };

    VRPSX_CPP_WRAPPER(embedding_t);

    class embedding_t :
        private embedding_system_padding_t,
        public system_t     // because system_t has nonzero size,
                            // even when used as a base class, the
                            // above padding ensures the correct offset
    {
    public:
        vrpsx_errno_t internalize_pid_nt(
            void const *outer_pid,
            size_t outer_size,
            vrpsx_pid_t& inner_pid)
        {
            vrpsx_embedding_t *iface = from_cpp(this);
            return iface->vtbl->internalize_pid(
                iface, outer_pid, outer_size, &inner_pid);
        }

        vrpsx_pid_t internalize_pid(
            void const *outer_pid,
            size_t outer_size)
        {
            vrpsx_pid_t result;
            throw_on_error(
                "vrpsx::embedding::internalize_pid",
                internalize_pid_nt(outer_pid, outer_size, result));
            return result;
        }

        // NOTE: Be sure to destroy or release outer_pid before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t externalize_pid_nt(
            vrpsx_pid_t inner_pid,
            allocator_t *allocator,
            unique_ptr< void, free_memory_t >& outer_pid,
            size_t& outer_size)
        {
            vrpsx_embedding_t *iface = from_cpp(this);
            void *raw_outer_pid;
            vrpsx_errno_t ern = iface->vtbl->externalize_pid(
                iface, inner_pid, from_cpp(allocator),
                &raw_outer_pid, &outer_size);
            unique_allocation(raw_outer_pid, allocator).swap(outer_pid);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< void, free_memory_t > externalize_pid(
            vrpsx_pid_t inner_pid,
            allocator_t *allocator,
            size_t& outer_size)
        {
            unique_ptr< void, free_memory_t > result;
            throw_on_error(
                "vrpsx::embedding::externalize_pid",
                externalize_pid_nt(inner_pid, allocator, result, outer_size));
            return move(result);
        }

        // NOTE: Be sure to destroy or release inner_path before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t internalize_path_nt(
            char const *outer_path,
            allocator_t *allocator,
            unique_ptr< char, free_memory_t >& inner_path)
        {
            vrpsx_embedding_t *iface = from_cpp(this);
            char *raw_inner_path;
            vrpsx_errno_t ern = iface->vtbl->internalize_path(
                iface, outer_path, from_cpp(allocator), &raw_inner_path);
            unique_allocation(raw_inner_path, allocator).swap(inner_path);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< char, free_memory_t > internalize_path(
            char const *outer_path,
            allocator_t *allocator)
        {
            unique_ptr< char, free_memory_t > result;
            throw_on_error(
                "vrpsx::embedding::internalize_path",
                internalize_path_nt(outer_path, allocator, result));
            return move(result);
        }

        // NOTE: Be sure to destroy or release outer_path before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t externalize_path_nt(
            char const *inner_path,
            allocator_t *allocator,
            unique_ptr< char, free_memory_t >& outer_path)
        {
            vrpsx_embedding_t *iface = from_cpp(this);
            char *raw_outer_path;
            vrpsx_errno_t ern = iface->vtbl->externalize_path(
                iface, inner_path, from_cpp(allocator), &raw_outer_path);
            unique_allocation(raw_outer_path, allocator).swap(outer_path);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< char, free_memory_t > externalize_path(
            char const *inner_path,
            allocator_t *allocator)
        {
            unique_ptr< char, free_memory_t > result;
            throw_on_error(
                "vrpsx::embedding::externalize_path",
                externalize_path_nt(inner_path, allocator, result));
            return move(result);
        }

        // NOTE: Be sure to destroy or release inner_path before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t internalize_execution_path_nt(
            char const *outer_path,
            allocator_t *allocator,
            unique_ptr< char, free_memory_t >& inner_path)
        {
            vrpsx_embedding_t *iface = from_cpp(this);
            char *raw_inner_path;
            vrpsx_errno_t ern = iface->vtbl->internalize_execution_path(
                iface, outer_path, from_cpp(allocator), &raw_inner_path);
            unique_allocation(raw_inner_path, allocator).swap(inner_path);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< char, free_memory_t > internalize_execution_path(
            char const *outer_path,
            allocator_t *allocator)
        {
            unique_ptr< char, free_memory_t > result;
            throw_on_error(
                "vrpsx::embedding::internalize_execution_path",
                internalize_execution_path_nt(outer_path, allocator, result));
            return move(result);
        }

        // NOTE: Be sure to destroy or release outer_path before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t externalize_execution_path_nt(
            char const *inner_path,
            allocator_t *allocator,
            unique_ptr< char, free_memory_t >& outer_path)
        {
            vrpsx_embedding_t *iface = from_cpp(this);
            char *raw_outer_path;
            vrpsx_errno_t ern = iface->vtbl->externalize_execution_path(
                iface, inner_path, from_cpp(allocator), &raw_outer_path);
            unique_allocation(raw_outer_path, allocator).swap(outer_path);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< char, free_memory_t > externalize_execution_path(
            char const *inner_path,
            allocator_t *allocator)
        {
            unique_ptr< char, free_memory_t > result;
            throw_on_error(
                "vrpsx::embedding::externalize_execution_path",
                externalize_execution_path_nt(inner_path, allocator, result));
            return move(result);
        }
    };
}


#endif /* VRPSX_EMBEDDING_HPP */
