// Copyright (c) 2015, Electric Cloud, Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   - Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//
//   - Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// (As an historical note, the above license instantiates
// the "BSD 2-Clause License" template that was presented
// by the Open Source Initiative at
//
//   http://www.opensource.org/licenses/bsd-license.php
//
// on 2012-02-23.)

// Summary:
//
// This header defines utilities for using vrpsx interfaces
// that have single-owner interface lifetime control.
//
// NOTE: This convenience header does not affect the
// definition of any any binary interface; see README.txt.

#ifndef VRPSX_SINGLE_OWNER_HPP
#define VRPSX_SINGLE_OWNER_HPP 1

#include "vrpsx_move.hpp"


namespace vrpsx {
    // We use this kind of smart pointer for C++ wrappers
    // for single-owner vrpsx interfaces.  That is, there
    // can be at most one owner, and when that one owner
    // disowns the interface, then it is cleaned up.

    using boost::interprocess::unique_ptr;

    // unique_ptr< W, disown_t< W > > is the correct type
    // for smart pointers to single-owner vrpsx wrappers.
    //
    // Note that some single-owner vrpsx interfaces use "next"
    // instead of "disown", but for uniformity the C++ wrapper
    // defines a "disown_mt" member function in terms of "next".
    //
    // The "_mt" stands for "maybe throw":
    // throw if there is no uncaught exception.

    template< typename W >
    struct disown_t
    {
        void operator()(W *wrapper) const
        {
            wrapper->disown_mt();
        }
    };

    // single_owner_ptr< W >::type means unique_ptr< W, disown_t< W > >
    //
    // NOTE: Right now we support C++98, but C++11 adds "alias declarations",
    // and if single_owner_ptr were one of those, we could omit the "::type".

    template< typename W >
    struct single_owner_ptr_t
    {
        typedef unique_ptr< W, disown_t< W > > type;
    };

    template< typename C >
    inline
    typename single_owner_ptr_t< typename cpp_type_for_t< C >::type >::type
    to_cpp_single_owner(
        C *iface)
    {
        typedef typename cpp_type_for_t< C >::type CPP;
        typedef typename single_owner_ptr_t< CPP >::type ptr;
        return ptr(to_cpp(iface));
    }
}


#endif /* VRPSX_SINGLE_OWNER_HPP */
