/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * This is an example program that makes use of a virtual POSIX-like interface
 * and an embedding of that interface inside some possibly-larger system.
 *
 * This program walks a directory tree, reporting the virtual and actual
 * pathnames encountered, as well as various information about those paths.
 */

#define __STRICT_ANSI__ 1

#include <vrpsx_embedding_main.h>
#include <vrpsx_embedding_main_identifier.h>
#include <vrpsx_identified_function.h>
#include <vrpsx_identified_function_identifier.h>
#include <vrpsx_utility.h>
#include <string.h>


typedef struct path_buffer_struct {
    char *begin;
    size_t length;
    size_t capacity;
    vrpsx_allocator_t *allocator;
} path_buffer_t;

static void path_buffer_init(
    path_buffer_t *buffer,
    vrpsx_allocator_t *allocator)
{
    buffer->begin = NULL;
    buffer->length = 0;
    buffer->capacity = 0;
    buffer->allocator = allocator;
}

static void path_buffer_destroy(
    path_buffer_t *buffer)
{
    VRPSX_CALL1(buffer->allocator, free_memory, buffer->begin);

    /* Not strictly necessary; helps detect buggy use of destroyed buffer:
     */
    buffer->begin = NULL;
    buffer->length = 0;
    buffer->capacity = 0;
    buffer->allocator = NULL;
}

static char *path_buffer_reserve(
    path_buffer_t *buffer,
    size_t desired_capacity)
{
    char *allocation = NULL;

    if (desired_capacity <= buffer->capacity) {
        return buffer->begin;
    }

    if (desired_capacity < (buffer->capacity << 1)) {
        desired_capacity = (buffer->capacity << 1);
    }

    allocation = (char *) VRPSX_CALL1(
        buffer->allocator, allocate_memory, desired_capacity);
    if (! allocation) {
        return NULL;
    }

    if (buffer->begin) {
        memcpy(allocation, buffer->begin, buffer->length + 1);
        VRPSX_CALL1(buffer->allocator, free_memory, buffer->begin);
    }

    buffer->begin = allocation;
    buffer->capacity = desired_capacity;
    return buffer->begin;
}

static char *path_buffer_assign(
    path_buffer_t *buffer,
    char const *path)
{
    size_t length = strlen(path);

    buffer->length = 0;
    if (! path_buffer_reserve(buffer, length + 1)) {
        return NULL;
    }

    memcpy(buffer->begin, path, length + 1);
    buffer->length = length;
    return buffer->begin;
}

static void path_buffer_shorten(
    path_buffer_t *buffer,
    size_t new_length)
{
    buffer->length = new_length;
    buffer->begin[new_length] = '\0';
}

static char const *path_buffer_append(
    path_buffer_t *buffer,
    char const *name)
{
    size_t separated = buffer->length > 1
            || (buffer->length == 1 && buffer->begin[0] != '/');
    size_t name_length = strlen(name);
    size_t full_length = buffer->length + separated + name_length;
    char *next;

    if (! path_buffer_reserve(buffer, full_length + 1)) {
        return NULL;
    }

    next = buffer->begin + buffer->length;

    if (separated) {
        *next++ = '/';
    }

    memcpy(next, name, name_length + 1);

    buffer->length = full_length;
    return buffer->begin;
}


/* Reports a memory allocation failure and returns enomem.
 */
static int memory_error(
    vrpsx_system_t *system)
{
    static char const msg[] = "Not enough space.\n";

    VRPSX_CALL4(system, write, 2, msg, sizeof msg - 1, NULL);

    return VRPSX_CONSTANT(system, enomem);
}


/* Reports an error for the given path with the given system error number.
 */
static void path_error(
    vrpsx_system_t *system,
    char const *path,
    vrpsx_errno_t ern)
{
    vrpsx_system_error_t *error = NULL;
    char const *base;

    VRPSX_CALL2(system, error_acquire, ern, &error);
    base = VRPSX_CALL0(error, text);

    VRPSX_CALL4(system, write, 2, path, strlen(path), NULL);
    VRPSX_CALL4(system, write, 2, ": ", 2, NULL);
    VRPSX_CALL4(system, write, 2, base, strlen(base), NULL);

    VRPSX_COUNTED_REFERENCE_CLEAR(vrpsx_system_error_t, &error);
    return;
}


static int report(
    vrpsx_embedding_t *embedding,
    char const *path)
{
    vrpsx_system_t *system = &embedding->system;
    char *outer_path = NULL;
    char *text = NULL;
    char *next;
    size_t path_length;
    size_t outer_path_length;
    size_t text_length;
    vrpsx_errno_t ern = 0;

    ern = VRPSX_CALL3(embedding, externalize_path, path, NULL, &outer_path);
    if (ern) {
        path_error(system, path, ern);
        goto cleanup;
    }

    path_length = strlen(path);
    outer_path_length = outer_path ? strlen(outer_path) : 0;
    text_length = path_length + 1 + outer_path_length + 1;

    text = VRPSX_CALL1(system, allocate_memory, text_length);
    if (! text) {
        ern = memory_error(system);
        goto cleanup;
    }

    next = text;
    MEMCPY_END(next, path, path_length);
    *next++ = '\t';
    MEMCPY_END(next, outer_path, outer_path_length);
    *next++ = '\n';

    ern = VRPSX_CALL4(system, write, 1, text, text_length, NULL);
    if (ern) {
        path_error(system, "stdout", ern);
        goto cleanup;
    }

cleanup:
    VRPSX_CALL1(system, free_memory, text);
    VRPSX_CALL1(system, free_memory, outer_path);
    return ern;
}


static int walk(
    vrpsx_embedding_t *embedding,
    path_buffer_t *buffer)
{
    vrpsx_system_t *system = &embedding->system;
    vrpsx_directory_t *dir = NULL;
    vrpsx_dirent_t *ent = NULL;
    size_t dir_length = buffer->length;
    char const *name;
    vrpsx_errno_t ern = 0, unused_ern;

    ern = report(embedding, buffer->begin);
    if (ern) {
        goto cleanup;
    }

    ern = VRPSX_CALL2(system, opendir, buffer->begin, &dir);
    if (ern) {
        if (ern == VRPSX_CONSTANT(system, enotdir)) {
            ern = 0;
        } else {
            path_error(system, buffer->begin, ern);
        }
        goto cleanup;
    }

    for (;;) {
        ern = VRPSX_CALL1(dir, readdir, &ent);
        if (ern) {
            path_error(system, buffer->begin, ern);
            goto cleanup;
        }

        if (! ent) {
            break;
        }

        name = VRPSX_CALL0(ent, name);
        if (name[0] == '.'
                && (name[1] == '\0' || (name[1] == '.' && name[2] == '\0'))) {
            continue;
        }

        if (! path_buffer_append(buffer, name)) {
            ern = memory_error(system);
            goto cleanup;
        }

        ern = walk(embedding, buffer);
        if (ern) {
            goto cleanup;
        }

        path_buffer_shorten(buffer, dir_length);
    }

cleanup:
    VRPSX_SINGLE_OWNER_CLOSE_SROE(vrpsx_directory_t, &dir, &unused_ern, NULL);
    path_buffer_shorten(buffer, dir_length);
    return ern;
}


VRPSX_DECLARE_IDENTIFIED_FUNCTION(static, main, vrpsx_embedding_main);

static int VRPSX_IDENTIFIED_FUNCTION_NAME(main, vrpsx_embedding_main)(
    vrpsx_embedding_initial_t *initial)
{
    static char const usage1[] = "Usage: ";
    static char const usage2[] = " <path> [<path> ...]\n";

    vrpsx_embedding_t *embedding = initial->embedding;
    vrpsx_system_t *system = &embedding->system;
    int argc = initial->arguments->dimension;
    char const *const *argv = initial->arguments->array;
    path_buffer_t buffer;
    int index;
    vrpsx_errno_t ern = 0;

    if (argc == 1) {
        VRPSX_CALL4(system, write, 2, usage1, sizeof usage1 - 1, NULL);
        VRPSX_CALL4(system, write, 2, argv[0], strlen(argv[0]), NULL);
        VRPSX_CALL4(system, write, 2, usage2, sizeof usage2 - 1, NULL);
        return VRPSX_CONSTANT(system, exit_failure);
    }

    path_buffer_init(&buffer, &system->allocator);

    for (index = 1; index < argc; ++index) {
        if (! path_buffer_assign(&buffer, argv[index])) {
            ern = memory_error(system);
            goto cleanup;
        }

        ern = walk(embedding, &buffer);
        if (ern) {
            goto cleanup;
        }
    }

cleanup:
    path_buffer_destroy(&buffer);

    return ern
            ? VRPSX_CONSTANT(system, exit_failure)
            : VRPSX_CONSTANT(system, exit_success);
}

vrpsx_identified_function_t const VRPSX_IDENTIFIED_FUNCTION_LIST[] = {
    VRPSX_MENTION_IDENTIFIED_FUNCTION(main, vrpsx_embedding_main),
    VRPSX_IDENTIFIED_FUNCTION_LIST_TERMINATOR
};
