/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Implements a virtual POSIX-like interface on top of a POSIX-like system.
 * Also implements the (sometimes rather trivial) embedding thereof.
 */

#define _POSIX_C_SOURCE 200809L
#define _XOPEN_SOURCE   500
#define _SVID_SOURCE 1 /* for getmntent_r */

#include "implementation.h"
#include "support_map.h"
#include "impl_config.h"
#include <direct_identifier.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <vrpsx_embedding_identifier.h>
#include <vrpsx_system_identifier.h>
#include <vrpsx_system_error_identifier.h>
#include <hybrid_allocator.h>
#include <sys/stat.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <sys/types.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <sys/utsname.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <assert.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <ctype.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <dirent.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <errno.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <fcntl.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <iconv.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <limits.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <locale.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <pthread.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <pwd.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <signal.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <stdlib.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <string.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <unistd.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <wchar.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <time.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */

#if defined(__linux__) || defined(__CYGWIN__)
#include <sys/mount.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <mntent.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#endif

#if defined(__CYGWIN__)
#include <cygwin/version.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <sys/cygwin.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#endif


#include <implementation_support_map.c>
#include <sa_unsupported.h>


/* The version to return from the "virtualization" operation of vrpsx_utsname_t
 * for this implementation of the virtual POSIX-like interface.
 *
 * Currently the numeric portion of this version is essentially a hash of
 * the source files, so that it changes when they do.  You could consider
 * pinning it down on some kind of compatibility branch, but only at the
 * risk of undermining interface client workarounds that are based upon
 * the assumption that a particular implementation bug remains present.
 *
 * Currently we hash using the interface identifier scheme, rather
 * than a sparate tool.  However, one could argue that comments should
 * be excluded in the case of identifying an implementation.  We could
 * either write a separate tool or augment the existing one.
 */
#define DIRECT_VERSION_STR "vrpsx direct " direct_IDENTIFIER_STR


#define MALLOC(TYPE)  ((TYPE *)malloc(sizeof(TYPE)))

#define ZERO(PTR)     (memset((PTR), 0, sizeof *(PTR)))


#if defined(__CYGWIN__)

/* On Cygwin a POSIX thread ID is a pointer; in order
 * to avoid compiler warnings we cast through uintptr_t.
 * Each of these macros evaluates its argument exactly once.
 */
#define VRPSX_THREAD_ID(ID)   ((vrpsx_thread_t)(uintptr_t)ID)
#define NATIVE_THREAD_ID(ID)  ((pthread_t)(uintptr_t)ID)

#else /* not Cygwin */

/* Early versions of POSIX required pthread_t to be an integer type,
 * and in most cases it still is, in which case we can directly cast.
 * Each of these macros evaluates its argument exactly once.
 */
#define VRPSX_THREAD_ID(ID)   ((vrpsx_thread_t)ID)
#define NATIVE_THREAD_ID(ID)  ((pthread_t)ID)

#endif /* Cygwin? */


extern int direct_interface_identifier_compare(
    void const *va,
    void const *vb)
{
    vrpsx_interface_identifier_t ia
            = *(vrpsx_interface_identifier_t const *)va;

    vrpsx_interface_identifier_t ib
            = *(vrpsx_interface_identifier_t const *)vb;

    return ia < ib
            ? -1
            : ia > ib
                ? +1
                : 0;
}


extern vrpsx_interface_identifier_t const *object_supports(
    object_binding_t const support_map[],
    size_t dimension,
    vrpsx_interface_identifier_t interface_identifier,
    size_t *offset)
{
    object_binding_t const *binding
            = (object_binding_t const *) bsearch(
                &interface_identifier,
                support_map,
                dimension,
                sizeof support_map[0],
                &direct_interface_identifier_compare);

    if (binding) {
        *offset = binding->offset;
        return &binding->provided;
    } else {
        *offset = 0;
        return NULL;
    }
}


#define DIRECT_ALLOCATION(PREFIX, TYPE)                 \
                                                        \
static void VRPSX_CAT2(PREFIX, _free_memory)(           \
    TYPE *allocator, /* unused */                       \
    void *memory)                                       \
{                                                       \
    free(memory);                                       \
}                                                       \
                                                        \
static void *VRPSX_CAT2(PREFIX, _allocate_memory)(      \
    TYPE *allocator, /* unused */                       \
    size_t size)                                        \
{                                                       \
    return malloc(size);                                \
}

DIRECT_ALLOCATION(allocator, vrpsx_allocator_t);

static vrpsx_allocator_end_t const allocator_end;

static vrpsx_allocator_vtbl_t const allocator_vtbl = {
    &allocator_free_memory,
    &allocator_allocate_memory,
    &allocator_end
};

/* Equivalent to direct_embedding_singleton.interface.system.allocator.
 *
 * If in future we trace vrpsx_system_t allocations, then this allocator
 * will differ from trace_embedding_singleton.interface.system.allocator
 * in that it performs no tracing.
 */
static vrpsx_allocator_t allocator_singleton = {
    &allocator_vtbl
};


/* We can share these methods across all virtual function
 * tables for interfaces that extend vrpsx_system_associate_t.
 */

static vrpsx_system_t *direct_get_system(void const *interface);
static vrpsx_system_t *trace_get_system(void const *interface);

#define VRPSX_GET_SYSTEM_METHOD(PREFIX, TYPE)                           \
 ((vrpsx_system_t *(*)(TYPE const *interface))&VRPSX_CAT2(PREFIX, _get_system))


/* The typical implementation of the "System Reference On Error" mechanism.
 *
 * This macro evaluates SROE exactly once.
 * If SROE == 0, then this macro does not evaluate either SROE or INTERFACE.
 * Otherwise:
 *    If ERN == 0, then this macro does NOT evaluate INTERFACE.
 *    If ERN != 0, then this macro evaluates INTERFACE twice.
 *
 * WARNING: If SROE != NULL && ERN != 0, then this macro must be
 * used BEFORE freeing or otherwise modifying (INTERFACE)->vtbl
 * or anything used by (INTERFACE)->vtbl->system.  (But currently,
 * all such "get_system" methods ignore their first argument.)
 */
#define VRPSX_SYSTEM_REFERENCE_ON_ERROR(ERN, SROE, INTERFACE)  {        \
    vrpsx_system_t **sroe_7db5e535 = (SROE);                            \
    if (sroe_7db5e535) {                                                \
        vrpsx_errno_t ern_7db5e535 = (ERN);                             \
        if (ern_7db5e535 == 0) {                                        \
            *sroe_7db5e535 = NULL;                                      \
        } else {                                                        \
            *sroe_7db5e535 = (INTERFACE)->vtbl->get_system((INTERFACE)); \
            VRPSX_CALL0(*sroe_7db5e535, acquire);                       \
        }                                                               \
    }                                                                   \
}


typedef struct direct_system_error_struct {
    vrpsx_system_error_t interface;
    vrpsx_reference_count_t reference_count;
    vrpsx_errno_t number;
    char text[17]; /* This size is appropriate to direct_enomem_singleton;
                    * for other errors the size is computed at run time. */
} direct_system_error_t;

static vrpsx_system_error_vtbl_t const direct_system_error_vtbl;

/* Statically allocated fallback.  Even for ENOMEM we
 * prefer the dynamically allocated object because it
 * may have better phrasing this hard-coded object.
 */
static direct_system_error_t direct_enomem_singleton = {
    {
        &direct_system_error_vtbl
    },
    1, /* always alive */
    ENOMEM,
    "Not enough space"
};

/* Allocates and constructs a new error object for the specified
 * error number.  Degrades gracefully under low memory conditions.
 */
static void direct_system_error_create(
    vrpsx_errno_t ern,
    vrpsx_system_error_t **out_error)
{
    direct_system_error_t *obj = NULL;
    char *temporary = NULL;
    size_t text_size;

#if defined(__CYGWIN__)

    /* Cygwin 1.5.25 does not provide strerror_r or strerror_l.
     * Cygwin 1.7.9 does provides strerror_r but not strerror_l.
     * But the man pages for both claim that their strerror
     * is reentrant, and this claim appears to be correct.
     */

    char unknown_error[14 + 1 + (((sizeof(int) * CHAR_BIT) + 2) / 3) + 1];

    temporary = strerror(ern);
    if (! temporary) {
        sprintf(unknown_error, "Unknown error %d", ern);
        temporary = unknown_error;
    }

#else /* not Cygwin */

#if POSIX_STRERROR_R

    /* We use POSIX strerror_r on progressively larger temporary buffers.
     */

    HYBRID_ALLOCATOR_STACK(tmp_stack, 256);
    hybrid_allocator_t tmp_allocator;
    size_t capacity;
    vrpsx_errno_t ern2;

    hybrid_allocator_construct_from_array(
        &tmp_allocator,
        &allocator_singleton,
        tmp_stack);

    temporary = VRPSX_CALL1(&tmp_allocator, allocate_available, &capacity);

    while (0 != (ern2 = strerror_r(ern, temporary, capacity))) {
        VRPSX_CALL1(&tmp_allocator, free_memory, temporary);
        temporary = NULL;

        if (ern2 != ERANGE) {
            temporary = VRPSX_CALL1(&tmp_allocator, allocate_available,
                    &capacity);
            sprintf(temporary, "Unknown error %d", ern);
            break;
        }

        if ((capacity << 1) <= capacity) {
            break;
        }
        capacity <<= 1;

        temporary = VRPSX_CALL1(&tmp_allocator, allocate_memory, capacity);
        if (! temporary) {
            break;
        }
    }

    if (! temporary) {
        temporary = VRPSX_CALL1(&tmp_allocator, allocate_available,
                &capacity);
        strcpy(temporary, "Not enough space.");
    }

#else /* non-POSIX strerror_r */

    /* We assume that the non-POSIX strerror_r is like the GNU one,
     * and use it with what we hope is a sufficiently large buffer.
     * If the buffer is too small, then the text gets truncated.
     */

    char buffer[1024];
    temporary = strerror_r(ern, buffer, sizeof buffer);

#endif /* POSIX_STRERROR_R? */

#endif /* Cygwin? */

    text_size = strlen(temporary) + 1;
    obj = (direct_system_error_t *)malloc(
        offsetof(direct_system_error_t, text) + text_size);
    if (! obj) {
#if ! defined(__CYGWIN__) && POSIX_STRERROR_R
        VRPSX_CALL1(&tmp_allocator, free_memory, temporary);
#endif
        VRPSX_REFERENCE_COUNT_INCR(&direct_enomem_singleton.reference_count);
        *out_error = &direct_enomem_singleton.interface;
        return;
    }

    obj->interface.vtbl = &direct_system_error_vtbl;
    obj->reference_count = 1;
    obj->number = ern;
    memcpy(obj->text, temporary, text_size);

#if ! defined(__CYGWIN__) && POSIX_STRERROR_R
    VRPSX_CALL1(&tmp_allocator, free_memory, temporary);
#endif

    *out_error = &obj->interface;
    return;
}

/* Like direct_system_error_create but downcasts to narrower interface.
 */
extern void direct_error_create(
    vrpsx_errno_t ern,
    vrpsx_error_t **out_error)
{
    direct_system_error_create(ern, (vrpsx_system_error_t **)out_error);
}

static void direct_system_error_replace_interface(
    vrpsx_system_error_t **identified,
    vrpsx_interface_identifier_t interface_identifier,
    void **out_alternative,
    vrpsx_error_t **out_error)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_system_error_t, , obj, interface, *identified);
    size_t offset;

    *out_alternative = NULL;
    *out_error = NULL;

    vrpsx_interface_identifier_t const *provided
            = direct_system_error_supports(interface_identifier, &offset);

    if (! provided) {
        direct_error_create(ENOSYS, out_error);
        return;
    }

    switch (*provided) {
        case vrpsx_system_error_IDENTIFIER: {
            vrpsx_system_error_t *system_error = &obj->interface;
            *out_alternative = (char *)system_error + offset;
            *identified = NULL;
            return;
        }
    }

    assert(0 /* provided interface needs case in above switch */);
    direct_error_create(ENOSYS, out_error);
}

static void direct_system_error_release(
    vrpsx_system_error_t **error)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_system_error_t, , obj, interface, *error);
    vrpsx_reference_count_t new_count;

    *error = NULL;

    VRPSX_REFERENCE_COUNT_DECR_READ(&obj->reference_count, &new_count);

    if (new_count == 0) {
        assert(obj != &direct_enomem_singleton);
        free(obj);
    }
}

static void direct_system_error_acquire_op(
    vrpsx_system_error_t *error)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_system_error_t, , obj, interface, error);
    VRPSX_REFERENCE_COUNT_INCR(&obj->reference_count);
}

static char const *direct_system_error_text(
    vrpsx_system_error_t const *error)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_system_error_t, const, obj, interface, error);
    return obj->text;
}

static vrpsx_errno_t direct_system_error_number(
    vrpsx_system_error_t const *system_error)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_system_error_t, const, obj, interface, system_error);
    return obj->number;
}

static vrpsx_identified_end_t const identified_end;

static vrpsx_counted_identified_end_t const counted_identified_end;

static vrpsx_error_end_t const error_end;

static vrpsx_system_error_end_t const system_error_end;

static vrpsx_system_error_vtbl_t const direct_system_error_vtbl = {
    vrpsx_system_error_IDENTIFIER,
    &direct_system_error_replace_interface,
    &identified_end,
    &direct_system_error_release,
    &direct_system_error_acquire_op,
    &counted_identified_end,
    &direct_system_error_text,
    &error_end,
    &direct_system_error_number,
    &system_error_end
};


/* Protects access to trace_file and *trace_file.
 */
static pthread_mutex_t trace_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Trace output goes to this file; NULL until tracing is enabled.
 */
static FILE *trace_file = NULL;

static trace_state_t const trace_argument_tail
    = { ", ", ")\n", &trace_argument_tail };

static trace_state_t const trace_argument_list
    = { "(", "()\n", &trace_argument_tail };

static trace_state_t const trace_value_tail
    = { ", ", "\n", &trace_value_tail };

static trace_state_t const trace_value_list
    = { ": ", "\n", &trace_value_tail };

static trace_state_t const trace_extra_output_tail
    = { ", ", ")\n", &trace_extra_output_tail };

static trace_state_t const trace_extra_output_list
    = { " (", "\n", &trace_extra_output_tail };

static trace_state_t const trace_return
    = { " = ", "\n", &trace_extra_output_list };

static trace_state_t const space_separated_tail
    = { " ", "", &space_separated_tail };

static trace_state_t const space_separated_list
    = { "", "", &space_separated_tail };

static trace_state_t const bitwise_or_tail
    = { " | ", ")", &bitwise_or_tail };

static trace_state_t const bitwise_or
    = { "(", "0", &bitwise_or_tail };

static trace_state_t const brace_tail
    = { ", ", "}", &brace_tail };

static trace_state_t const brace_list
    = { "{", "{}", &brace_tail };

static trace_state_t const data_constructor
    = { "", "{}", &brace_list };

static trace_state_t const association_tail
    = { " : ", "", &association_tail };

static trace_state_t const association
    = { "", "", &association_tail };

extern void direct_trace(
    FILE *new_trace_file)
{
    pthread_mutex_lock(&trace_mutex);

    if (trace_file) {
        fclose(trace_file);
    }
    trace_file = new_trace_file;

    pthread_mutex_unlock(&trace_mutex);
}

/* Prints the thread ID of the calling thread to trace_file.
 */
static void trace_thread_raw(
    vrpsx_thread_t thread)
{
    if (sizeof(pthread_t) * CHAR_BIT <= 32) {
        fprintf(trace_file, "%.8x", (unsigned)thread);
    } else {
        fprintf(trace_file, "%.16llx", (unsigned long long)thread);
    }
}

/* Prints the given timespec to trace_file; the argument must be non-NULL.
 */
static void trace_timespec_raw(
    struct timespec const *ts)
{
    if (sizeof(ts->tv_sec) <= sizeof(long)) {
        fprintf(trace_file, "%ld.%09ld",
                (long)ts->tv_sec, (long)ts->tv_nsec);
    } else {
        fprintf(trace_file, "%lld.%09ld",
                (long long)ts->tv_sec, (long)ts->tv_nsec);
    }
}

extern void trace_begin(
    trace_stack_t *inner,
    trace_stack_t *outer,
    trace_state_t const *first)
{
    struct timespec now;

    inner->outer = outer;
    inner->current = first;

    if (outer) {
        fputs(outer->current->prefix, trace_file);
    } else {
        pthread_mutex_lock(&trace_mutex);

        if (clock_gettime(CLOCK_REALTIME, &now)) {
            now.tv_sec = 0;
            now.tv_nsec = 0;
        }

        trace_thread_raw(VRPSX_THREAD_ID(pthread_self()));
        fputs(": ", trace_file);

        trace_timespec_raw(&now);
        fputs(": ", trace_file);
    }
}

extern void trace_end(
    trace_stack_t *inner)
{
    trace_stack_t *outer = inner->outer;

    fputs(inner->current->closer, trace_file);

    /* Helps to detect accidental misuse.
     */
    inner->outer = NULL;
    inner->current = NULL;

    if (outer) {
        outer->current = outer->current->next;
    } else {
        fflush(trace_file);
        pthread_mutex_unlock(&trace_mutex);
    }
}

/* An optimized alternative to "trace_begin(stack, inner); trace_end(stack);",
 * where "inner->closer" is empty.
 */
static void trace_atom(
    trace_stack_t *stack)
{
    fputs(stack->current->prefix, trace_file);
    stack->current = stack->current->next;
}

/* Prints a C string literal representation of the given argument string.
 */
static void print_substring(
    FILE *fp,
    vrpsx_substring_t const *sub)
{
    char const *ptr = sub->begin;
    char const *end = ptr + sub->size;

    fputc('"', fp);

    for (; ptr != end; ++ptr) {
        switch (*ptr) {
            case '\t':
                fputs("\\\t", fp);
                break;

            case '\n':
                fputs("\\\n", fp);
                break;

            case '\r':
                fputs("\\\r", fp);
                break;

            case '"':
                fputs("\\\"", fp);
                break;

            case '\\':
                fputs("\\\\", fp);
                break;

            default:
                if (*ptr >= 0x20 && *ptr <= 0x7E) {
                    putc(*ptr, fp);
                } else {
                    fprintf(fp, "\\x%x", (unsigned)*ptr);

                    /* We escape hexadecimal digits that immediately
                     * follow a hexadecimal escape; otherwise they
                     * might be interpreted as more such digits.
                     */
                    while (ptr + 1 != end && isxdigit((unsigned char)ptr[1])) {
                        fprintf(fp, "\\x%x", (unsigned)*++ptr);
                    }
                    break;
                }
        }
    }

    fputc('"', fp);
}

/* Traces a C string literal representation of the given argument string.
 */
static void trace_substring(
    trace_stack_t *stack,
    vrpsx_substring_t const *sub)
{
    trace_atom(stack);
    print_substring(trace_file, sub);
}

extern void trace_decimal_int(
    trace_stack_t *stack,
    int value)
{
    trace_atom(stack);
    fprintf(trace_file, "%d", value);
}

extern void trace_null(
    trace_stack_t *stack)
{
    trace_atom(stack);
    fputs("NULL", trace_file);
}

extern void print_string(
    FILE *fp,
    char const *str)
{
    if (! str) {
        fputs("NULL", fp);
        return;
    }

    vrpsx_substring_t sub;
    sub.begin = str;
    sub.size = strlen(str);
    print_substring(fp, &sub);
}

extern void trace_string(
    trace_stack_t *stack,
    char const *str)
{
    trace_atom(stack);
    print_string(trace_file, str);
}

extern void trace_system_error(
    trace_stack_t *stack,
    vrpsx_errno_t ern)
{
    trace_decimal_int(stack, ern);

    if (ern) {
        vrpsx_system_error_t *error;
        direct_system_error_create(ern, &error);
        trace_string(stack, VRPSX_CALL0(error, text));
        VRPSX_COUNTED_REFERENCE_CLEAR(vrpsx_system_error_t, &error);
    }
}

extern void trace_enter_begin(
    trace_stack_t *inner,
    char const *func)
{
    trace_begin(inner, NULL, &trace_argument_list);
    fputs(func, trace_file);
}

extern void trace_leave_begin(
    trace_stack_t *inner,
    char const *func)
{
    trace_begin(inner, NULL, &trace_return);
    fputs(func, trace_file);
}

extern void trace_leave_begin_ern(
    trace_stack_t *inner,
    char const *func,
    vrpsx_errno_t ern)
{
    trace_leave_begin(inner, func);
    trace_system_error(inner, ern);
}

extern void trace_note_begin(
    trace_stack_t *inner,
    char const *name)
{
    trace_begin(inner, NULL, &trace_value_list);
    fputs(name, trace_file);
}

extern void trace_address(
    trace_stack_t *stack,
    void const *ptr)
{
    if (ptr) {
        trace_atom(stack);
        fprintf(trace_file, "%p", ptr);
    } else {
        trace_null(stack);
    }
}

extern void trace_uintptr(
    trace_stack_t *stack,
    uintptr_t value)
{
    trace_atom(stack);

    if (sizeof value <= sizeof(unsigned long)) {
        fprintf(trace_file, "%#lx", (unsigned long)value);
    } else {
        fprintf(trace_file, "%#llx", (unsigned long long)value);
    }
}

extern void trace_identifier(
    trace_stack_t *stack,
    char const *id)
{
    trace_atom(stack);
    fputs(id, trace_file);
}

extern void trace_interface_identifier(
    trace_stack_t *stack,
    vrpsx_interface_identifier_t id)
{
    trace_atom(stack);
    fprintf(trace_file, "%#04x", (unsigned)id);
}

extern void trace_flagset(
    trace_stack_t *stack,
    flag_definition_t const *flagdefs,
    vrpsx_uint64_t flags)
{
    trace_stack_t top;
    vrpsx_uint64_t named = 0;

    trace_begin(&top, stack, &bitwise_or);

    for (; flagdefs->name; ++flagdefs) {
        if ((flagdefs->mask & flags) == flagdefs->value) {
            named |= flagdefs->mask;
            trace_atom(&top);
            fputs(flagdefs->name, trace_file);
        }
    }

    flags &= ~ named;
    if (flags) {
        trace_atom(&top);
        fprintf(trace_file, "%#llx", (unsigned long long)flags);
    }

    trace_end(&top);
}

/* Traces "size" octets in hexadecimal, starting at "begin".
 */
static void trace_octets(
    trace_stack_t *stack,
    char const *begin,
    size_t size)
{
    trace_stack_t top;

    trace_begin(&top, stack, &space_separated_list);

    while (size--) {
        trace_atom(&top);
        fprintf(trace_file, "%.2x", (unsigned)(unsigned char)*begin++);
    }

    trace_end(&top);
}

/* Traces an argument expressing a thread ID.
 */
static void trace_thread(
    trace_stack_t *stack,
    vrpsx_thread_t thread)
{
    trace_atom(stack);
    trace_thread_raw(thread);
}

/* Traces an argument expressing a process ID.
 */
static void trace_pid(
    trace_stack_t *stack,
    vrpsx_pid_t pid)
{
    trace_atom(stack);
    fprintf(trace_file, "%llu", (unsigned long long)pid);
}

static char const *safe_strsignal(
    vrpsx_signo_t signum)
{
#if defined(__CYGWIN__)

    /* The Cygwin strsignal() implementation uses
     * static and/or thread-local storage for thread safety.
     */

    return strsignal(signum);

#elif defined(__linux__)

    /* The Linux strsignal() implementation uses
     * static and/or thread-local storage for thread safety.
     *
     * In older Linux environments we must allow GNU extensions
     * generally or else declare strsignal ourselves.  We prefer
     * the latter alternative, at least for now.
     */

    extern char *strsignal(int signo);

    return strsignal(signum);

#else

#error Thread-safe strsignal not yet implemented on this platform.

#endif
}

/* Traces an argument expressing a signal number.
 */
static void trace_signo(
    trace_stack_t *stack,
    vrpsx_signo_t signum)
{
    trace_stack_t top;
    trace_begin(&top, stack, &association);
    trace_decimal_int(&top, signum);
    trace_string(&top, safe_strsignal(signum));
    trace_end(&top);
}

/* Traces an argument expressing a set of signal numbers,
 * unless the argument is NULL, in which case the effect
 * is the same as if you had called "trace_null(stack)".
 */
static void trace_sigset(
    trace_stack_t *stack,
    sigset_t const *set)
{
    trace_stack_t top;
    vrpsx_signo_t signum;
    size_t count[2];
    int member;

    if (! set) {
        trace_null(stack);
        return;
    }

    /* Count the number of members and the number of non-members.
     */
    count[0] = 0;
    count[1] = 0;

    /* We do not have a good, portable way to know the full range of
     * valid signal numbers.  Instead, we assume that signal numbers
     * are positive, and that sigset_t is a simple bit map, and then
     * infer an upper bound on the range.
     */
    for (signum = 1; signum <= (sizeof *set) * CHAR_BIT; ++signum) {
        /* According to POSIX, sigismember is supposed to return 1 if
         * the signal is in the set, 0 if it is not, and -1 if there
         * is some error, such as the signal number being invalid.
         */
        member = sigismember(set, signum);
        if (member == 0 || member == 1) {
            ++ count[member];
        }
    }

    /* List the members or the non-members, whichever list is smaller.
     * When listing the non-members, prefix the result with ~.
     */
    if (count[1] < count[0]) {
        trace_begin(&top, stack, &brace_list);
        member = 1;
    } else {
        trace_begin(&top, stack, &data_constructor);
        trace_identifier(&top, "~");
        member = 0;
    }

    for (signum = 1; signum <= (sizeof *set) * CHAR_BIT; ++signum) {
        if (sigismember(set, signum) == member) {
            trace_signo(&top, signum);
        }
    }

    trace_end(&top);
}

static void trace_vrpsx_sigset(
    trace_stack_t *stack,
    vrpsx_sigset_t const *set);


typedef struct direct_utsname_struct {
    vrpsx_utsname_t interface;
    struct utsname native;
} direct_utsname_t;

static void direct_utsname_disown(
    vrpsx_utsname_t **utsname)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_utsname_t, , obj, interface, *utsname);
    *utsname = NULL;
    free(obj);
}

static char const *direct_utsname_machine(
    vrpsx_utsname_t const *utsname)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_utsname_t,const, obj, interface, utsname);
    return obj->native.machine;
}

static char const *direct_utsname_nodename(
    vrpsx_utsname_t const *utsname)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_utsname_t,const, obj, interface, utsname);
    return obj->native.nodename;
}

static char const *direct_utsname_release(
    vrpsx_utsname_t const *utsname)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_utsname_t,const, obj, interface, utsname);
    return obj->native.release;
}

static char const *direct_utsname_sysname(
    vrpsx_utsname_t const *utsname)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_utsname_t,const, obj, interface, utsname);
    return obj->native.sysname;
}

static char const *direct_utsname_version(
    vrpsx_utsname_t const *utsname)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_utsname_t,const, obj, interface, utsname);
    return obj->native.version;
}

static char const *direct_utsname_virtualization(
    vrpsx_utsname_t const *utsname)
{
    return DIRECT_VERSION_STR;
}

static vrpsx_utsname_end_t const utsname_end;

static vrpsx_utsname_vtbl_t const direct_utsname_vtbl = {
    &direct_utsname_disown,
    &direct_utsname_machine,
    &direct_utsname_nodename,
    &direct_utsname_release,
    &direct_utsname_sysname,
    &direct_utsname_version,
    &direct_utsname_virtualization,
    &utsname_end
};


typedef struct direct_passwd_struct {
    vrpsx_passwd_t interface;
    struct passwd native;
} direct_passwd_t;

static void direct_passwd_disown(
    vrpsx_passwd_t **passwd)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_passwd_t, , obj, interface, *passwd);
    *passwd = NULL;
    free(obj);
}

static char const *direct_passwd_dir(
    vrpsx_passwd_t const *passwd)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_passwd_t, const, obj, interface, passwd);
    return obj->native.pw_dir;
}

static char const *direct_passwd_name(
    vrpsx_passwd_t const *passwd)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_passwd_t, const, obj, interface, passwd);
    return obj->native.pw_name;
}

static char const *direct_passwd_shell(
        vrpsx_passwd_t const *passwd)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_passwd_t, const, obj, interface, passwd);
    return obj->native.pw_shell;
}

static vrpsx_gid_t direct_passwd_gid(
        vrpsx_passwd_t const *passwd)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_passwd_t, const, obj, interface, passwd);
    return obj->native.pw_gid;
}

static vrpsx_uid_t direct_passwd_uid(
        vrpsx_passwd_t const *passwd)
{
    VRPSX_UPCAST_NONNULL_DECL(direct_passwd_t, const, obj, interface, passwd);
    return obj->native.pw_uid;
}

static vrpsx_passwd_end_t const passwd_end;

static vrpsx_passwd_vtbl_t const direct_passwd_vtbl = {
    &direct_passwd_disown,
    &direct_passwd_dir,
    &direct_passwd_name,
    &direct_passwd_shell,
    &direct_passwd_gid,
    &direct_passwd_uid,
    &passwd_end
};


/* Converts a timestamp from the form presented by the
 * interface to the native form of the POSIX-like system.
 *
 * Evaluates macro arguments exactly once.
 */
#define DIRECT_TIMESPEC_TO_NATIVE(NATIVE, PRESENTED)  {         \
    struct timespec *native_083f5b08 = (NATIVE);                \
    vrpsx_timespec_t const *presented_083f5b08 = (PRESENTED);   \
                                                                \
    native_083f5b08->tv_sec = presented_083f5b08->tv_sec;       \
    native_083f5b08->tv_nsec = presented_083f5b08->tv_nsec;     \
}

/* Converts a timestamp to the form presented by the
 * interface from the native form of the POSIX-like system.
 *
 * Evaluates macro arguments exactly once.
 */
#define DIRECT_TIMESPEC_FROM_NATIVE(PRESENTED, NATIVE)  {       \
    vrpsx_timespec_t *presented_5c06e227 = (PRESENTED);         \
    struct timespec const *native_5c06e227 = (NATIVE);          \
                                                                \
    presented_5c06e227->tv_sec = native_5c06e227->tv_sec;       \
    presented_5c06e227->tv_nsec = native_5c06e227->tv_nsec;     \
}


typedef struct direct_pointer_to_sigset_struct {
    vrpsx_sigset_t interface;
    sigset_t *native;
} direct_pointer_to_sigset_t;

typedef struct direct_bundled_sigset_struct {
    direct_pointer_to_sigset_t pointer;
    sigset_t native;
} direct_bundled_sigset_t;

typedef struct direct_bundled_sigset_alignof_struct {
    char bump;
    direct_bundled_sigset_t object;
} direct_bundled_sigset_alignof_t;

#define DIRECT_BUNDLED_SIGSET_ALIGNOF                   \
    (offsetof(direct_bundled_sigset_alignof_t, object))

static vrpsx_sigset_vtbl_t const direct_pointer_to_sigset_vtbl;

static void direct_pointer_to_sigset_init(
    sigset_t *native,
    direct_pointer_to_sigset_t *obj)
{
    obj->interface.vtbl = &direct_pointer_to_sigset_vtbl;
    obj->native = native;
}

static void direct_pointer_to_sigset_assign_sigset(
    vrpsx_sigset_t *set,
    vrpsx_sigset_t const *source)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, , obj, interface, set);
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, const, source_obj, interface, source);

    assert(source->vtbl == &direct_pointer_to_sigset_vtbl);

    /* POSIX requires that sigset_t have integer or structure type,
     * and it does not distinguish construction and destruction from
     * other modifications, and it does not provide a copy operation.
     * So it seems safe to use C assignment for copying.
     */
    *obj->native = *source_obj->native;
}

static void direct_pointer_to_sigset_swap_sigset(
    vrpsx_sigset_t *set,
    vrpsx_sigset_t *other)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, , obj, interface, set);
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, , other_obj, interface, other);
    sigset_t tmp;

    assert(other->vtbl == &direct_pointer_to_sigset_vtbl);

    /* See the comments in direct_pointer_to_sigset_assign_sigset()
     * for why we believe that C assignment is safe.
     */
    tmp = *obj->native;
    *obj->native = *other_obj->native;
    *other_obj->native = tmp;
}

static vrpsx_errno_t direct_pointer_to_sigset_empty(
    vrpsx_sigset_t *set)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, , obj, interface, set);
    if (sigemptyset(obj->native)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_pointer_to_sigset_fill(
    vrpsx_sigset_t *set)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, , obj, interface, set);
    if (sigfillset(obj->native)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_pointer_to_sigset_del(
    vrpsx_sigset_t *set,
    vrpsx_signo_t signum)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, , obj, interface, set);
    if (sigdelset(obj->native, signum)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_pointer_to_sigset_add(
    vrpsx_sigset_t *set,
    vrpsx_signo_t signum)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, , obj, interface, set);
    if (sigaddset(obj->native, signum)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_pointer_to_sigset_ismember(
    vrpsx_sigset_t const *set,
    vrpsx_signo_t signum,
    int *out_is_member)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, const, obj, interface, set);
    int ret = sigismember(obj->native, signum);
    if (ret < 0) {
        /* The value of *out_is_member is unspecified on error.
         * Should the interface specify a value for this case?
         */
        return errno;
    }
    *out_is_member = !!ret;
    return 0;
}

static vrpsx_system_associate_end_t const system_associate_end;

static vrpsx_sigset_end_t const sigset_end;

static vrpsx_sigset_vtbl_t const direct_pointer_to_sigset_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_sigset_t),
    &system_associate_end,
    &direct_pointer_to_sigset_assign_sigset,
    &direct_pointer_to_sigset_swap_sigset,
    &direct_pointer_to_sigset_empty,
    &direct_pointer_to_sigset_fill,
    &direct_pointer_to_sigset_del,
    &direct_pointer_to_sigset_add,
    &direct_pointer_to_sigset_ismember,
    &sigset_end
};

/* If "set == NULL", then the effect is the same as "trace_null(stack)".
 * Otherwise, if "set->vtbl == &direct_pointer_to_sigset_vtbl", then this
 * function applies trace_sigset to the contained native signal set.
 * For interfaces to other implementations it prints the addresses
 * "set" and "set->vtbl".
 */
static void trace_vrpsx_sigset(
    trace_stack_t *stack,
    vrpsx_sigset_t const *set)
{
    sigset_t const *native = NULL;

    if (set) {
        if (set->vtbl != &direct_pointer_to_sigset_vtbl) {
            /* Trace interface to unrecognized implementation by tracing
             * its address and the address of its function pointer table.
             */
            trace_stack_t assoc, ctor;
            trace_begin(&assoc, stack, &association);
            trace_address(&assoc, set);
            trace_begin(&ctor, &assoc, &data_constructor);
            trace_identifier(&ctor, "vrpsx_sigset_t");
            trace_address(&ctor, set->vtbl);
            trace_end(&ctor);
            trace_end(&assoc);
            return;
        } else {
            VRPSX_UPCAST_NONNULL_DECL(
                direct_pointer_to_sigset_t, const, obj, interface, set);
            native = obj->native;
        }
    }

    trace_sigset(stack, native);
}


typedef struct direct_sigaction_struct {
    vrpsx_sigaction_t interface;
    direct_pointer_to_sigset_t pointer_to_mask;
    struct sigaction native;
} direct_sigaction_t;

typedef struct direct_sigaction_alignof_struct {
    char bump;
    direct_sigaction_t object;
} direct_sigaction_alignof_t;

#define DIRECT_SIGACTION_ALIGNOF                        \
    (offsetof(direct_sigaction_alignof_t, object))

static vrpsx_sigaction_vtbl_t const direct_sigaction_vtbl;

static vrpsx_errno_t direct_sigaction_init(
    direct_sigaction_t *obj)
{
    obj->interface.vtbl = &direct_sigaction_vtbl;

    ZERO(&obj->native);
    obj->native.sa_flags = 0;
    obj->native.sa_handler = NULL;
    if (sigemptyset(&obj->native.sa_mask)) {
        return errno;
    }

    direct_pointer_to_sigset_init(
        &obj->native.sa_mask, &obj->pointer_to_mask);
    return 0;
}

static void direct_sigaction_assign_sigaction(
    vrpsx_sigaction_t *action,
    vrpsx_sigaction_t const *source)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, , obj, interface, action);
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, const, source_obj, interface, source);

    assert(source->vtbl == &direct_sigaction_vtbl);

    /* POSIX does not distinguish construction and destruction from other
     * modifications on struct sigaction, and it does not provide a copy
     * operation.  So it seems safe to use C struct assignment for copying.
     */
    obj->native = source_obj->native;
}

static void direct_sigaction_swap_sigaction(
    vrpsx_sigaction_t *action,
    vrpsx_sigaction_t *other)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, , obj, interface, action);
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, , other_obj, interface, other);
    struct sigaction tmp;

    assert(other->vtbl == &direct_sigaction_vtbl);

    /* See the comments in direct_sigaction_assign_sigaction()
     * for why we believe that C struct assignment is safe.
     */
    tmp = obj->native;
    obj->native = other_obj->native;
    other_obj->native = tmp;
}

static vrpsx_sa_handler_t *direct_sigaction_get_handler(
    vrpsx_sigaction_t const *action)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, const, obj, interface, action);
    return (obj->native.sa_flags & SA_SIGINFO) ? 0 : obj->native.sa_handler;
}

static void direct_sigaction_set_handler(
    vrpsx_sigaction_t *action,
    vrpsx_sa_handler_t *handler)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, , obj, interface, action);
    obj->native.sa_handler = handler;
    obj->native.sa_flags = obj->native.sa_flags & (~ SA_SIGINFO);
}

static vrpsx_sa_sigaction_t *direct_sigaction_get_sigaction(
    vrpsx_sigaction_t const *action)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, const, obj, interface, action);

    void (*sigaction)(int, siginfo_t *, void *)
        = (obj->native.sa_flags & SA_SIGINFO) ? obj->native.sa_sigaction : 0;

    /* Replace the implementation type siginfo_t with the abstract type
     * vrpsx_siginfo_t; because these are both struct types, their pointer
     * types should be binary compatible.  Note that we never dereference
     * though vrpsx_siginfo_t--it is just a placeholder for siginfo_t.
     */
    return (void (*)(int, vrpsx_siginfo_t *, void *))sigaction;
}

static void direct_sigaction_set_sigaction(
    vrpsx_sigaction_t *action,
    vrpsx_sa_sigaction_t *handler)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, , obj, interface, action);

    /* Replace the abstract type vrpsx_siginfo_t with the implementation
     * type siginfo_t; because these are both struct types, their pointer
     * types should be binary compatible.  Note that we never dereference
     * though vrpsx_siginfo_t--it is just a placeholder for siginfo_t.
     */
    void (*sigaction)(int, siginfo_t *, void *) =
            (void (*)(int, siginfo_t *, void *))handler;

    obj->native.sa_sigaction = sigaction;
    obj->native.sa_flags = obj->native.sa_flags | SA_SIGINFO;
}

static vrpsx_sigset_t const *direct_sigaction_get_mask_const(
    vrpsx_sigaction_t const *action)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, const, obj, interface, action);
    return &obj->pointer_to_mask.interface;
}

static vrpsx_sigset_t *direct_sigaction_get_mask(
    vrpsx_sigaction_t *action)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, , obj, interface, action);
    return &obj->pointer_to_mask.interface;
}

static vrpsx_sa_flags_t direct_sigaction_get_flags(
    vrpsx_sigaction_t const *action)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, const, obj, interface, action);
    return obj->native.sa_flags;
}

#ifndef SA_NOCLDWAIT
#define SA_NOCLDWAIT SA_UNSUPPORTED
#endif

#ifndef SA_ONSTACK
#define SA_ONSTACK SA_UNSUPPORTED
#endif

static vrpsx_errno_t direct_sigaction_set_flags(
    vrpsx_sigaction_t *action,
    vrpsx_sa_flags_t flags)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_sigaction_t, , obj, interface, action);

    obj->native.sa_flags =
            ( obj->native.sa_flags & SA_SIGINFO )
            | ( flags & (~ (SA_SIGINFO | SA_UNSUPPORTED)) );

    return (flags & SA_UNSUPPORTED) ? ENOSYS : 0;
}

static vrpsx_sigaction_end_t const sigaction_end;

static vrpsx_sigaction_vtbl_t const direct_sigaction_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_sigaction_t),
    &system_associate_end,
    &direct_sigaction_assign_sigaction,
    &direct_sigaction_swap_sigaction,
    &direct_sigaction_get_handler,
    &direct_sigaction_set_handler,
    &direct_sigaction_get_sigaction,
    &direct_sigaction_set_sigaction,
    &direct_sigaction_get_mask_const,
    &direct_sigaction_get_mask,
    &direct_sigaction_get_flags,
    &direct_sigaction_set_flags,
    &sigaction_end
};


/* The callback information passed to "thread_create".
 */
typedef struct trace_thread_callback_struct {
    void *(*function)(void*);
    void *argument;
} trace_thread_callback_t;

static char const thread_callback[] = "thread_callback";

/* Traces the end of the execution of a thread
 * that called pthread_exit() or was cancelled.
 */
static void trace_thread_cancellation(
    void *ptr)
{
    trace_thread_callback_t *delegate = (trace_thread_callback_t *)ptr;
    trace_stack_t top;

    trace_leave_begin(&top, thread_callback);
    trace_identifier(&top, "cancelled");
    trace_address(&top, delegate->function);
    trace_address(&top, delegate->argument);
    trace_end(&top);
}

/* Traces the beginning and the end of the execution
 * of a thread callback given by its argument.
 */
static void *trace_thread_callback(
    void *ptr)
{
    trace_thread_callback_t *delegate = (trace_thread_callback_t *)ptr;
    trace_stack_t top;
    void *result;

    /* Arrange to free the allocated delegation information.
     */
    pthread_cleanup_push(&free, ptr);

    /* Trace the callback that we are about to invoke in this new thread.
     */
    trace_enter_begin(&top, thread_callback);
    trace_address(&top, delegate->function);
    trace_address(&top, delegate->argument);
    trace_end(&top);

    /* Arrange to trace cancellation, if it happens.
     */
    pthread_cleanup_push(&trace_thread_cancellation, ptr);

    /* Actually invoke the client-specified callback.
     */
    result = (*delegate->function)(delegate->argument);

    /* We will trace the normal return, but just in case an asynchronous
     * cancellation arrives, we first unhook the cancellation handler,
     * so that we do not risk deadlock on the trace mutex.  (Ideally all
     * of our trace code would protect against an asynchronous cancellation
     * leaving the trace mutex locked, but that requires more investigation.)
     */
    pthread_cleanup_pop(0);

    /* Trace the normal return.
     */
    trace_leave_begin(&top, thread_callback);
    trace_address(&top, result);
    trace_address(&top, delegate->function);
    trace_address(&top, delegate->argument);
    trace_end(&top);

    /* And finally, free the delegation information.
     */
    pthread_cleanup_pop(1);

    return result;
}


typedef struct direct_thread_attr_struct {
    vrpsx_thread_attr_t interface;
    pthread_attr_t native;
} direct_thread_attr_t;

static vrpsx_errno_t direct_thread_attr_disown(
    vrpsx_thread_attr_t **attr,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, , obj, interface, *attr);
    vrpsx_errno_t ern;

    *attr = NULL;

    ern = pthread_attr_destroy(&obj->native);
    VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
    free(obj);
    return ern;
}

static vrpsx_errno_t direct_thread_attr_create(
    vrpsx_thread_attr_t const *attr,
    vrpsx_thread_t *thread,
    void *(*function)(void*),
    void *argument)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, const, obj, interface, attr);

    pthread_t tid;
    vrpsx_errno_t ern = 0;

    ern = pthread_create(&tid, &obj->native, function, argument);

    /* In the case of an error, POSIX leaves the thread ID value undefined.
     * Let us help to detect accidental misuse of that undefined value by
     * choosing a distinctive constant for that value ("0DEF" = "unDEFined").
     */
    *thread = ern ? 0x0DEF0DEF0DEF0DEFLL : VRPSX_THREAD_ID(tid);

    return ern;
}

static vrpsx_errno_t trace_thread_attr_create(
    vrpsx_thread_attr_t const *attr,
    vrpsx_thread_t *thread,
    void *(*function)(void*),
    void *argument)
{
    static char const me[] = "thread_attr::create";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, attr);
    trace_address(&top, thread);
    trace_address(&top, function);
    trace_address(&top, argument);
    trace_end(&top);

    /* Rather than directly invoking the caller-supplied function,
     * we invoke our own wrapper that will trace the beginning and
     * the end of execution of that thread.
     */
    vrpsx_errno_t ern;
    trace_thread_callback_t *delegate = MALLOC(trace_thread_callback_t);
    if (! delegate) {
        ern = errno;
    } else {
        delegate->function = function;
        delegate->argument = argument;

        ern = direct_thread_attr_create(
            attr,
            thread,
            trace_thread_callback,
            delegate);

        if (ern) {
            /* Because we failed to create the thread, we failed to transfer
             * responsibility to that thread for freeing the delegation info.
             */
            free(delegate);
        }
    }

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && thread) {
        trace_thread(&top, *thread);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_thread_attr_getdetachstate(
    vrpsx_thread_attr_t const *attr,
    int *detach_state)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, const, obj, interface, attr);
    return pthread_attr_getdetachstate(&obj->native, detach_state);
}

static vrpsx_errno_t direct_thread_attr_setdetachstate(
    vrpsx_thread_attr_t *attr,
    int detach_state)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, , obj, interface, attr);
    return pthread_attr_setdetachstate(&obj->native, detach_state);
}

static vrpsx_errno_t direct_thread_attr_getscope(
    vrpsx_thread_attr_t const *attr,
    int *contention_scope)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, const, obj, interface, attr);
    return pthread_attr_getscope(&obj->native, contention_scope);
}

static vrpsx_errno_t direct_thread_attr_setscope(
    vrpsx_thread_attr_t *attr,
    int contention_scope)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, , obj, interface, attr);
    return pthread_attr_setscope(&obj->native, contention_scope);
}

static vrpsx_errno_t direct_thread_attr_getstacksize(
    vrpsx_thread_attr_t const *attr,
    size_t *stack_size)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, const, obj, interface, attr);
    return pthread_attr_getstacksize(&obj->native, stack_size);
}

static vrpsx_errno_t direct_thread_attr_setstacksize(
    vrpsx_thread_attr_t *attr,
    size_t stack_size)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_attr_t, , obj, interface, attr);
    return pthread_attr_setstacksize(&obj->native, stack_size);
}

static vrpsx_thread_attr_end_t const thread_attr_end;

#define THREAD_ATTR_VTBL_INITIALIZATION(PREFIX)                         \
                                                                        \
static vrpsx_thread_attr_vtbl_t const VRPSX_CAT2(PREFIX, _thread_attr_vtbl)={ \
    VRPSX_GET_SYSTEM_METHOD(PREFIX, vrpsx_thread_attr_t),               \
    &system_associate_end,                                              \
    &direct_thread_attr_disown,                                         \
    &VRPSX_CAT2(PREFIX, _thread_attr_create),                           \
    &direct_thread_attr_getdetachstate,                                 \
    &direct_thread_attr_setdetachstate,                                 \
    &direct_thread_attr_getscope,                                       \
    &direct_thread_attr_setscope,                                       \
    &direct_thread_attr_getstacksize,                                   \
    &direct_thread_attr_setstacksize,                                   \
    &thread_attr_end                                                    \
}

THREAD_ATTR_VTBL_INITIALIZATION(direct);
THREAD_ATTR_VTBL_INITIALIZATION(trace);


typedef struct direct_thread_key_struct {
    vrpsx_thread_key_t interface;
    pthread_key_t native;
} direct_thread_key_t;

static vrpsx_errno_t direct_thread_key_disown(
    vrpsx_thread_key_t **key,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_key_t, , obj, interface, *key);
    vrpsx_errno_t ern;

    *key = NULL;

    ern = pthread_key_delete(obj->native);
    VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
    free(obj);
    return ern;
}

static void *direct_thread_key_get(
    vrpsx_thread_key_t const *key)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_key_t, const, obj, interface, key);
    return pthread_getspecific(obj->native);
}

static vrpsx_errno_t direct_thread_key_set(
    vrpsx_thread_key_t *key,
    void const *value)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_thread_key_t, , obj, interface, key);
    return pthread_setspecific(obj->native, value);
}

static vrpsx_thread_key_end_t const thread_key_end;

static vrpsx_thread_key_vtbl_t const direct_thread_key_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_thread_key_t),
    &system_associate_end,
    &direct_thread_key_disown,
    &direct_thread_key_get,
    &direct_thread_key_set,
    &thread_key_end
};


typedef struct direct_mutex_struct {
    vrpsx_mutex_t interface;
    pthread_mutex_t native;
} direct_mutex_t;

static vrpsx_errno_t direct_mutex_disown(
    vrpsx_mutex_t **mutex,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_t, , obj, interface, *mutex);
    vrpsx_errno_t ern;

    *mutex = NULL;

    ern = pthread_mutex_destroy(&obj->native);
    VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
    if (! ern) {
        free(obj);
    }
    return ern;
}

static vrpsx_errno_t direct_mutex_unlock(
    vrpsx_mutex_t *mutex)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_t, , obj, interface, mutex);
    return pthread_mutex_unlock(&obj->native);
}

static vrpsx_errno_t direct_mutex_lock(
    vrpsx_mutex_t *mutex)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_t, , obj, interface, mutex);
    return pthread_mutex_lock(&obj->native);
}

static vrpsx_errno_t direct_mutex_trylock(
    vrpsx_mutex_t *mutex)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_t, , obj, interface, mutex);
    return pthread_mutex_trylock(&obj->native);
}

static vrpsx_mutex_end_t const mutex_end;

static vrpsx_mutex_vtbl_t const direct_mutex_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_mutex_t),
    &system_associate_end,
    &direct_mutex_disown,
    &direct_mutex_unlock,
    &direct_mutex_lock,
    &direct_mutex_trylock,
    &mutex_end
};


typedef struct direct_mutex_attr_struct {
    vrpsx_mutex_attr_t interface;
    pthread_mutexattr_t native;
} direct_mutex_attr_t;

static vrpsx_mutex_attr_vtbl_t const direct_mutex_attr_vtbl;

static vrpsx_errno_t direct_mutex_attr_disown(
    vrpsx_mutex_attr_t **attr,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_attr_t, , obj, interface, *attr);
    vrpsx_errno_t ern;

    *attr = NULL;

    ern = pthread_mutexattr_destroy(&obj->native);
    VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
    free(obj);
    return ern;
}

/* This is not just an attribute method; you can
 * pass NULL attributes to get default attributes.
 */
static vrpsx_errno_t direct_mutex_attr_create(
    vrpsx_mutex_attr_t const *attr,
    vrpsx_mutex_t **mutex)
{
    pthread_mutexattr_t const *attr_native = NULL;
    direct_mutex_t *obj;
    vrpsx_errno_t ern;

    *mutex = NULL;

    if (attr) {
        assert(attr->vtbl == &direct_mutex_attr_vtbl);

        attr_native = & VRPSX_UPCAST_NONNULL(
            direct_mutex_attr_t, interface, attr) -> native;
    }

    obj = MALLOC(direct_mutex_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_mutex_vtbl;
    ern = pthread_mutex_init(&obj->native, attr_native);
    if (ern) {
        ern = errno;
        free(obj);
        return ern;
    }

    *mutex = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_mutex_attr_gettype(
    vrpsx_mutex_attr_t const *attr,
    int *type)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_attr_t, const, obj, interface, attr);
    return pthread_mutexattr_gettype(&obj->native, type);
}

static vrpsx_errno_t direct_mutex_attr_settype(
    vrpsx_mutex_attr_t *attr,
    int type)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_attr_t, , obj, interface, attr);
    return pthread_mutexattr_settype(&obj->native, type);
}

static vrpsx_mutex_attr_end_t const mutex_attr_end;

static vrpsx_mutex_attr_vtbl_t const direct_mutex_attr_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_mutex_attr_t),
    &system_associate_end,
    &direct_mutex_attr_disown,
    &direct_mutex_attr_create,
    &direct_mutex_attr_gettype,
    &direct_mutex_attr_settype,
    &mutex_attr_end
};


typedef struct direct_rwlock_struct {
    vrpsx_rwlock_t interface;
    pthread_rwlock_t native;
} direct_rwlock_t;

static vrpsx_errno_t direct_rwlock_disown(
    vrpsx_rwlock_t **rwlock,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_rwlock_t, , obj, interface, *rwlock);
    vrpsx_errno_t ern;

    *rwlock = NULL;

    ern = pthread_rwlock_destroy(&obj->native);
    VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
    if (! ern) {
        free(obj);
    }
    return ern;
}

static vrpsx_errno_t direct_rwlock_unlock(
    vrpsx_rwlock_t *rwlock)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_rwlock_t, , obj, interface, rwlock);
    return pthread_rwlock_unlock(&obj->native);
}

static vrpsx_errno_t direct_rwlock_rdlock(
    vrpsx_rwlock_t *rwlock)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_rwlock_t, , obj, interface, rwlock);
    return pthread_rwlock_rdlock(&obj->native);
}

static vrpsx_errno_t direct_rwlock_wrlock(
    vrpsx_rwlock_t *rwlock)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_rwlock_t, , obj, interface, rwlock);
    return pthread_rwlock_wrlock(&obj->native);
}

static vrpsx_errno_t direct_rwlock_tryrdlock(
    vrpsx_rwlock_t *rwlock)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_rwlock_t, , obj, interface, rwlock);
    return pthread_rwlock_tryrdlock(&obj->native);
}

static vrpsx_errno_t direct_rwlock_trywrlock(
    vrpsx_rwlock_t *rwlock)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_rwlock_t, , obj, interface, rwlock);
    return pthread_rwlock_trywrlock(&obj->native);
}

static vrpsx_rwlock_end_t const rwlock_end;

static vrpsx_rwlock_vtbl_t const direct_rwlock_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_rwlock_t),
    &system_associate_end,
    &direct_rwlock_disown,
    &direct_rwlock_unlock,
    &direct_rwlock_rdlock,
    &direct_rwlock_wrlock,
    &direct_rwlock_tryrdlock,
    &direct_rwlock_trywrlock,
    &rwlock_end
};


typedef struct direct_rwlock_attr_struct {
    vrpsx_rwlock_attr_t interface;
} direct_rwlock_attr_t;

static vrpsx_rwlock_attr_vtbl_t const direct_rwlock_attr_vtbl;

static vrpsx_errno_t direct_rwlock_attr_disown(
    vrpsx_rwlock_attr_t **attr,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_rwlock_attr_t, , obj, interface, *attr);
    *attr = NULL;
    if (sroe) {
        *sroe = NULL;
    }
    free(obj);
    return 0;
}

/* This is not just an attribute method; you can
 * pass NULL attributes to get default attributes
 * (though currently that makes no difference
 * because there are no settable attributes).
 */
static vrpsx_errno_t direct_rwlock_attr_create(
    vrpsx_rwlock_attr_t const *attr,
    vrpsx_rwlock_t **rwlock)
{
    direct_rwlock_t *obj;
    vrpsx_errno_t ern;

    *rwlock = NULL;

    if (attr) {
        assert(attr->vtbl == &direct_rwlock_attr_vtbl);
    }

    obj = MALLOC(direct_rwlock_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_rwlock_vtbl;
    ern = pthread_rwlock_init(&obj->native, NULL);
    if (ern) {
        ern = errno;
        free(obj);
        return ern;
    }

    *rwlock = &obj->interface;
    return 0;
}

static vrpsx_rwlock_attr_end_t const rwlock_attr_end;

static vrpsx_rwlock_attr_vtbl_t const direct_rwlock_attr_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_rwlock_attr_t),
    &system_associate_end,
    &direct_rwlock_attr_disown,
    &direct_rwlock_attr_create,
    &rwlock_attr_end
};


typedef struct direct_cond_struct {
    vrpsx_cond_t interface;
    pthread_cond_t native;
} direct_cond_t;

static vrpsx_errno_t direct_cond_disown(
    vrpsx_cond_t **cond,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_cond_t, , obj, interface, *cond);
    vrpsx_errno_t ern;

    *cond = NULL;

    ern = pthread_cond_destroy(&obj->native);
    VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
    if (! ern) {
        free(obj);
    }
    return ern;
}

static vrpsx_errno_t direct_cond_broadcast(
    vrpsx_cond_t *cond)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_cond_t, , obj, interface, cond);
    return pthread_cond_broadcast(&obj->native);
}

static vrpsx_errno_t direct_cond_signal(
    vrpsx_cond_t *cond)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_cond_t, , obj, interface, cond);
    return pthread_cond_signal(&obj->native);
}

static vrpsx_errno_t direct_cond_wait(
    vrpsx_cond_t *cond,
    vrpsx_mutex_t *mutex)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_cond_t, , obj, interface, cond);
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_t, , mutex_obj, interface, mutex);

    assert(mutex->vtbl == &direct_mutex_vtbl);

    return pthread_cond_wait(&obj->native, &mutex_obj->native);
}

static vrpsx_errno_t direct_cond_timedwait(
    vrpsx_cond_t *cond,
    vrpsx_mutex_t *mutex,
    vrpsx_timespec_t const *absolute_time)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_cond_t, , obj, interface, cond);
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mutex_t, , mutex_obj, interface, mutex);
    struct timespec native_absolute_time;

    assert(mutex->vtbl == &direct_mutex_vtbl);

    ZERO(&native_absolute_time);
    DIRECT_TIMESPEC_TO_NATIVE(&native_absolute_time, absolute_time);

    return pthread_cond_timedwait(
        &obj->native, &mutex_obj->native, &native_absolute_time);
}

static vrpsx_cond_end_t const cond_end;

static vrpsx_cond_vtbl_t const direct_cond_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_cond_t),
    &system_associate_end,
    &direct_cond_disown,
    &direct_cond_broadcast,
    &direct_cond_signal,
    &direct_cond_wait,
    &direct_cond_timedwait,
    &cond_end
};


typedef struct direct_cond_attr_struct {
    vrpsx_cond_attr_t interface;
} direct_cond_attr_t;

static vrpsx_cond_attr_vtbl_t const direct_cond_attr_vtbl;

static vrpsx_errno_t direct_cond_attr_disown(
    vrpsx_cond_attr_t **attr,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_cond_attr_t, , obj, interface, *attr);
    *attr = NULL;
    if (sroe) {
        *sroe = NULL;
    }
    free(obj);
    return 0;
}

/* This is not just an attribute method; you can
 * pass NULL attributes to get default attributes
 * (though currently that makes no difference
 * because there are no settable attributes).
 */
static vrpsx_errno_t direct_cond_attr_create(
    vrpsx_cond_attr_t const *attr,
    vrpsx_cond_t **cond)
{
    direct_cond_t *obj;
    vrpsx_errno_t ern;

    *cond = NULL;

    if (attr) {
        assert(attr->vtbl == &direct_cond_attr_vtbl);
    }

    obj = MALLOC(direct_cond_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_cond_vtbl;
    ern = pthread_cond_init(&obj->native, NULL);
    if (ern) {
        ern = errno;
        free(obj);
        return ern;
    }

    *cond = &obj->interface;
    return 0;
}

static vrpsx_cond_attr_end_t const cond_attr_end;

static vrpsx_cond_attr_vtbl_t const direct_cond_attr_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_cond_attr_t),
    &system_associate_end,
    &direct_cond_attr_disown,
    &direct_cond_attr_create,
    &cond_attr_end
};


typedef struct direct_directory_struct {
    vrpsx_directory_t interface;
    vrpsx_dirent_t dirent_interface;
    DIR *native;
    struct dirent dirent_native; /* must be at end of struct */
} direct_directory_t;

static vrpsx_directory_vtbl_t const direct_directory_vtbl;
static vrpsx_dirent_vtbl_t const direct_directory_dirent_vtbl;

static vrpsx_errno_t direct_directory_create(
    DIR *native,                    /* ownership transfers */
    vrpsx_directory_t **dir)
{
    direct_directory_t *obj;
    size_t size;
    int fd;
    vrpsx_errno_t ern;

    *dir = NULL;

    fd = dirfd(native);
    if (fd == -1) {
        ern = errno;
        closedir(native);
        return ern;
    }

    size = offsetof(direct_directory_t, dirent_native.d_name)
            + fpathconf(fd, _PC_NAME_MAX) + 1;

    obj = (direct_directory_t *)malloc(size);
    if (! obj) {
        ern = errno;
        closedir(native);
        return ern;
    }

    obj->interface.vtbl = &direct_directory_vtbl;
    obj->dirent_interface.vtbl = &direct_directory_dirent_vtbl;
    obj->native = native;

    *dir = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_directory_disown(
    vrpsx_directory_t **dir,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_directory_t, , obj, interface, *dir);
    vrpsx_errno_t ern = 0;

    *dir = NULL;

    if (closedir(obj->native)) {
        ern = errno;
    }
    VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
    free(obj);
    return ern;
}

static vrpsx_errno_t direct_directory_dirfd(
    vrpsx_directory_t *dir,
    vrpsx_fd_t *out_fd)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_directory_t, , obj, interface, dir);
    *out_fd = dirfd(obj->native);
    if (*out_fd == -1) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_directory_readdir(
    vrpsx_directory_t *dir,
    vrpsx_dirent_t **next)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_directory_t, , obj, interface, dir);
    struct dirent *result_native;

    *next = NULL;

    if (readdir_r(obj->native, &obj->dirent_native, &result_native)) {
        return errno;
    }

    if (result_native) {
        *next = &obj->dirent_interface;
    }
    return 0;
}

static vrpsx_errno_t direct_directory_rewinddir(
    vrpsx_directory_t *dir)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_directory_t, , obj, interface, dir);
    errno = 0;
    rewinddir(obj->native);
    return errno;
}

static vrpsx_directory_end_t const directory_end;

static vrpsx_directory_vtbl_t const direct_directory_vtbl = {
    VRPSX_GET_SYSTEM_METHOD(direct, vrpsx_directory_t),
    &system_associate_end,
    &direct_directory_disown,
    &direct_directory_dirfd,
    &direct_directory_readdir,
    &direct_directory_rewinddir,
    &directory_end
};

static char const *direct_directory_dirent_name(
    vrpsx_dirent_t const *dir)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_directory_t, const, obj, dirent_interface, dir);
    return obj->dirent_native.d_name;
}

static vrpsx_ino_t direct_directory_dirent_ino(
    vrpsx_dirent_t const *dir)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_directory_t, const, obj, dirent_interface, dir);
    return obj->dirent_native.d_ino;
}

static vrpsx_dirent_end_t const dirent_end;

static vrpsx_dirent_vtbl_t const direct_directory_dirent_vtbl = {
    &direct_directory_dirent_name,
    &direct_directory_dirent_ino,
    &dirent_end
};


#if defined(__linux__) || defined(__CYGWIN__)

typedef struct direct_mount_struct {
    vrpsx_mount_t interface;
    vrpsx_mount_option_t option_interface;
    vrpsx_substring_t current_option;
    FILE *fp_native;
    struct mntent mntent_native;
#if defined(__CYGWIN__)
    int reported_cygdrive_prefix;
#endif
    /* WARNING: We are just guessing at the maximum size of the options list.
     */
    char buffer[PATH_MAX * 2 /*fsname, dir*/ + 80 /*type*/ + 4096 /*options*/];
} direct_mount_t;

#if defined(__CYGWIN__)

/* Copies "str" to "*current" and advances "*current" past the copied NUL,
 * unless doing so would advance "*current" past "end", in which case this
 * function does nothing except to fail with ENOMEM.
 */
static vrpsx_errno_t direct_mount_next_strcpy(
    char **current,
    char *end,
    char const *str)
{
    size_t size = strlen(str) + 1;
    if (end - *current < size) {
        return ENOMEM;
    }

    MEMCPY_END(*current, str, size);
    return 0;
}

#endif

static vrpsx_errno_t direct_mount_next_unfiltered(
    vrpsx_mount_t **mount,
    vrpsx_mount_t **tail,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, , obj, interface, *mount);
    struct mntent *pmnt = NULL;
    vrpsx_errno_t ern = 0;

    *mount = NULL;

    if (tail) {
        *tail = NULL;

#if defined(__CYGWIN__)

        /* If and when Cygwin adds support for getmntent_r
         * we can eliminate this workaround, which is thread-safe
         * only if no other threads are calling getmntent without
         * first locking the same mutex.
         */
        {
            static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
            vrpsx_errno_t lock_error = pthread_mutex_lock(&mutex);
            if (lock_error) {
                VRPSX_SYSTEM_REFERENCE_ON_ERROR(
                    lock_error, sroe, &obj->interface);
                return lock_error;
            }

            errno = 0;
            pmnt = getmntent(
                obj->fp_native);
            ern = errno;

            if (pmnt) {
                char *current = obj->buffer;
                char *end = obj->buffer + sizeof obj->buffer;

                obj->mntent_native = *pmnt;

                obj->mntent_native.mnt_fsname = current;
                if (direct_mount_next_strcpy(&current, end, pmnt->mnt_fsname)){
                    pthread_mutex_unlock(&mutex);
                    VRPSX_SYSTEM_REFERENCE_ON_ERROR(
                        ENOMEM, sroe, &obj->interface);
                    return ENOMEM;
                }

                obj->mntent_native.mnt_dir = current;
                if (direct_mount_next_strcpy(&current, end, pmnt->mnt_dir)) {
                    pthread_mutex_unlock(&mutex);
                    VRPSX_SYSTEM_REFERENCE_ON_ERROR(
                        ENOMEM, sroe, &obj->interface);
                    return ENOMEM;
                }

                obj->mntent_native.mnt_type = current;
                if (direct_mount_next_strcpy(&current, end, pmnt->mnt_type)) {
                    pthread_mutex_unlock(&mutex);
                    VRPSX_SYSTEM_REFERENCE_ON_ERROR(
                        ENOMEM, sroe, &obj->interface);
                    return ENOMEM;
                }

                obj->mntent_native.mnt_opts = current;
                if (direct_mount_next_strcpy(&current, end, pmnt->mnt_opts)) {
                    pthread_mutex_unlock(&mutex);
                    VRPSX_SYSTEM_REFERENCE_ON_ERROR(
                        ENOMEM, sroe, &obj->interface);
                    return ENOMEM;
                }

                pmnt = &obj->mntent_native;
            }

            lock_error = pthread_mutex_unlock(&mutex);
            if (lock_error) {
                VRPSX_SYSTEM_REFERENCE_ON_ERROR(
                    lock_error, sroe, &obj->interface);
                return lock_error;
            }
        }

#else /* not Cygwin */

        errno = 0;
        pmnt = getmntent_r(
            obj->fp_native,
            &obj->mntent_native,
            obj->buffer,
            sizeof obj->buffer);
        ern = errno;

#endif /* Cygwin? */
    }

    if (! pmnt) {
        errno = ern;
        endmntent(obj->fp_native);
        if (! ern) {
            ern = errno;
        }
        VRPSX_SYSTEM_REFERENCE_ON_ERROR(ern, sroe, &obj->interface);
        free(obj);
        return ern;
    }

    *tail = &obj->interface;
    if (sroe) {
        *sroe = NULL;
    }
    return 0;
}

static vrpsx_errno_t direct_mount_device(
    vrpsx_mount_t const *mount,
    vrpsx_substring_t *device)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, const, obj, interface, mount);
    device->begin = obj->mntent_native.mnt_fsname;
    device->size = strlen(device->begin);
    return 0;
}

static vrpsx_errno_t trace_mount_device(
    vrpsx_mount_t const *mount,
    vrpsx_substring_t *device)
{
    static char const me[] = "mount_device";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, mount);
    trace_address(&top, device);
    trace_end(&top);

    vrpsx_errno_t ern = direct_mount_device(
        mount,
        device);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && device) {
        trace_substring(&top, device);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_mount_root(
    vrpsx_mount_t const *mount,
    vrpsx_substring_t *root)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, const, obj, interface, mount);
    root->begin = obj->mntent_native.mnt_dir;
    root->size = strlen(root->begin);
    return 0;
}

static vrpsx_errno_t trace_mount_root(
    vrpsx_mount_t const *mount,
    vrpsx_substring_t *root)
{
    static char const me[] = "mount_root";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, mount);
    trace_address(&top, root);
    trace_end(&top);

    vrpsx_errno_t ern = direct_mount_root(
        mount,
        root);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && root) {
        trace_substring(&top, root);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_mount_type(
    vrpsx_mount_t const *mount,
    vrpsx_substring_t *type)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, const, obj, interface, mount);
    type->begin = obj->mntent_native.mnt_type;
    type->size = strlen(type->begin);
    return 0;
}

static vrpsx_errno_t trace_mount_type(
    vrpsx_mount_t const *mount,
    vrpsx_substring_t *type)
{
    static char const me[] = "mount_type";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, mount);
    trace_address(&top, type);
    trace_end(&top);

    vrpsx_errno_t ern = direct_mount_type(
        mount,
        type);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && type) {
        trace_substring(&top, type);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_mount_mount_option_next_unfiltered(
    vrpsx_mount_option_t **option,
    vrpsx_mount_option_t **tail,
    vrpsx_system_t **sroe)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, , obj, option_interface, *option);
    char const *begin;
    char const *comma;

    *option = NULL;

    if (! tail) {
        if (sroe) {
            *sroe = NULL;
        }
        return 0;
    }

    *tail = NULL;

    begin = obj->current_option.begin + obj->current_option.size;
    while (*begin == ',') {
        ++begin;
    }
    obj->current_option.begin = begin;

    comma = strchr(begin, ',');
    obj->current_option.size = comma ? comma - begin : strlen(begin);
    if (obj->current_option.size) {
        *tail = &obj->option_interface;
    }
    if (sroe) {
        *sroe = NULL;
    }
    return 0;
}

static vrpsx_errno_t direct_mount_mount_option_text(
    vrpsx_mount_option_t const *option,
    vrpsx_substring_t *text)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, const, obj, option_interface, option);
    *text = obj->current_option;
    return 0;
}

static vrpsx_errno_t trace_mount_mount_option_text(
    vrpsx_mount_option_t const *option,
    vrpsx_substring_t *text)
{
    static char const me[] = "mount_option_text";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, option);
    trace_address(&top, text);
    trace_end(&top);

    vrpsx_errno_t ern = direct_mount_mount_option_text(
        option,
        text);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && text) {
        trace_substring(&top, text);
    }
    trace_end(&top);

    return ern;
}

#if defined(__CYGWIN__)

/* Returns true if this is the "noumount" option; otherwise false. Never fails.
 */
static int direct_mount_mount_option_is_noumount(
    vrpsx_mount_option_t const *option)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, const, obj, option_interface, option);
    return obj->current_option.size == 8
            && (0 == memcmp(obj->current_option.begin, "noumount", 8));
}

static vrpsx_errno_t direct_mount_next(
    vrpsx_mount_t **mount,
    vrpsx_mount_t **tail,
    vrpsx_system_t **sroe)
{
    direct_mount_t *obj;
    vrpsx_mount_option_t *option;
    char *slash;
    int is_cygdrive;
    vrpsx_errno_t ern;

    for (;;) {
        ern = direct_mount_next_unfiltered(mount, tail, sroe);
        if (! tail || ! *tail) {
            return ern;
        }
        obj = VRPSX_UPCAST_NONNULL(direct_mount_t, interface, *tail);

        is_cygdrive = 0;
        obj->current_option.begin = obj->mntent_native.mnt_opts;
        obj->current_option.size = 0;
        for (option = &obj->option_interface;;) {
            direct_mount_mount_option_next_unfiltered(
                &option, &option, NULL);  /* never fails */
            if (! option) {
                break;
            }
            if (direct_mount_mount_option_is_noumount(option)) {
                is_cygdrive = 1;
                break;
            }
        }

        if (! is_cygdrive) {
            if (sroe) {
                *sroe = NULL;
            }
            return 0;
        }

        if (! obj->reported_cygdrive_prefix) {
            obj->reported_cygdrive_prefix = 1;

            obj->mntent_native.mnt_type = "cygdrive";
            obj->mntent_native.mnt_fsname = "none";
            slash = strrchr(obj->mntent_native.mnt_dir, '/');
            if (slash) {
                slash[slash == obj->mntent_native.mnt_dir] = '\0';
            }

            if (! sroe) {
                *sroe = NULL;
            }
            return 0;
        }

        VRPSX_SINGLE_OWNER_LIST_MOVE(vrpsx_mount_t, mount, tail);
    }
}

static vrpsx_errno_t direct_mount_mount_option_next(
    vrpsx_mount_option_t **option,
    vrpsx_mount_option_t **tail,
    vrpsx_system_t **sroe)
{
    /* Skip "noumount" because we can change the cygdrive prefix
     * even if we cannot change the drive mounts that it implies.
     */
    for (;;) {
        direct_mount_mount_option_next_unfiltered(option, tail, sroe);

        if (! tail || ! *tail
                || ! direct_mount_mount_option_is_noumount(*tail)) {
            return 0;
        }

        VRPSX_SINGLE_OWNER_LIST_MOVE(vrpsx_mount_option_t, option, tail);
    }
}

#elif defined(__linux__)

#define direct_mount_next \
        direct_mount_next_unfiltered

#define direct_mount_mount_option_next \
        direct_mount_mount_option_next_unfiltered

#else /* operating system */

#error unsupported operating system

#endif /* operating system */

static vrpsx_errno_t trace_mount_mount_option_next(
    vrpsx_mount_option_t **option,
    vrpsx_mount_option_t **tail,
    vrpsx_system_t **sroe)
{
    static char const me[] = "mount_option_next";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, *option);
    trace_address(&top, tail);
    trace_address(&top, sroe);
    trace_end(&top);

    vrpsx_errno_t ern = direct_mount_mount_option_next(
        option,
        tail,
        sroe);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && tail) {
        trace_address(&top, *tail);
    }
    if (sroe) {
        trace_address(&top, *sroe);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t trace_mount_next(
    vrpsx_mount_t **mount,
    vrpsx_mount_t **tail,
    vrpsx_system_t **sroe)
{
    static char const me[] = "mount_next";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, *mount);
    trace_address(&top, tail);
    trace_address(&top, sroe);
    trace_end(&top);

    vrpsx_errno_t ern = direct_mount_next(
        mount,
        tail,
        sroe);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && tail) {
        trace_address(&top, *tail);
    }
    if (sroe) {
        trace_address(&top, *sroe);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_mount_options(
    vrpsx_mount_t *mount,
    vrpsx_mount_option_t **options)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_mount_t, , obj, interface, mount);

    obj->current_option.begin = obj->mntent_native.mnt_opts;
    obj->current_option.size = 0;

    *options = &obj->option_interface;
    return direct_mount_mount_option_next(options, options, NULL);
}

static vrpsx_errno_t trace_mount_options(
    vrpsx_mount_t *mount,
    vrpsx_mount_option_t **options)
{
    static char const me[] = "mount_options";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, mount);
    trace_address(&top, options);
    trace_end(&top);

    vrpsx_errno_t ern = direct_mount_options(
        mount,
        options);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && options) {
        trace_address(&top, *options);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_mount_end_t const mount_end;

#define MOUNT_VTBL_INITIALIZATION(PREFIX)                               \
                                                                        \
static vrpsx_mount_vtbl_t const VRPSX_CAT2(PREFIX, _mount_vtbl) = {     \
    VRPSX_GET_SYSTEM_METHOD(PREFIX, vrpsx_mount_t),                     \
    &system_associate_end,                                              \
    &VRPSX_CAT2(PREFIX, _mount_next),                                   \
    &VRPSX_CAT2(PREFIX, _mount_device),                                 \
    &VRPSX_CAT2(PREFIX, _mount_root),                                   \
    &VRPSX_CAT2(PREFIX, _mount_type),                                   \
    &VRPSX_CAT2(PREFIX, _mount_options),                                \
    &mount_end                                                          \
}

MOUNT_VTBL_INITIALIZATION(direct);
MOUNT_VTBL_INITIALIZATION(trace);

static vrpsx_mount_option_end_t const mount_option_end;

#define MOUNT_OPTION_VTBL_INITIALIZATION(PREFIX)                \
                                                                \
static vrpsx_mount_option_vtbl_t const                          \
VRPSX_CAT2(PREFIX, _mount_mount_option_vtbl) = {                \
    VRPSX_GET_SYSTEM_METHOD(PREFIX, vrpsx_mount_option_t),      \
    &system_associate_end,                                      \
    &VRPSX_CAT2(PREFIX, _mount_mount_option_next),              \
    &VRPSX_CAT2(PREFIX, _mount_mount_option_text),              \
    &mount_option_end                                           \
}

MOUNT_OPTION_VTBL_INITIALIZATION(direct);
MOUNT_OPTION_VTBL_INITIALIZATION(trace);


typedef struct direct_mount_flag_struct {
    vrpsx_substring_t text;
    unsigned long or_mask;
    unsigned long and_mask;
} direct_mount_flag_t;

#define DIRECT_MOUNT_FLAG(T, OR, AND_NOT)               \
    { { T, sizeof T }, OR, (~(unsigned long)AND_NOT) }


#if defined(__linux__)

/* Linux mount flags, ALREADY SORTED by textual name.
 */
static direct_mount_flag_t const direct_mount_flags[] = {
#if defined(MS_SYNCHRONOUS)
    DIRECT_MOUNT_FLAG("async", 0, MS_SYNCHRONOUS),
#endif
#if defined(MS_NOATIME)
    DIRECT_MOUNT_FLAG("atime", 0, MS_NOATIME),
#endif
#if defined(MS_BIND)
    DIRECT_MOUNT_FLAG("bind", MS_BIND, 0),
#endif
#if defined(MS_NODEV)
    DIRECT_MOUNT_FLAG("dev", 0, MS_NODEV),
#endif
#if defined(MS_NODIRATIME)
    DIRECT_MOUNT_FLAG("diratime", 0, MS_NODIRATIME),
#endif
#if defined(MS_DIRSYNC)
    DIRECT_MOUNT_FLAG("dirsync", MS_DIRSYNC, 0),
#endif
#if defined(MS_NOEXEC)
    DIRECT_MOUNT_FLAG("exec", 0, MS_NOEXEC),
#endif
#if defined(MS_MANDLOCK)
    DIRECT_MOUNT_FLAG("mand", MS_MANDLOCK, 0),
#endif
#if defined(MS_MOVE)
    DIRECT_MOUNT_FLAG("move", MS_MOVE, 0),
#endif
#if defined(MS_NOATIME)
    DIRECT_MOUNT_FLAG("noatime", MS_NOATIME, 0),
#endif
#if defined(MS_NODEV)
    DIRECT_MOUNT_FLAG("nodev", MS_NODEV, 0),
#endif
#if defined(MS_NODIRATIME)
    DIRECT_MOUNT_FLAG("nodiratime", MS_NODIRATIME, 0),
#endif
#if defined(MS_NOEXEC)
    DIRECT_MOUNT_FLAG("noexec", MS_NOEXEC, 0),
#endif
#if defined(MS_MANDLOCK)
    DIRECT_MOUNT_FLAG("nomand", 0, MS_MANDLOCK),
#endif
#if defined(MS_RELATIME)
    DIRECT_MOUNT_FLAG("norelatime", 0, MS_RELATIME),
#endif
#if defined(MS_STRICTATIME)
    DIRECT_MOUNT_FLAG("nostrictatime", 0, MS_STRICTATIME),
#endif
#if defined(MS_NOSUID)
    DIRECT_MOUNT_FLAG("nosuid", MS_NOSUID, 0),
#endif
#if defined(MS_RELATIME)
    DIRECT_MOUNT_FLAG("relatime", MS_RELATIME, 0),
#endif
#if defined(MS_REMOUNT)
    DIRECT_MOUNT_FLAG("remount", MS_REMOUNT, 0),
#endif
#if defined(MS_RDONLY)
    DIRECT_MOUNT_FLAG("ro", MS_RDONLY, 0),
#endif
#if defined(MS_RDONLY)
    DIRECT_MOUNT_FLAG("rw", 0, MS_RDONLY),
#endif
#if defined(MS_SILENT)
    DIRECT_MOUNT_FLAG("silent", MS_SILENT, 0), /* not in Linux "mount" cmd */
#endif
#if defined(MS_STRICTATIME)
    DIRECT_MOUNT_FLAG("strictatime", MS_STRICTATIME, 0),
#endif
#if defined(MS_NOSUID)
    DIRECT_MOUNT_FLAG("suid", 0, MS_NOSUID),
#endif
#if defined(MS_SYNCHRONOUS)
    DIRECT_MOUNT_FLAG("sync", MS_SYNCHRONOUS, 0),
#endif
};

/* Linux umount flags, ALREADY SORTED by textual name.
 */
static direct_mount_flag_t const direct_umount_flags[] = {
#if defined(MNT_DETACH)
    DIRECT_MOUNT_FLAG("detach", MNT_DETACH, 0),
#endif
#if defined(MNT_EXPIRE)
    DIRECT_MOUNT_FLAG("expire", MNT_EXPIRE, 0),
#endif
#if defined(MNT_FORCE)
    DIRECT_MOUNT_FLAG("force", MNT_FORCE, 0),
#endif
#if defined(UMOUNT_NOFOLLOW)
    DIRECT_MOUNT_FLAG("nofollow", UMOUNT_NOFOLLOW, 0),
#endif
};

#endif /* defined(__linux__) */


#if defined(__CYGWIN__)

#if CYGWIN_VERSION_DLL_MAJOR < 1005

#error Cygwin versions before 1.5 are unsupported

#elif CYGWIN_VERSION_DLL_MAJOR == 1005

#if CYGWIN_VERSION_DLL_MINOR < 18
#error Cygwin 1.5.X supported only for X >= 18
#endif

/* Cygwin 1.5 mount flags, ALREADY SORTED by textual name.
 *
 * The Cygwin 1.5 "mount" command accepts "-o user" and "-o system",
 * but seems to ignore them.  If you want to control the user/system
 * distinction, you have to use "-u" and "-s".  But for purposes of
 * this interface, we present "user" and "system" as options.
 */
static direct_mount_flag_t const direct_mount_flags[] = {
    DIRECT_MOUNT_FLAG("binmode", MOUNT_BINARY, 0),
    DIRECT_MOUNT_FLAG("cygexec", MOUNT_CYGWIN_EXEC, 0),
    DIRECT_MOUNT_FLAG("exec", MOUNT_EXEC, 0),
    DIRECT_MOUNT_FLAG("managed", MOUNT_ENC, 0),
    DIRECT_MOUNT_FLAG("noexec", MOUNT_NOTEXEC, 0),
    DIRECT_MOUNT_FLAG("textmode", 0, MOUNT_BINARY),
};

/* Cygwin 1.5 umount flags, ALREADY SORTED by textual name.
 */
static direct_mount_flag_t const direct_umount_flags[] = {
    DIRECT_MOUNT_FLAG("cygdrive", MOUNT_CYGDRIVE, 0),
    DIRECT_MOUNT_FLAG("system", MOUNT_SYSTEM, 0),
    DIRECT_MOUNT_FLAG("user", 0, MOUNT_SYSTEM)
};

#elif CYGWIN_VERSION_DLL_MAJOR == 1006

#error Cygwin version 1.6 unsupported

#elif CYGWIN_VERSION_DLL_MAJOR >= 1007

#if CYGWIN_VERSION_DLL_MAJOR == 1007 && CYGWIN_VERSION_DLL_MINOR < 7
#error Cygwin 1.7.X supported only for X >= 7
#endif

/* Cygwin 3.x removed MOUNT_BINARY, apparently it became the default.
 * To make this code more compatible between 1.7, 2.x, and 3.x, we
 * redefine MOUNT_BINARY and hope for the best.
 */

#ifndef MOUNT_BINARY
#define MOUNT_BINARY 0x00002
#endif

/* Cygwin 1.7+ mount flags, ALREADY SORTED by textual name.
 */
static direct_mount_flag_t const direct_mount_flags[] = {
    DIRECT_MOUNT_FLAG("acl", 0, MOUNT_NOACL),
    DIRECT_MOUNT_FLAG("auto", 0, 0), /* cannot directly make auto mounts */
    DIRECT_MOUNT_FLAG("binary", MOUNT_BINARY, 0),
    DIRECT_MOUNT_FLAG("bind", MOUNT_BIND, 0),
    DIRECT_MOUNT_FLAG("cygexec", MOUNT_CYGWIN_EXEC, 0),
    DIRECT_MOUNT_FLAG("dos", MOUNT_DOS, 0),
    DIRECT_MOUNT_FLAG("exec", MOUNT_EXEC, 0),
    DIRECT_MOUNT_FLAG("ihash", MOUNT_IHASH, 0),
    DIRECT_MOUNT_FLAG("noacl", MOUNT_NOACL, 0),
    DIRECT_MOUNT_FLAG("nosuid", 0, 0), /* "mount" allows this but ignores it */
    DIRECT_MOUNT_FLAG("notexec", MOUNT_NOTEXEC, 0),
    DIRECT_MOUNT_FLAG("override", MOUNT_OVERRIDE, 0),
    DIRECT_MOUNT_FLAG("posix=0", MOUNT_NOPOSIX, 0),
    DIRECT_MOUNT_FLAG("posix=1", 0, MOUNT_NOPOSIX),
#if defined(MOUNT_SPARSE)
    DIRECT_MOUNT_FLAG("sparse", 0, MOUNT_SPARSE),
#endif
    DIRECT_MOUNT_FLAG("text", 0, MOUNT_BINARY)
};

/* Cygwin 1.7 umount flags, ALREADY SORTED by textual name.
 */
static direct_mount_flag_t const direct_umount_flags[] = {
};

#endif /* CYGWIN_VERSION_DLL_MAJOR */

#endif /* defined(__CYGWIN__) */


/* Given the addresses of two direct_mount_flag_t instances,
 * compares those flag definitions by text encoding.
 */
static int direct_mount_flag_compare(
    void const *va,
    void const *vb)
{
    direct_mount_flag_t const *fa = (direct_mount_flag_t const *)va;
    direct_mount_flag_t const *fb = (direct_mount_flag_t const *)vb;

    size_t sa = fa->text.size;
    size_t sb = fb->text.size;

    int smin = sa < sb ? sa : sb;
    int cmp = memcmp(fa->text.begin, fb->text.begin, smin);

    return cmp
            ? cmp
            : sa < sb
                ? -1
                : sa > sb
                    ? +1
                    : 0;
}

/* Converts a list of mount options into a bitmap of flags
 * and a comma-separated list, which is stored into *csv.
 * On failure, *csv is set to NULL.
 *
 * You pass in the address of a bitmap variable whose initial value
 * specifies the default flags.  Options that modify flags are used
 * to modify that bitmap and all other options are placed into *csv.
 *
 * On success, be sure to eventually call free(*csv).
 */
static vrpsx_errno_t direct_mount_option_parse(
    direct_mount_flag_t const *mount_flags_begin,
    size_t mount_flags_count,
    unsigned long *flags,
    vrpsx_mount_option_t **options,  /* always consumed */
    char **csv)
{
    size_t used = 0, size = 0;
    char *buffer = NULL, *new_buffer;
    direct_mount_flag_t key;
    direct_mount_flag_t const *flag;
    vrpsx_mount_option_t *option;
    vrpsx_errno_t ern;

    *csv = NULL;

    option = *options;
    *options = NULL;

    while (option) {
        ern = option->vtbl->text(option, &key.text);
        if (ern) {
            goto consume_options;
        }

        flag = (direct_mount_flag_t const *) bsearch(
            &key,
            mount_flags_begin,
            mount_flags_count,
            sizeof direct_mount_flags[0],
            &direct_mount_flag_compare);

        if (flag) {
            *flags = (*flags | flag->or_mask) & flag->and_mask;
            continue;
        }

        /* Not a recognized flag, but may be a filesystem-specific option.
         * Pass it on to the filesystem handler as text.
         */
        if (size - used < key.text.size) {
            if ((size_t)-1 - used < key.text.size) {
                ern = ENOMEM;
                goto consume_options;
            }
            size *= 2;
            if (size < used + key.text.size) {
                size = used + key.text.size;
            }
            new_buffer = realloc(buffer, size);
            if (! new_buffer) {
                ern = errno;
                goto consume_options;
            }
            buffer = new_buffer;
        }

        memcpy(buffer + used, key.text.begin, key.text.size);
        used += key.text.size;

        VRPSX_SINGLE_OWNER_LIST_NEXT_CHECK_SROE(
            vrpsx_mount_option_t, &option, &option, &ern, NULL);
        if (ern) {
            goto free_buffer;
        }
    }

    new_buffer = realloc(buffer, used + 1);
    if (! new_buffer) {
        ern = errno;
        goto free_buffer;
    }
    buffer = new_buffer;
    buffer[used] = '\0';

    *csv = buffer;
    return 0;

consume_options:
    VRPSX_SINGLE_OWNER_LIST_CLEAR_SROE(vrpsx_mount_option_t, &option);
free_buffer:
    free(buffer);
    return ern;
}

#endif /* defined(__linux__) || defined(__CYGWIN__) */


typedef struct direct_embedding_struct {
    vrpsx_embedding_t interface;
    vrpsx_thread_attr_vtbl_t const *thread_attr_vtbl;
    vrpsx_mount_option_vtbl_t const *mount_mount_option_vtbl;
    vrpsx_mount_vtbl_t const *mount_vtbl;
} direct_embedding_t;

static vrpsx_embedding_vtbl_t const direct_embedding_vtbl;

static vrpsx_embedding_vtbl_t const trace_embedding_vtbl;

static vrpsx_system_vtbl_t const direct_system_vtbl;

static vrpsx_system_vtbl_t const trace_system_vtbl;

static direct_embedding_t direct_embedding_singleton = {
    {
        &direct_embedding_vtbl,
        {
            &direct_system_vtbl,
            {
                &allocator_vtbl
            }
        }
    },
    &direct_thread_attr_vtbl,
    &direct_mount_mount_option_vtbl,
    &direct_mount_vtbl
};

static direct_embedding_t trace_embedding_singleton = {
    {
        &trace_embedding_vtbl,
        {
            &trace_system_vtbl,
            {
                &allocator_vtbl
            }
        }
    },
    &trace_thread_attr_vtbl,
    &trace_mount_mount_option_vtbl,
    &trace_mount_vtbl
};

static vrpsx_system_t *direct_get_system(
    void const *interface)  /* unused */
{
    return &direct_embedding_singleton.interface.system;
}

static vrpsx_system_t *trace_get_system(
    void const *interface)  /* unused */
{
    return &trace_embedding_singleton.interface.system;
}

extern vrpsx_errno_t direct_embedding_acquire(
    vrpsx_embedding_t **embedding)
{
    pthread_mutex_lock(&trace_mutex);

    *embedding = trace_file
            ? &trace_embedding_singleton.interface
            : &direct_embedding_singleton.interface;

    pthread_mutex_unlock(&trace_mutex);
    return 0;
}

static void direct_replace_interface(
    vrpsx_interface_identifier_t interface_identifier,
    void **out_alternative,
    vrpsx_error_t **out_error)
{
    size_t offset;

    *out_alternative = NULL;
    *out_error = NULL;

    vrpsx_interface_identifier_t const *provided
            = direct_embedding_supports(interface_identifier, &offset);

    if (! provided) {
        direct_error_create(ENOSYS, out_error);
        return;
    }

    switch (*provided) {
        case vrpsx_embedding_IDENTIFIER: {
            vrpsx_embedding_t *embedding=&direct_embedding_singleton.interface;
            *out_alternative = (char *)embedding + offset;
            return;
        }
    }

    assert(0 /* provided interface needs case in above switch */);
    direct_error_create(ENOSYS, out_error);
}

static void direct_embedding_replace_interface(
    vrpsx_embedding_t **identified,
    vrpsx_interface_identifier_t interface_identifier,
    void **out_alternative,
    vrpsx_error_t **out_error)
{
    direct_replace_interface(
        interface_identifier,
        out_alternative,
        out_error);

    if (*out_alternative) {
        *identified = NULL;
    }
}

static void direct_system_replace_interface(
    vrpsx_system_t **identified,
    vrpsx_interface_identifier_t interface_identifier,
    void **out_alternative,
    vrpsx_error_t **out_error)
{
    direct_replace_interface(
        interface_identifier,
        out_alternative,
        out_error);

    if (*out_alternative) {
        *identified = NULL;
    }
}

static void direct_embedding_release(
    vrpsx_embedding_t **counted)
{
    *counted = NULL;
}

static void direct_embedding_acquire_op(
    vrpsx_embedding_t *counted)
{
    /* NOTE: If this implementation becomes substantive,
     * then please update VRPSX_SYSTEM_SINGLETON_ACQUIRE!
     */
}

static void direct_system_release(
    vrpsx_system_t **counted)
{
    *counted = NULL;
}

static void direct_system_acquire(
    vrpsx_system_t *counted)
{
    /* NOTE: If this implementation becomes substantive,
     * then please update VRPSX_SYSTEM_SINGLETON_ACQUIRE!
     */
}

/* The error number for those error names that are defined by POSIX
 * and mandated by the virtual POSIX-like interface but are undefined
 * on the platform on which this implementation is built.
 *
 * We chose a high code at random in order to make it unlikely that
 * we collide with an actual error number.  NOT the same as ENOSYS.
 */
#define EUNSUPPORTED 0x52d51603

#ifndef ECANCELED
#define ECANCELED EUNSUPPORTED
#endif

static void direct_system_error_acquire(
    vrpsx_system_t *system,
    vrpsx_errno_t ern,
    vrpsx_system_error_t **out_error)
{
    direct_system_error_create(ern, out_error);
}

/* Must remain compatible with allocator_allocate_memory and
 * allocator_free_memory--it should not matter which is used.
 */
DIRECT_ALLOCATION(direct_system, vrpsx_system_t);

static void direct_system_exit(
    vrpsx_system_t *system,
    int code)
{
    exit(code);
}

static vrpsx_errno_t direct_system_uname(
    vrpsx_system_t *system,
    vrpsx_utsname_t **name)
{
    direct_utsname_t *obj;
    vrpsx_errno_t ern;

    *name = NULL;

    obj = MALLOC(direct_utsname_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_utsname_vtbl;
    if (uname(&obj->native) < 0) {
        ern = errno;
        free(obj);
        return ern;
    }

    *name = &obj->interface;
    return 0;
}

/* The sysconf name code for those names that are defined by POSIX
 * and mandated by the virtual POSIX-like interface but are undefined
 * on the platform on which this implementation is built.
 *
 * We chose a high (but positive) code at random in order to make it
 * unlikely that we collide with an actual sysconf name.
 */
#define _SC_UNSUPPORTED 0x6ed4e30e

#ifndef _SC_2_C_BIND
#define _SC_2_C_BIND _SC_UNSUPPORTED
#endif

#ifndef _SC_2_C_DEV
#define _SC_2_C_DEV _SC_UNSUPPORTED
#endif

#ifndef _SC_2_CHAR_TERM
#define _SC_2_CHAR_TERM _SC_UNSUPPORTED
#endif

#ifndef _SC_2_FORT_DEV
#define _SC_2_FORT_DEV _SC_UNSUPPORTED
#endif

#ifndef _SC_2_FORT_RUN
#define _SC_2_FORT_RUN _SC_UNSUPPORTED
#endif

#ifndef _SC_2_LOCALEDEF
#define _SC_2_LOCALEDEF _SC_UNSUPPORTED
#endif

#ifndef _SC_2_PBS
#define _SC_2_PBS _SC_UNSUPPORTED
#endif

#ifndef _SC_2_PBS_ACCOUNTING
#define _SC_2_PBS_ACCOUNTING _SC_UNSUPPORTED
#endif

#ifndef _SC_2_PBS_CHECKPOINT
#define _SC_2_PBS_CHECKPOINT _SC_UNSUPPORTED
#endif

#ifndef _SC_2_PBS_LOCATE
#define _SC_2_PBS_LOCATE _SC_UNSUPPORTED
#endif

#ifndef _SC_2_PBS_MESSAGE
#define _SC_2_PBS_MESSAGE _SC_UNSUPPORTED
#endif

#ifndef _SC_2_PBS_TRACK
#define _SC_2_PBS_TRACK _SC_UNSUPPORTED
#endif

#ifndef _SC_2_SW_DEV
#define _SC_2_SW_DEV _SC_UNSUPPORTED
#endif

#ifndef _SC_2_UPE
#define _SC_2_UPE _SC_UNSUPPORTED
#endif

#ifndef _SC_2_VERSION
#define _SC_2_VERSION _SC_UNSUPPORTED
#endif

#ifndef _SC_ADVISORY_INFO
#define _SC_ADVISORY_INFO _SC_UNSUPPORTED
#endif

#ifndef _SC_ATEXIT_MAX
#define _SC_ATEXIT_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_BARRIERS
#define _SC_BARRIERS _SC_UNSUPPORTED
#endif

#ifndef _SC_BC_BASE_MAX
#define _SC_BC_BASE_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_BC_DIM_MAX
#define _SC_BC_DIM_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_BC_SCALE_MAX
#define _SC_BC_SCALE_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_BC_STRING_MAX
#define _SC_BC_STRING_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_CLOCK_SELECTION
#define _SC_CLOCK_SELECTION _SC_UNSUPPORTED
#endif

#ifndef _SC_COLL_WEIGHTS_MAX
#define _SC_COLL_WEIGHTS_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_CPUTIME
#define _SC_CPUTIME _SC_UNSUPPORTED
#endif

#ifndef _SC_EXPR_NEST_MAX
#define _SC_EXPR_NEST_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_HOST_NAME_MAX
#define _SC_HOST_NAME_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_IOV_MAX
#define _SC_IOV_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_IPV6
#define _SC_IPV6 _SC_UNSUPPORTED
#endif

#ifndef _SC_LINE_MAX
#define _SC_LINE_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_MONOTONIC_CLOCK
#define _SC_MONOTONIC_CLOCK _SC_UNSUPPORTED
#endif

#ifndef _SC_PAGE_SIZE
#define _SC_PAGE_SIZE _SC_UNSUPPORTED
#endif

#ifndef _SC_PRIORITY_SCHEDULING
#define _SC_PRIORITY_SCHEDULING _SC_UNSUPPORTED
#endif

#ifndef _SC_RAW_SOCKETS
#define _SC_RAW_SOCKETS _SC_UNSUPPORTED
#endif

#ifndef _SC_READER_WRITER_LOCKS
#define _SC_READER_WRITER_LOCKS _SC_UNSUPPORTED
#endif

#ifndef _SC_REGEXP
#define _SC_REGEXP _SC_UNSUPPORTED
#endif

#ifndef _SC_RE_DUP_MAX
#define _SC_RE_DUP_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_SHELL
#define _SC_SHELL _SC_UNSUPPORTED
#endif

#ifndef _SC_SPAWN
#define _SC_SPAWN _SC_UNSUPPORTED
#endif

#ifndef _SC_SPIN_LOCKS
#define _SC_SPIN_LOCKS _SC_UNSUPPORTED
#endif

#ifndef _SC_SPORADIC_SERVER
#define _SC_SPORADIC_SERVER _SC_UNSUPPORTED
#endif

#ifndef _SC_SS_REPL_MAX
#define _SC_SS_REPL_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_STREAM_MAX
#define _SC_STREAM_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_SYMLOOP_MAX
#define _SC_SYMLOOP_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_THREAD_CPUTIME
#define _SC_THREAD_CPUTIME _SC_UNSUPPORTED
#endif

#ifndef _SC_THREAD_ROBUST_PRIO_INHERIT
#define _SC_THREAD_ROBUST_PRIO_INHERIT _SC_UNSUPPORTED
#endif

#ifndef _SC_THREAD_ROBUST_PRIO_PROTECT
#define _SC_THREAD_ROBUST_PRIO_PROTECT _SC_UNSUPPORTED
#endif

#ifndef _SC_THREAD_SPORADIC_SERVER
#define _SC_THREAD_SPORADIC_SERVER _SC_UNSUPPORTED
#endif

#ifndef _SC_TIMEOUTS
#define _SC_TIMEOUTS _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE
#define _SC_TRACE _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE_EVENT_FILTER
#define _SC_TRACE_EVENT_FILTER _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE_EVENT_NAME_MAX
#define _SC_TRACE_EVENT_NAME_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE_INHERIT
#define _SC_TRACE_INHERIT _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE_LOG
#define _SC_TRACE_LOG _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE_NAME_MAX
#define _SC_TRACE_NAME_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE_SYS_MAX
#define _SC_TRACE_SYS_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_TRACE_USER_EVENT_MAX
#define _SC_TRACE_USER_EVENT_MAX _SC_UNSUPPORTED
#endif

#ifndef _SC_TYPED_MEMORY_OBJECTS
#define _SC_TYPED_MEMORY_OBJECTS _SC_UNSUPPORTED
#endif

#ifndef _SC_V6_ILP32_OFF32
#define _SC_V6_ILP32_OFF32 _SC_UNSUPPORTED
#endif

#ifndef _SC_V6_ILP32_OFFBIG
#define _SC_V6_ILP32_OFFBIG _SC_UNSUPPORTED
#endif

#ifndef _SC_V6_LP64_OFF64
#define _SC_V6_LP64_OFF64 _SC_UNSUPPORTED
#endif

#ifndef _SC_V6_LPBIG_OFFBIG
#define _SC_V6_LPBIG_OFFBIG _SC_UNSUPPORTED
#endif

#ifndef _SC_V7_ILP32_OFF32
#define _SC_V7_ILP32_OFF32 _SC_UNSUPPORTED
#endif

#ifndef _SC_V7_ILP32_OFFBIG
#define _SC_V7_ILP32_OFFBIG _SC_UNSUPPORTED
#endif

#ifndef _SC_V7_LP64_OFF64
#define _SC_V7_LP64_OFF64 _SC_UNSUPPORTED
#endif

#ifndef _SC_V7_LPBIG_OFFBIG
#define _SC_V7_LPBIG_OFFBIG _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_CRYPT
#define _SC_XOPEN_CRYPT _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_ENH_I18N
#define _SC_XOPEN_ENH_I18N _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_REALTIME
#define _SC_XOPEN_REALTIME _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_REALTIME_THREADS
#define _SC_XOPEN_REALTIME_THREADS _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_SHM
#define _SC_XOPEN_SHM _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_STREAMS
#define _SC_XOPEN_STREAMS _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_UNIX
#define _SC_XOPEN_UNIX _SC_UNSUPPORTED
#endif

#ifndef _SC_XOPEN_VERSION
#define _SC_XOPEN_VERSION _SC_UNSUPPORTED
#endif

static vrpsx_errno_t direct_system_sysconf(
    vrpsx_system_t *system,
    int name,
    long *out_value)
{
    if (name == _SC_UNSUPPORTED) {
        *out_value = -1;
        return ENOSYS;
    }

    errno = 0;
    *out_value = sysconf(name);
    return errno;
}

/* The [f]pathconf name code for those names that are defined by POSIX
 * and mandated by the virtual POSIX-like interface but are undefined
 * on the platform on which this implementation is built.
 *
 * We chose a high (but positive) code at random in order to make it
 * unlikely that we collide with an actual [f]pathconf/fpathconf name.
 */
#define _PC_UNSUPPORTED 0x633a1518

#ifndef _PC_2_SYMLINKS
#define _PC_2_SYMLINKS _PC_UNSUPPORTED
#endif

#ifndef _PC_ALLOC_SIZE_MIN
#define _PC_ALLOC_SIZE_MIN _PC_UNSUPPORTED
#endif

#ifndef _PC_FILESIZEBITS
#define _PC_FILESIZEBITS _PC_UNSUPPORTED
#endif

#ifndef _PC_REC_INCR_XFER_SIZE
#define _PC_REC_INCR_XFER_SIZE _PC_UNSUPPORTED
#endif

#ifndef _PC_REC_MAX_XFER_SIZE
#define _PC_REC_MAX_XFER_SIZE _PC_UNSUPPORTED
#endif

#ifndef _PC_REC_MIN_XFER_SIZE
#define _PC_REC_MIN_XFER_SIZE _PC_UNSUPPORTED
#endif

#ifndef _PC_REC_XFER_ALIGN
#define _PC_REC_XFER_ALIGN _PC_UNSUPPORTED
#endif

#ifndef _PC_SYMLINK_MAX
#define _PC_SYMLINK_MAX _PC_UNSUPPORTED
#endif

#ifndef _PC_TIMESTAMP_RESOLUTION
#define _PC_TIMESTAMP_RESOLUTION _PC_UNSUPPORTED
#endif

static vrpsx_errno_t direct_system_pathconf(
    vrpsx_system_t *system,
    char const *path,
    int name,
    long *out_value)
{
    if (name == _PC_UNSUPPORTED) {
        *out_value = -1;
        return ENOSYS;
    }

    errno = 0;
    *out_value = pathconf(path, name);
    return errno;
}

static vrpsx_errno_t direct_system_fpathconf(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    int name,
    long *out_value)
{
    if (name == _PC_UNSUPPORTED) {
        *out_value = -1;
        return ENOSYS;
    }

    errno = 0;
    *out_value = fpathconf(fd, name);
    return errno;
}

static vrpsx_errno_t direct_system_getcwd(
    vrpsx_system_t *system,
    vrpsx_allocator_t *allocator,
    char **out_cwd)
{
    char buffer[PATH_MAX];
    size_t length;

    *out_cwd = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    if (! getcwd(buffer, sizeof buffer)) {
        /* TO DO: If errno == ERANGE then use pathconf(".") to make a new guess
         * at how big to make the buffer, with special handling for indefinite
         * or very large limits, as suggested by POSIX.
         */
        return errno;
    }

    length = strlen(buffer);
    *out_cwd = (char *) VRPSX_CALL1(allocator, allocate_memory, length + 1);
    if (! *out_cwd) {
        return ENOMEM;
    }

    memcpy(*out_cwd, buffer, length + 1);
    return 0;
}

static vrpsx_int64_t direct_system_get_timezone(
    vrpsx_system_t *system)
{
#if defined(__CYGWIN__)
    return _timezone;
#else
    return timezone;
#endif
}

static char const *const *direct_system_get_tzname(
    vrpsx_system_t *system)
{
    char **tzn = tzname;
    return (char const *const *)tzn;
}

static vrpsx_errno_t direct_system_gmtime(
    vrpsx_system_t *system,
    vrpsx_time_t in_time,
    vrpsx_tm_t *out_time)
{
    time_t in_native = in_time;
    struct tm out_native;

    ZERO(out_time);

    if (! gmtime_r(&in_native, &out_native)) {
        return errno;
    }

    out_time->tm_sec   = out_native.tm_sec;
    out_time->tm_min   = out_native.tm_min;
    out_time->tm_hour  = out_native.tm_hour;
    out_time->tm_mday  = out_native.tm_mday;
    out_time->tm_mon   = out_native.tm_mon;
    out_time->tm_year  = out_native.tm_year;
    out_time->tm_wday  = out_native.tm_wday;
    out_time->tm_yday  = out_native.tm_yday;
    out_time->tm_isdst = out_native.tm_isdst;
    return 0;
}

static vrpsx_errno_t direct_system_localtime(
    vrpsx_system_t *system,
    vrpsx_time_t in_time,
    vrpsx_tm_t *out_time)
{
    time_t in_native = in_time;
    struct tm out_native;
    vrpsx_errno_t ern;

    ZERO(out_time);

    ern = localtime_r(&in_native, &out_native) ? 0 : errno;
    if (ern) {
        return ern;
    }

    out_time->tm_sec   = out_native.tm_sec;
    out_time->tm_min   = out_native.tm_min;
    out_time->tm_hour  = out_native.tm_hour;
    out_time->tm_mday  = out_native.tm_mday;
    out_time->tm_mon   = out_native.tm_mon;
    out_time->tm_year  = out_native.tm_year;
    out_time->tm_wday  = out_native.tm_wday;
    out_time->tm_yday  = out_native.tm_yday;
    out_time->tm_isdst = out_native.tm_isdst;
    return 0;
}

static vrpsx_errno_t direct_system_mktime(
    vrpsx_system_t *system,
    vrpsx_tm_t *io_tm,
    vrpsx_time_t *out_time)
{
    struct tm io_tm_native;
    time_t out_time_native;
    vrpsx_errno_t ern;

    io_tm_native.tm_sec   = io_tm->tm_sec;
    io_tm_native.tm_min   = io_tm->tm_min;
    io_tm_native.tm_hour  = io_tm->tm_hour;
    io_tm_native.tm_mday  = io_tm->tm_mday;
    io_tm_native.tm_mon   = io_tm->tm_mon;
    io_tm_native.tm_year  = io_tm->tm_year;
    io_tm_native.tm_isdst = io_tm->tm_isdst;

    *out_time = -1;

    out_time_native = mktime(&io_tm_native);
    ern = (out_time_native == (time_t)-1) ? errno : 0;
    if (ern) {
        return ern;
    }

    io_tm->tm_sec   = io_tm_native.tm_sec;
    io_tm->tm_min   = io_tm_native.tm_min;
    io_tm->tm_hour  = io_tm_native.tm_hour;
    io_tm->tm_mday  = io_tm_native.tm_mday;
    io_tm->tm_mon   = io_tm_native.tm_mon;
    io_tm->tm_year  = io_tm_native.tm_year;
    io_tm->tm_wday  = io_tm_native.tm_wday;
    io_tm->tm_yday  = io_tm_native.tm_yday;
    io_tm->tm_isdst = io_tm_native.tm_isdst;

    *out_time = out_time_native;

    return 0;
}

static vrpsx_errno_t direct_system_time(
    vrpsx_system_t *system,
    vrpsx_time_t *out_time)
{
    time_t out_native;
    if (time(&out_native) == (time_t)-1) {
        *out_time = -1;
        return errno;
    }
    *out_time = out_native;
    return 0;
}

static vrpsx_errno_t direct_system_clock_gettime(
    vrpsx_system_t *system,
    vrpsx_clockid_t clock,
    vrpsx_timespec_t *out_time)
{
    struct timespec out_native;
    if (clock_gettime(clock, &out_native) == (time_t)-1) {
        out_time->tv_sec = 0;   /* undocumented but deterministic value */
        out_time->tv_nsec = 0;  /* undocumented but deterministic value */
        return errno;
    }
    DIRECT_TIMESPEC_FROM_NATIVE(out_time, &out_native);
    return 0;
}

static vrpsx_errno_t direct_system_nanosleep(
    vrpsx_system_t *system,
    vrpsx_timespec_t const *duration,
    vrpsx_timespec_t *remaining)
{
    struct timespec native_duration;
    struct timespec native_remaining;
    vrpsx_errno_t ern;

    ZERO(&native_duration);
    DIRECT_TIMESPEC_TO_NATIVE(&native_duration, duration);

    if (remaining) {
        ZERO(&native_remaining);

        ern = 0;
        if (nanosleep(&native_duration, &native_remaining)) {
            ern = errno;
        }

        DIRECT_TIMESPEC_FROM_NATIVE(remaining, &native_remaining);
    } else {
        ern = 0;
        if (nanosleep(&native_duration, NULL)) {
            ern = errno;
        }
    }

    return ern;
}

static vrpsx_errno_t direct_system_getlogin(
    vrpsx_system_t *system,
    vrpsx_allocator_t *allocator,
    char **out_name)
{
    char buffer[LOGIN_NAME_MAX];
    size_t length;
    vrpsx_errno_t ern;

    *out_name = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    ern = getlogin_r(buffer, sizeof buffer);
    if (ern) {
        return ern;
    }

    length = strlen(buffer);
    *out_name = (char *) VRPSX_CALL1(allocator, allocate_memory, length + 1);
    if (! *out_name) {
        return ENOMEM;
    }

    memcpy(*out_name, buffer, length + 1);
    return 0;
}

static vrpsx_errno_t direct_system_getpwnam(
    vrpsx_system_t *system,
    char const *name,
    vrpsx_passwd_t **result)
{
    direct_passwd_t *obj;
    struct passwd native, *nativep;
    char *buffer, *next;
    size_t buffer_size, doubled;
    size_t dir_size, name_size, shell_size;
    vrpsx_errno_t ern;

    *result = NULL;

    buffer_size = sysconf(_SC_GETPW_R_SIZE_MAX);
    if (buffer_size == (size_t)-1) {
        buffer_size = 256;
    }

    for (;;) {
        buffer = (char *)malloc(buffer_size);
        if (! buffer) {
            return errno;
        }

        ZERO(&native);
        ern = getpwnam_r(
            name,
            &native,
            buffer,
            buffer_size,
            &nativep);
        if (ern != ERANGE) {
            break;
        }

        free(buffer);
        doubled = buffer_size << 1;
        if (doubled <= buffer_size) {
            return ENOMEM;
        }
        buffer_size = doubled;
    }

    if (nativep && !ern) {
        dir_size = strlen(native.pw_dir) + 1;
        name_size = strlen(native.pw_name) + 1;
        shell_size = strlen(native.pw_shell) + 1;

        obj = (direct_passwd_t *)malloc(
            sizeof(direct_passwd_t) + dir_size + name_size + shell_size);
        if (! obj) {
            ern = errno;
            free(buffer);
            return ern;
        }

        obj->interface.vtbl = &direct_passwd_vtbl;
        obj->native = native;
        next = (char*)(obj + 1);

        obj->native.pw_dir = next;
        MEMCPY_END(next, native.pw_dir, dir_size);

        obj->native.pw_name = next;
        MEMCPY_END(next, native.pw_name, name_size);

        obj->native.pw_shell = next;
        memcpy(next, native.pw_shell, shell_size);

        *result = &obj->interface;
    }

    free(buffer);
    return ern;
}

static vrpsx_pid_t direct_system_getpid(
    vrpsx_system_t *system)
{
    return getpid();
}

static vrpsx_pid_t direct_system_getpgrp(
    vrpsx_system_t *system)
{
    return getpgrp();
}

static vrpsx_errno_t direct_system_getpgid(
    vrpsx_system_t *system,
    vrpsx_pid_t pid,
    vrpsx_pid_t *out_pgid)
{
    pid_t native;

    *out_pgid = -1;

    native = getpgid(pid);
    if (native == (pid_t)-1) {
        return errno;
    }

    *out_pgid = native;
    return 0;
}

static vrpsx_errno_t direct_system_kill(
    vrpsx_system_t *system,
    vrpsx_pid_t pid,
    vrpsx_signo_t signum)
{
    if (kill(pid, signum)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_raise(
    vrpsx_system_t *system,
    vrpsx_signo_t signum)
{
    if (raise(signum)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t trace_system_raise(
    vrpsx_system_t *system,
    vrpsx_signo_t signum)
{
    static char const me[] = "raise";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_signo(&top, signum);
    trace_end(&top);

    vrpsx_errno_t ern = direct_system_raise(
        system,
        signum);

    trace_leave_begin_ern(&top, me, ern);
    trace_end(&top);

    return ern;
}

static char const *direct_system_strsignal(
    vrpsx_system_t *system,
    vrpsx_signo_t signum)
{
    return safe_strsignal(signum);
}

static vrpsx_errno_t direct_system_sigset_init(
    vrpsx_system_t *system,
    void *storage,
    vrpsx_sigset_t **set)
{
    direct_bundled_sigset_t *obj;

    *set = NULL;

    if (! storage) {
        return EFAULT;
    }

    if ((uintptr_t)(char *)storage % DIRECT_BUNDLED_SIGSET_ALIGNOF) {
        return EFAULT;
    }

    obj = (direct_bundled_sigset_t *)storage;
    if (sigemptyset(&obj->native)) {
        return errno;
    }

    direct_pointer_to_sigset_init(&obj->native, &obj->pointer);
    *set = &obj->pointer.interface;
    return 0;
}

static vrpsx_errno_t direct_system_sigaction_init(
    vrpsx_system_t *system,
    void *storage,
    vrpsx_sigaction_t **action)
{
    direct_sigaction_t *obj;
    vrpsx_errno_t ern;

    *action = NULL;

    if (! storage) {
        return EFAULT;
    }

    if ((uintptr_t)(char *)storage % DIRECT_SIGACTION_ALIGNOF) {
        return EFAULT;
    }

    obj = (direct_sigaction_t *)storage;

    ern = direct_sigaction_init(obj);
    if (ern) {
        return ern;
    }

    *action = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_system_sigaction(
    vrpsx_system_t *system,
    vrpsx_signo_t signum,
    vrpsx_sigaction_t const *action,
    vrpsx_sigaction_t *old_action)
{
    struct sigaction const *native = NULL;
    struct sigaction *old_native = NULL;

    if (action) {
        assert(action->vtbl == &direct_sigaction_vtbl);

        native = & VRPSX_UPCAST_NONNULL(
            direct_sigaction_t, interface, action) -> native;
    }

    if (old_action) {
        assert(old_action->vtbl == &direct_sigaction_vtbl);

        old_native = & VRPSX_UPCAST_NONNULL(
            direct_sigaction_t, interface, old_action) -> native;
    }

    return sigaction(signum, native, old_native);
}

static vrpsx_errno_t direct_system_sigwait(
    vrpsx_system_t *system,
    vrpsx_sigset_t const *set,
    vrpsx_signo_t *signum)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_pointer_to_sigset_t, const, set_obj, interface, set);

    assert(set->vtbl == &direct_pointer_to_sigset_vtbl);

    return sigwait(set_obj->native, signum);
}

static vrpsx_errno_t trace_system_sigwait(
    vrpsx_system_t *system,
    vrpsx_sigset_t const *set,
    vrpsx_signo_t *signum)
{
    static char const me[] = "sigwait";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_vrpsx_sigset(&top, set);
    trace_address(&top, signum);
    trace_end(&top);

    vrpsx_errno_t ern = direct_system_sigwait(
        system,
        set,
        signum);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && signum) {
        trace_signo(&top, *signum);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_system_thread_sigmask(
    vrpsx_system_t *system,
    int how,
    vrpsx_sigset_t const *change,
    vrpsx_sigset_t *old_set)
{
    sigset_t const *change_native = NULL;
    sigset_t *old_set_native = NULL;

    if (change) {
        assert(change->vtbl == &direct_pointer_to_sigset_vtbl);

        change_native = VRPSX_UPCAST_NONNULL(
            direct_pointer_to_sigset_t, interface, change) -> native;
    }

    if (old_set) {
        assert(old_set->vtbl == &direct_pointer_to_sigset_vtbl);

        old_set_native = VRPSX_UPCAST_NONNULL(
            direct_pointer_to_sigset_t, interface, old_set) -> native;
    }

    return pthread_sigmask(how, change_native, old_set_native);
}

static vrpsx_errno_t trace_system_thread_sigmask(
    vrpsx_system_t *system,
    int how,
    vrpsx_sigset_t const *change,
    vrpsx_sigset_t *old_set)
{
    static char const me[] = "thread_sigmask";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    switch (how) {
        case SIG_BLOCK:
            trace_identifier(&top, "SIG_BLOCK");
            break;
        case SIG_SETMASK:
            trace_identifier(&top, "SIG_SETMASK");
            break;
        case SIG_UNBLOCK:
            trace_identifier(&top, "SIG_UNBLOCK");
            break;
        default:
            trace_decimal_int(&top, how);
            break;
    }
    trace_vrpsx_sigset(&top, change);
    trace_address(&top, old_set);
    trace_end(&top);

    vrpsx_errno_t ern = direct_system_thread_sigmask(
        system,
        how,
        change,
        old_set);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && old_set) {
        trace_vrpsx_sigset(&top, old_set);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_system_thread_attr_create(
    vrpsx_system_t *system,
    vrpsx_thread_attr_t **attr)
{
    VRPSX_UPCAST_NONNULL_DECL(
        direct_embedding_t, , embedding, interface.system, system);
    direct_thread_attr_t *obj;
    vrpsx_errno_t ern;

    *attr = NULL;

    obj = MALLOC(direct_thread_attr_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = embedding->thread_attr_vtbl;
    ern = pthread_attr_init(&obj->native);
    if (ern) {
        free(obj);
        return ern;
    }

    *attr = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_system_thread_create(
    vrpsx_system_t *system,
    vrpsx_thread_t *thread,
    void *(*function)(void*),
    void *argument)
{
    pthread_t tid;
    vrpsx_errno_t ern = 0;

    ern = pthread_create(&tid, NULL, function, argument);

    /* In the case of an error, POSIX leaves the thread ID value undefined.
     * Let us help to detect accidental misuse of that undefined value by
     * choosing a distinctive constant for that value ("0DEF" = "unDEFined").
     */
    *thread = ern ? 0x0DEF0DEF0DEF0DEFLL : VRPSX_THREAD_ID(tid);

    return ern;
}

static vrpsx_errno_t trace_system_thread_create(
    vrpsx_system_t *system,
    vrpsx_thread_t *thread,
    void *(*function)(void*),
    void *argument)
{
    static char const me[] = "thread_create";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, thread);
    trace_address(&top, function);
    trace_address(&top, argument);
    trace_end(&top);

    /* Rather than directly invoking the caller-supplied function,
     * we invoke our own wrapper that will trace the beginning and
     * the end of execution of that thread.
     */
    vrpsx_errno_t ern;
    trace_thread_callback_t *delegate = MALLOC(trace_thread_callback_t);
    if (! delegate) {
        ern = errno;
    } else {
        delegate->function = function;
        delegate->argument = argument;

        ern = direct_system_thread_create(
            system,
            thread,
            trace_thread_callback,
            delegate);

        if (ern) {
            /* Because we failed to create the thread, we failed to transfer
             * responsibility to that thread for freeing the delegation info.
             */
            free(delegate);
        }
    }

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && thread) {
        trace_thread(&top, *thread);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_system_thread_detach(
    vrpsx_system_t *system,
    vrpsx_thread_t thread)
{
    return pthread_detach(NATIVE_THREAD_ID(thread));
}

static vrpsx_errno_t direct_system_thread_join(
    vrpsx_system_t *system,
    vrpsx_thread_t thread,
    void **value)
{
    return pthread_join(NATIVE_THREAD_ID(thread), value);
}

static vrpsx_errno_t trace_system_thread_join(
    vrpsx_system_t *system,
    vrpsx_thread_t thread,
    void **value)
{
    static char const me[] = "thread_join";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_thread(&top, thread);
    trace_address(&top, value);
    trace_end(&top);

    vrpsx_errno_t ern = direct_system_thread_join(
        system,
        thread,
        value);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && value) {
        trace_address(&top, *value);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_system_thread_kill(
    vrpsx_system_t *system,
    vrpsx_thread_t thread,
    vrpsx_signo_t signum)
{
    return pthread_kill(NATIVE_THREAD_ID(thread), signum);
}

static vrpsx_errno_t trace_system_thread_kill(
    vrpsx_system_t *system,
    vrpsx_thread_t thread,
    vrpsx_signo_t signum)
{
    static char const me[] = "thread_kill";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_thread(&top, thread);
    trace_signo(&top, signum);
    trace_end(&top);

    vrpsx_errno_t ern = direct_system_thread_kill(
        system,
        thread,
        signum);

    trace_leave_begin_ern(&top, me, ern);
    trace_end(&top);

    return ern;
}

static int direct_system_thread_equal(
    vrpsx_system_t *system,
    vrpsx_thread_t thread1,
    vrpsx_thread_t thread2)
{
    return pthread_equal(
        NATIVE_THREAD_ID(thread1),
        NATIVE_THREAD_ID(thread2));
}

static vrpsx_thread_t direct_system_thread_self(
    vrpsx_system_t *system)
{
    return VRPSX_THREAD_ID(pthread_self());
}

static vrpsx_errno_t direct_system_thread_key_create(
    vrpsx_system_t *system,
    vrpsx_thread_key_t **key,
    void (*destructor)(void*))
{
    direct_thread_key_t *obj;
    vrpsx_errno_t ern;

    *key = NULL;

    obj = MALLOC(direct_thread_key_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_thread_key_vtbl;
    ern = pthread_key_create(&obj->native, destructor);
    if (ern) {
        ern = errno;
        free(obj);
        return ern;
    }

    *key = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_system_mutex_attr_create(
    vrpsx_system_t *system,
    vrpsx_mutex_attr_t **attr)
{
    direct_mutex_attr_t *obj;
    vrpsx_errno_t ern;

    *attr = NULL;

    obj = MALLOC(direct_mutex_attr_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_mutex_attr_vtbl;
    ern = pthread_mutexattr_init(&obj->native);
    if (ern) {
        ern = errno;
        free(obj);
        return ern;
    }

    *attr = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_system_mutex_create(
    vrpsx_system_t *system,
    vrpsx_mutex_t **mutex)
{
    return direct_mutex_attr_create(NULL, mutex);
}

static vrpsx_errno_t direct_system_rwlock_attr_create(
    vrpsx_system_t *system,
    vrpsx_rwlock_attr_t **attr)
{
    direct_rwlock_attr_t *obj;

    *attr = NULL;

    obj = MALLOC(direct_rwlock_attr_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_rwlock_attr_vtbl;

    *attr = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_system_rwlock_create(
    vrpsx_system_t *system,
    vrpsx_rwlock_t **rwlock)
{
    return direct_rwlock_attr_create(NULL, rwlock);
}

static vrpsx_errno_t direct_system_cond_attr_create(
    vrpsx_system_t *system,
    vrpsx_cond_attr_t **attr)
{
    direct_cond_attr_t *obj;

    *attr = NULL;

    obj = MALLOC(direct_cond_attr_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &direct_cond_attr_vtbl;

    *attr = &obj->interface;
    return 0;
}

static vrpsx_errno_t direct_system_cond_create(
    vrpsx_system_t *system,
    vrpsx_cond_t **cond)
{
    return direct_cond_attr_create(NULL, cond);
}

static vrpsx_errno_t direct_system_open(
    vrpsx_system_t *system,
    char const *path,
    int flags,
    int permissions,
    vrpsx_fd_t *out_fd)
{
    *out_fd = open(path, flags, permissions);
    if (*out_fd == -1) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_close(
    vrpsx_system_t *system,
    vrpsx_fd_t fd)
{
    if (close(fd)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_read(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    void *buffer,
    size_t in_size,
    size_t *out_size)
{
    ssize_t result = read(fd, buffer, in_size);
    if (result == -1) {
        *out_size = 0;
        return errno;
    }

    *out_size = (size_t)result;
    return 0;
}

static vrpsx_errno_t direct_system_write(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    void const *buffer,
    size_t in_size,
    size_t *out_size)
{
    ssize_t result = write(fd, buffer, in_size);
    if (result == -1) {
        if (out_size) {
            *out_size = 0;
        }
        return errno;
    }

    if (out_size) {
        *out_size = (size_t)result;
    }
    return 0;
}

static vrpsx_errno_t direct_system_pread(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    void *buffer,
    size_t in_size,
    size_t offset,
    size_t *out_size)
{
    ssize_t result = pread(fd, buffer, in_size, offset);
    if (result == -1) {
        *out_size = 0;
        return errno;
    }

    *out_size = (size_t)result;
    return 0;
}

static vrpsx_errno_t direct_system_pwrite(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    void const *buffer,
    size_t in_size,
    size_t offset,
    size_t *out_size)
{
    ssize_t result = pwrite(fd, buffer, in_size, offset);
    if (result == -1) {
        if (out_size) {
            *out_size = 0;
        }
        return errno;
    }

    if (out_size) {
        *out_size = (size_t)result;
    }
    return 0;
}

static vrpsx_errno_t direct_system_unlink(
    vrpsx_system_t *system,
    char const *path)
{
    if (unlink(path)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_link(
    vrpsx_system_t *system,
    char const *existing,
    char const *new_name)
{
    if (link(existing, new_name)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_symlink(
    vrpsx_system_t *system,
    char const *target,
    char const *symlink_name)
{
    if (symlink(target, symlink_name)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_readlink(
    vrpsx_system_t *system,
    char const *path,
    vrpsx_allocator_t *allocator,
    char **result,
    size_t *result_length)
{
    char buffer[PATH_MAX];
    ssize_t length;

    *result = NULL;
    *result_length = 0;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    length = readlink(path, buffer, sizeof buffer);
    if (length == -1) {
        return errno;
    }

    *result = (char *) VRPSX_CALL1(allocator, allocate_memory, length + 1);
    if (! *result) {
        return ENOMEM;
    }

    *result_length = length;
    memcpy(*result, buffer, length);
    buffer[length] = '\0';
    return 0;
}

static vrpsx_errno_t direct_system_chmod(
    vrpsx_system_t *system,
    char const *path,
    vrpsx_mode_t mode)
{
    if (chmod(path, mode)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_fchmod(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    vrpsx_mode_t mode)
{
    if (fchmod(fd, mode)) {
        return errno;
    }
    return 0;
}

static void direct_stat_from_native(
    vrpsx_stat_t *presented,
    struct stat const *native)
{
    DIRECT_TIMESPEC_FROM_NATIVE(&presented->st_atim, &native->st_atim);
    DIRECT_TIMESPEC_FROM_NATIVE(&presented->st_ctim, &native->st_ctim);
    DIRECT_TIMESPEC_FROM_NATIVE(&presented->st_mtim, &native->st_mtim);
    presented->st_size = native->st_size;
    presented->st_blocks = native->st_blocks;
    presented->st_blksize = native->st_blksize;
    presented->st_nlink = native->st_nlink;
    presented->st_rdev = native->st_rdev;
    presented->st_dev = native->st_dev;
    presented->st_ino = native->st_ino;
    presented->st_uid = native->st_uid;
    presented->st_gid = native->st_gid;
    presented->st_mode = native->st_mode;
}

static vrpsx_errno_t direct_system_stat(
    vrpsx_system_t *system,
    char const *path,
    vrpsx_stat_t *buffer)
{
    struct stat native;

    ZERO(buffer);

    if (stat(path, &native)) {
        return errno;
    }

    direct_stat_from_native(buffer, &native);
    return 0;
}

static vrpsx_errno_t direct_system_lstat(
    vrpsx_system_t *system,
    char const *path,
    vrpsx_stat_t *buffer)
{
    struct stat native;

    ZERO(buffer);

    if (lstat(path, &native)) {
        return errno;
    }

    direct_stat_from_native(buffer, &native);
    return 0;
}

static vrpsx_errno_t direct_system_fstat(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    vrpsx_stat_t *buffer)
{
    struct stat native;

    ZERO(buffer);

    if (fstat(fd, &native)) {
        return errno;
    }

    direct_stat_from_native(buffer, &native);
    return 0;
}

static vrpsx_errno_t direct_system_mkdir(
    vrpsx_system_t *system,
    char const *path,
    vrpsx_mode_t mode)
{
    if (mkdir(path, mode)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_rmdir(
    vrpsx_system_t *system,
    char const *path)
{
    if (rmdir(path)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_chdir(
    vrpsx_system_t *system,
    char const *path)
{
    if (chdir(path)) {
        return errno;
    }
    return 0;
}

static vrpsx_errno_t direct_system_opendir(
    vrpsx_system_t *system,
    char const *path,
    vrpsx_directory_t **dir)
{
    DIR *native;

    *dir = NULL;

    native = opendir(path);
    if (! native) {
        return errno;
    }

    return direct_directory_create(native, dir);
}

static vrpsx_errno_t direct_system_fdopendir(
    vrpsx_system_t *system,
    vrpsx_fd_t fd,
    vrpsx_directory_t **dir)
{
#if FDOPENDIR_SUPPORTED

    DIR *native;

    *dir = NULL;

    native = fdopendir(fd);
    if (! native) {
        return errno;
    }

    return direct_directory_create(native, dir);

#else

    *dir = NULL;
    return ENOSYS;

#endif
}

static vrpsx_errno_t direct_system_get_mounts(
    vrpsx_system_t *system,
    vrpsx_mount_t **mounts)
{
#if defined(__linux__) || defined(__CYGWIN__)

    VRPSX_UPCAST_NONNULL_DECL(
        direct_embedding_t, const, embedding, interface.system, system);
    direct_mount_t *obj;
    vrpsx_errno_t ern;

    *mounts = NULL;

    obj = MALLOC(direct_mount_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = embedding->mount_vtbl;
    obj->option_interface.vtbl = embedding->mount_mount_option_vtbl;
    obj->fp_native = setmntent("/proc/mounts", "r");
    if (! obj->fp_native) {
        ern = errno;
        free(obj);
        return ern;
    }

#if defined(__CYGWIN__)
    obj->reported_cygdrive_prefix = 0;
#endif

    *mounts = &obj->interface;
    return direct_mount_next(mounts, mounts, NULL);

#else

    /* Operating system not yet supported.
     */
    *mounts = NULL;
    return ENOSYS;

#endif
}

static vrpsx_errno_t trace_system_get_mounts(
    vrpsx_system_t *system,
    vrpsx_mount_t **mounts)
{
    static char const me[] = "get_mounts";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_address(&top, mounts);
    trace_end(&top);

    vrpsx_errno_t ern = direct_system_get_mounts(
        system,
        mounts);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && mounts) {
        trace_address(&top, *mounts);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_system_mount(
    vrpsx_system_t *system,
    char const *device,
    char const *root,
    char const *type,
    vrpsx_mount_option_t **options)
{
#if defined(__linux__)

    unsigned long flags = 0;
    char *csv;
    vrpsx_errno_t ern;

    ern = direct_mount_option_parse(
        direct_mount_flags,
        VRPSX_ARRAY_DIMENSION(direct_mount_flags),
        &flags,
        options,
        &csv);
    if (ern) {
        return ern;
    }

    if (mount(device, root, type, flags, csv)) {
        ern = errno;
    }

    free(csv);
    return ern;

#elif defined(__CYGWIN__)

    unsigned long flags = MOUNT_BINARY;
    char *csv;
    vrpsx_errno_t ern;

    ern = direct_mount_option_parse(
        direct_mount_flags,
        VRPSX_ARRAY_DIMENSION(direct_mount_flags),
        &flags,
        options,
        &csv);
    if (ern) {
        return ern;
    }

    if (*csv) {
        ern = EINVAL;
    }
    free(csv);
    csv = NULL;
    if (ern) {
        return ern;
    }

    if (0 == strcmp(type, "cygdrive")) {
        flags |= MOUNT_CYGDRIVE;
    }

#if CYGWIN_VERSION_DLL_MAJOR == 1005
    if (0 == strcmp(type, "system")) {
        flags |= MOUNT_SYSTEM;
    }
#endif

    if (mount(device, root, flags)) {
        return errno;
    }

    return 0;

#else

    return ENOSYS;

#endif /* operating system */
}

static vrpsx_errno_t direct_system_umount(
    vrpsx_system_t *system,
    char const *device_or_root,
    vrpsx_mount_option_t **options)
{
#if defined(__linux__)

    unsigned long flags = 0;
    char *csv;
    vrpsx_errno_t ern;

    ern = direct_mount_option_parse(
        direct_umount_flags,
        VRPSX_ARRAY_DIMENSION(direct_umount_flags),
        &flags,
        options,
        &csv);
    if (ern) {
        return ern;
    }

    if (*csv) {
        ern = EINVAL;
    }
    free(csv);
    csv = NULL;
    if (ern) {
        return ern;
    }

    if (umount2(device_or_root, flags)) {
        return errno;
    }

    return 0;

#elif defined(__CYGWIN__)

    unsigned long flags = MOUNT_SYSTEM;
    char *csv;
    vrpsx_errno_t ern;

    ern = direct_mount_option_parse(
        direct_umount_flags,
        VRPSX_ARRAY_DIMENSION(direct_umount_flags),
        &flags,
        options,
        &csv);
    if (ern) {
        return ern;
    }

    if (*csv) {
        ern = EINVAL;
    }
    free(csv);
    csv = NULL;
    if (ern) {
        return ern;
    }

    if (cygwin_umount(device_or_root, flags)) {
        return errno;
    }

    return ENOSYS;

#else

    return ENOSYS;

#endif /* operating system */
}

static vrpsx_errno_t direct_embedding_internalize_pid(
    vrpsx_embedding_t *embedding,
    void const *outer_pid,
    size_t outer_size,
    vrpsx_pid_t *inner_pid)
{
    /* In the error case the new value of *inner_pid is unspecified.
     * Let us help to detect accidental misuse of that undefined value by
     * choosing a distinctive constant for that value ("0DEF" = "unDEFined").
     */
    *inner_pid = 0x0DEF0DEF0DEF0DEFLL;

#if defined(__CYGWIN__)

    /* Windows identifies processes with a 32-bit unsigned integer,
     * though for whatever reason Cygwin expects a signed int.
     */
    vrpsx_uint32_t outer;
    pid_t cygwin_pid;

    if (outer_size != sizeof outer) {
        return EINVAL;
    }
    outer = *(vrpsx_uint32_t const *)outer_pid;

    cygwin_pid = cygwin_winpid_to_pid((int)outer);
    if (cygwin_pid == -1) {
        return ESRCH;
    }
    *inner_pid = cygwin_pid;
    return 0;

#else /* not Cygwin */

    pid_t outer;

    if (outer_size != sizeof outer) {
        return EINVAL;
    }
    outer = *(pid_t const *)outer_pid;

    *inner_pid = outer;
    return 0;

#endif /* Cygwin? */
}

static vrpsx_errno_t trace_embedding_internalize_pid(
    vrpsx_embedding_t *embedding,
    void const *outer_pid,
    size_t outer_size,
    vrpsx_pid_t *inner_pid)
{
    static char const me[] = "internalize_pid";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_octets(&top, outer_pid, outer_size);
    trace_address(&top, inner_pid);
    trace_end(&top);

    vrpsx_errno_t ern = direct_embedding_internalize_pid(
        embedding,
        outer_pid,
        outer_size,
        inner_pid);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && inner_pid) {
        trace_pid(&top, *inner_pid);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_embedding_externalize_pid(
    vrpsx_embedding_t *embedding,
    vrpsx_pid_t inner_pid,
    vrpsx_allocator_t *allocator,
    void **outer_pid,
    size_t *outer_size)
{
    *outer_pid = NULL;
    *outer_size = 0;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

#if defined(__CYGWIN__)

    /* As yet we do not know to convert from a Cygwin PID to a Windows PID.
     */
    return ENOSYS;

#else /* not Cygwin */

    pid_t *outer;

    outer = (pid_t *) VRPSX_CALL1(allocator, allocate_memory, sizeof *outer);
    if (! outer) {
        return ENOMEM;
    }

    *outer = inner_pid;
    *outer_pid = outer;
    *outer_size = sizeof *outer;
    return 0;

#endif /* Cygwin? */
}

static vrpsx_errno_t trace_embedding_externalize_pid(
    vrpsx_embedding_t *embedding,
    vrpsx_pid_t inner_pid,
    vrpsx_allocator_t *allocator,
    void **outer_pid,
    size_t *outer_size)
{
    static char const me[] = "externalize_pid";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_pid(&top, inner_pid);
    trace_address(&top, allocator);
    trace_address(&top, outer_pid);
    trace_address(&top, outer_size);
    trace_end(&top);

    vrpsx_errno_t ern = direct_embedding_externalize_pid(
        embedding,
        inner_pid,
        allocator,
        outer_pid,
        outer_size);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && outer_pid && outer_size) {
        trace_octets(&top, *outer_pid, *outer_size);
    }
    trace_end(&top);

    return ern;
}

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007

/* Given a string allocated by a given allocator, and the size of that
 * allocation, reallocates if necessary to have the minimum allocation
 * size that can express that string value.  Updates *begin and *size
 * to hold the modified location and allocation size figures.  Returns
 * 0 on success.  On error sets *begin to NULL, freeing any previous
 * allocation, sets size to 0, and returns ENOMEM.
 */
static vrpsx_errno_t shrink_singular_string(
    vrpsx_allocator_t *allocator,
    char **begin,
    size_t *size)
{
    size_t actual_size;
    HYBRID_ALLOCATOR_STACK(tmp_stack, PATH_MAX);
    hybrid_allocator_t tmp_allocator;
    char *tmp_begin;

    actual_size = strlen(*begin) + 1;
    if (actual_size == *size) {
        return 0;
    }

    assert(actual_size <= *size);

    hybrid_allocator_construct_from_array(
        &tmp_allocator,
        &allocator_singleton,
        tmp_stack);

    tmp_begin = VRPSX_CALL1(&tmp_allocator, allocate_memory, actual_size);
    if (! tmp_begin) {
        VRPSX_CALL1(allocator, free_memory, *begin);
        *begin = NULL;
        *size = 0;
        return ENOMEM;
    }
    memcpy(tmp_begin, *begin, actual_size);

    VRPSX_CALL1(allocator, free_memory, *begin);
    *begin = VRPSX_CALL1(allocator, allocate_memory, actual_size);
    if (! *begin) {
        VRPSX_CALL1(&tmp_allocator, free_memory, tmp_begin);
        *size = 0;
        return ENOMEM;
    }
    memcpy(*begin, tmp_begin, actual_size);
    *size = actual_size;

    VRPSX_CALL1(&tmp_allocator, free_memory, tmp_begin);
    return 0;
}

#endif /* defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007 */

static vrpsx_errno_t direct_embedding_internalize_path(
    vrpsx_embedding_t *embedding,
    char const *outer_path,
    vrpsx_allocator_t *allocator,
    char **inner_path)
{
#if defined(__CYGWIN__)

#if CYGWIN_VERSION_DLL_MAJOR < 1005

#error Cygwin versions before 1.5 are unsupported

#elif CYGWIN_VERSION_DLL_MAJOR == 1005

    char buffer[PATH_MAX];
    size_t size;
    char *allocation;

    *inner_path = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    if (cygwin_conv_to_posix_path(outer_path, buffer)) {
        return errno;
    }

    size = strlen(buffer) + 1;
    allocation = (char *) VRPSX_CALL1(allocator, allocate_memory, size);
    if (! allocation) {
        return ENOMEM;
    }
    memcpy(allocation, buffer, size);

    *inner_path = allocation;
    return 0;

#elif CYGWIN_VERSION_DLL_MAJOR == 1006

#error Cygwin version 1.6 unsupported

#elif CYGWIN_VERSION_DLL_MAJOR >= 1007

    char *allocation;
    ssize_t ssize;
    size_t size;
    vrpsx_errno_t ern;

    int const conv = CCP_WIN_A_TO_POSIX | CCP_RELATIVE;

    *inner_path = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    for (;;) {
        ssize = cygwin_conv_path(conv, outer_path, NULL, 0);
        if (ssize == -1) {
            return errno;
        }
        size = (size_t)ssize;

        allocation = (char *) VRPSX_CALL1(allocator, allocate_memory, size);
        if (! allocation) {
            return ENOMEM;
        }

        if (! cygwin_conv_path(conv, outer_path, allocation, size)) {
            break;
        }
        ern = errno;

        /* Just in case failure of cygwin_conv_path leaves a non-NUL char
         * at the end of the block, overwrite the end of the block with NUL
         * in order to meet the requirement that we write only NUL there.
         */
        if (size) {
            allocation[size - 1] = '\0';
        }

        VRPSX_CALL1(allocator, free_memory, allocation);

        if (ern != ENOSPC) {
            return ern;
        }

        /* The result size must have changed between we asked for the length
         * and when we then tried to actually perform the conversion, due to
         * some change in the filesystem.  Try again...
         */
    }

    /* If the filesystem changed between when we asked for the size
     * and when we performed the conversion, then we might have excess
     * capacity in the allocation.  Part of our singular allocation
     * guarantee is to avoid that, so here we reallocate if necessary.
     */
    *inner_path = allocation;
    return shrink_singular_string(allocator, inner_path, &size);

#endif /* CYGWIN_VERSION_DLL_MAJOR */

#else /* not Cygwin */

    /* Tell the caller that no conversion is required.
     */
    *inner_path = NULL;
    return 0;

#endif /* Cygwin? */
}

static vrpsx_errno_t trace_embedding_internalize_path(
    vrpsx_embedding_t *embedding,
    char const *outer_path,
    vrpsx_allocator_t *allocator,
    char **inner_path)
{
    static char const me[] = "internalize_path";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_string(&top, outer_path);
    trace_address(&top, allocator);
    trace_address(&top, inner_path);
    trace_end(&top);

    vrpsx_errno_t ern = direct_embedding_internalize_path(
        embedding,
        outer_path,
        allocator,
        inner_path);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && inner_path) {
        trace_string(&top, *inner_path);
    }
    trace_end(&top);

    return ern;
}

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007

/* Converts a narrow or wide string from from one character set to another.
 * On failure, any block that was allocated is freed.
 */
static vrpsx_errno_t typical_iconv(
    hybrid_allocator_t *allocator,  /* We make the singular allocation
                                     * guarantee for the delegate allocator,
                                     * and if the input is NUL-terminated
                                     * then only NUL will be written to the
                                     * end of any block allocated by the
                                     * delegate allocator, if any value
                                     * is written there. */
    char const *from_set,           /* Character set to convert from. */
    char const *from_begin,         /* First character to convert. */
    size_t from_size,               /* Number of bytes to convert. */
    char const *to_set,             /* Character set to convert to; */
    char const **to_begin,          /* Set to beginning of result. */
    size_t *to_size,                /* Set to number of bytes in result. */
    char **allocation)              /* Allocation backing *to_begin;
                                     * NULL if we can see that from_set and
                                     * to_set are the same codeset, in which
                                     * case we set *to_begin = from_begin. */
{
    iconv_t cd = (iconv_t)(-1);
    size_t available;
    char *in_ptr;
    size_t in_rem;
    char *out_ptr;
    size_t out_rem;
    size_t total_size;
    char count_buffer[256];
    vrpsx_errno_t ern;

    *allocation = NULL;
    *to_begin = NULL;
    *to_size = 0;

    if (! strcmp(to_set, from_set)) {
        *to_begin = from_begin;
        *to_size = from_size;
        ern = 0;
        goto done;
    }

    cd = iconv_open(to_set, from_set);
    if (cd == (iconv_t)(-1)) {
        ern = errno;
        goto done;
    }

    in_ptr = (char *)from_begin;
    in_rem = from_size;

    /* Start with whatever is left of the hybrid allocator's stack.
     */
    *allocation = (char *) VRPSX_CALL1(allocator, allocate_available,
            &available);
    if (*allocation) {
        out_ptr = *allocation;
        out_rem = available;
        ern = 0;
        if (iconv(cd, &in_ptr, &in_rem, &out_ptr, &out_rem) == (size_t)(-1)) {
            ern = errno;
        }
        if (! ern) {
            *to_begin = *allocation;
            *to_size = available - out_rem;
            VRPSX_CALL1(allocator, shrink_final, *to_size);
            goto done;
        }
        if (ern != E2BIG) {
            goto done;
        }
        VRPSX_CALL1(allocator, free_memory, *allocation);
        *allocation = NULL;
    }

    /* The available space in the hybrid allocator's stack was
     * insufficient to the task.  Finish the conversion while
     * discarding most of it, just to find out how much memory
     * we would actually need.
     */
    total_size = available - out_rem;

    for (;;) {
        out_ptr = count_buffer;
        out_rem = sizeof count_buffer;
        ern = 0;
        if (iconv(cd, &in_ptr, &in_rem, &out_ptr, &out_rem) == (size_t)(-1)) {
            ern = errno;
        }

        total_size += sizeof count_buffer - out_rem;

        if (! ern) {
            break;
        }
        if (ern != E2BIG) {
            goto done;
        }
    }

    /* Must reopen in case conversion is stateful.  It would
     * be faster if we could save and restore conversion state;
     * then we would not need to reconvert the initial portion.
     * Unfortunately the iconv API seems to omit save/restore.
     */
    if (iconv_close(cd)) {
        ern = errno;
        cd = (iconv_t)(-1);
        goto done;
    }
    cd = iconv_open(to_set, from_set);
    if (cd == (iconv_t)(-1)) {
        ern = errno;
        goto done;
    }

    /* Now that we know the final size of memory block we need,
     * allocate that block and write the conversion into it.
     */
    *allocation = (char *) VRPSX_CALL1(allocator, allocate_memory, total_size);
    if (! *allocation) {
        ern = ENOMEM;
        goto done;
    }

    in_ptr = (char *)from_begin;
    in_rem = from_size;
    out_ptr = *allocation;
    out_rem = total_size;
    ern = 0;
    if (iconv(cd, &in_ptr, &in_rem, &out_ptr, &out_rem) == (size_t)(-1)) {
        ern = errno;
    }
    if (ern) {
        /* We are obligated to leave NUL in the final allocated char if the
         * input is NUL-terminated.  Conceivably we might need to clean up
         * after iconv failure, as follows.
         */
        if (total_size) {
            (*allocation)[total_size - 1] = '\0';
        }
        goto done;
    }

    assert(out_rem == 0); /* conversion should be deterministic */

    *to_begin = *allocation;
    *to_size = total_size - out_rem;

done:
    if (ern && *allocation) {
        VRPSX_CALL1(allocator, free_memory, *allocation);
    }
    if (cd != (iconv_t)(-1)) {
        iconv_close(cd);
    }
    return ern;
}

#endif /* defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007 */

static vrpsx_errno_t direct_embedding_externalize_path(
    vrpsx_embedding_t *embedding,
    char const *inner_path,
    vrpsx_allocator_t *allocator,
    char **outer_path)
{
#if defined(__CYGWIN__)

#if CYGWIN_VERSION_DLL_MAJOR < 1005

#error Cygwin versions before 1.5 are unsupported

#elif CYGWIN_VERSION_DLL_MAJOR == 1005

    char buffer[PATH_MAX];
    size_t size;
    char *allocation;

    *outer_path = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    if (cygwin_conv_to_win32_path(inner_path, buffer)) {
        return errno;
    }

    size = strlen(buffer) + 1;
    allocation = (char *) VRPSX_CALL1(allocator, allocate_memory, size);
    if (! allocation) {
        return ENOMEM;
    }
    memcpy(allocation, buffer, size);

    *outer_path = allocation;
    return 0;

#elif CYGWIN_VERSION_DLL_MAJOR == 1006

#error Cygwin version 1.6 unsupported

#elif CYGWIN_VERSION_DLL_MAJOR >= 1007

    ssize_t wide_bytes;
    HYBRID_ALLOCATOR_STACK(wide_stack, PATH_MAX * sizeof(wchar_t));
    hybrid_allocator_t wide_allocator;
    wchar_t *wide_begin;

    HYBRID_ALLOCATOR_STACK(narrow_stack, PATH_MAX);
    hybrid_allocator_t narrow_allocator;
    char const *narrow_begin;
    size_t narrow_size;
    char *narrow_allocation;
    vrpsx_errno_t ern;

    /* We avoid CCP_POSIX_TO_WIN_A because it cannot convert very
     * long paths.  As a consequence we must convert to UTF-8 here,
     * though there should be little difference in run time between
     * the two conversion options.
     *
     * There is an additional side benefit to avoiding CCP_POSIX_TO_WIN_A,
     * but in order to explain it we must first provide a bit of background:
     *
     * Cygwin 1.7 permits characters such as '*' to appear in Cygwin
     * pathnames even though they are forbidden in Windows pathnames.
     * In order to make this work, Cygwin translates such characters
     * into Unicode private use areas when converting to Windows pathnames,
     * and reverses that translation when converting from Windows pathnames
     * to Cygwin pathnames.
     *
     * But when CCP_POSIX_TO_WIN_A converts from Cygwin pathnames to
     * a narrow encoding of a Windows pathname, it appears to apply
     * the same translation of Unicode private use areas that Cygwin
     * does when converting from Windows pathnames to Cygwin pathnames,
     * even though the conversion direction is quite the opposite.
     * The probably-unintended result is invalid Windows pathnames.
     *
     * Because we use CCP_POSIX_TO_WIN_W and then convert to UTF-8
     * using iconv--which does not perform this special translation
     * of Unicode private use areas--we avoid that particular problem.
     *
     * KNOWN BUG:
     *
     * Unfortunately, experiment indicates that in Cygwin 1.7.9,
     * CCP_POSIX_TO_WIN_W encodes forbidden characters such as '*'
     * into the Unicode private use areas only when producing an
     * absolute path (either because the input path was absolute
     * or because CCP_RELATIVE was not used).  In the relative case
     * the result is L'*', which is invalid in wide Windows pathnames.
     * As yet we have not decided how to address this particular bug.
     */
    int const conv = CCP_POSIX_TO_WIN_W | CCP_RELATIVE;

    *outer_path = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    hybrid_allocator_construct_from_array(
        &wide_allocator,
        &allocator_singleton,
        wide_stack);

    for (;;) {
        wide_bytes = cygwin_conv_path(conv, inner_path, NULL, 0);
        if (wide_bytes == -1) {
            return errno;
        }

        wide_begin = (wchar_t *) VRPSX_CALL1(&wide_allocator, allocate_memory,
                wide_bytes);
        if (! wide_begin) {
            return ENOMEM;
        }

        if (! cygwin_conv_path(conv, inner_path,
                        wide_begin, (size_t)wide_bytes)) {
            break;
        }
        ern = errno;

        VRPSX_CALL1(&wide_allocator, free_memory, wide_begin);

        if (ern != ENOSPC) {
            return ern;
        }

        /* The result size must have changed between we asked for the length
         * and when we then tried to actually perform the conversion, due to
         * some change in the filesystem.  Try again...
         */
    }

    /* If the filesystem changed between when we asked for the size
     * and when we performed the conversion, then we might have excess
     * capacity in the allocation.
     */
    wide_bytes = (ssize_t)(wcslen(wide_begin) + 1) * sizeof(wchar_t);

    hybrid_allocator_construct_from_array(
        &narrow_allocator,
        allocator,
        narrow_stack);

    ern = typical_iconv(
        &narrow_allocator,
        "UTF-16LE",
        (char *)wide_begin,
        (size_t)wide_bytes,
        "UTF-8",
        &narrow_begin,
        &narrow_size,
        &narrow_allocation);
    VRPSX_CALL1(&wide_allocator, free_memory, wide_begin);
    if (ern) {
        return ern;
    }

    if (STACK_FOR_HYBRID_ALLOCATOR_CONTAINS(narrow_stack, narrow_allocation)) {
        /* In this case we allow (narrow_allocation) to be freed
         * implicitly when (narrow_stack) goes out of scope.
         */
        *outer_path = VRPSX_CALL1(allocator, allocate_memory, narrow_size);
        if (! *outer_path) {
            return ENOMEM;
        }
        memcpy(*outer_path, narrow_begin, narrow_size);
        return 0;
    } else {
        /* In this case (narrow_allocation) was allocated by the
         * (allocator) and can be managed directly by (allocator)
         * without any further use of (narrow_allocator).
         */
        *outer_path = narrow_allocation;
        return 0;
    }

#endif /* CYGWIN_VERSION_DLL_MAJOR */

#else /* not Cygwin */

    /* Tell the caller that no conversion is required.
     */
    *outer_path = NULL;
    return 0;

#endif /* Cygwin? */
}

static vrpsx_errno_t trace_embedding_externalize_path(
    vrpsx_embedding_t *embedding,
    char const *inner_path,
    vrpsx_allocator_t *allocator,
    char **outer_path)
{
    static char const me[] = "externalize_path";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_string(&top, inner_path);
    trace_address(&top, allocator);
    trace_address(&top, outer_path);
    trace_end(&top);

    vrpsx_errno_t ern = direct_embedding_externalize_path(
        embedding,
        inner_path,
        allocator,
        outer_path);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && outer_path) {
        trace_string(&top, *outer_path);
    }
    trace_end(&top);

    return ern;
}

#if CYGWIN_VERSION_DLL_MAJOR == 1005

typedef struct path_node_struct {
    struct path_node_struct *next;
                                /* The next node in the list of pathnames. */
    size_t size;                /* The length of the pathname, not counting
                                 * the terminating NUL character. */
    char path[1];               /* The pathname.  The actual dimension
                                 * of this array is (size + 1). */
} path_node_t;

typedef struct path_list_struct {
    path_node_t *head;          /* The first path_node_t in the list,
                                 * but NULL if the list is empty. */
    path_node_t **tail;         /* The "next" member of the last path_node_t in
                                 * the list, but &head if the list is empty. */
} path_list_t;

/* Initializes the given path list variable to express an empty list.
 * Never fails.  Subsequent destruction is required only if nonempty.
*/
static void path_list_construct(
    path_list_t *list)
{
    *(list->tail = &list->head) = NULL;
}

/* Clears the path list.  Also functions as a destructor.
*/
static void path_list_clear(
    path_list_t *list)
{
    path_node_t *node = list->head;

    *(list->tail = &list->head) = NULL;

    while (node) {
        path_node_t *doomed = node;
        node = doomed->next;
        free(doomed);
    }
}

/* Allocates and appends a new node.
 */
static vrpsx_errno_t path_list_append(
    path_list_t *list,
    char const *path)
{
    size_t size = strlen(path);
    path_node_t *node = (path_node_t *)malloc(
        offsetof(path_node_t, path) + size + 1);
    if (! node) {
        return errno;
    }

    node->next = NULL;
    node->size = size;
    memcpy(&node->path, path, size);
    node->path[size] = '\0';

    *list->tail = node;
    list->tail = &node->next;
    return 0;
}

/* Creates a string representation of the path list,
 * using the given separator to separate the elements.
 */
static vrpsx_errno_t path_list_concat(
    path_list_t const *list,
    char separator,
    vrpsx_allocator_t *allocator, /* singular allocation;
                                   * NUL-terminated allocation;
                                   * must not be NULL */
    char **out)                   /* receives result, or NULL on failure */
{
    char *str, *dp;
    size_t size;
    path_node_t const *node = list->head;

    *out = NULL;

    /* Special case for empty list.
     */
    if (node == NULL) {
        str = (char *)VRPSX_CALL1(allocator, allocate_memory, 1);
        if (! str) {
            return ENOMEM;
        }

        *str = '\0';

        *out = str;
        return 0;
    }

    /* Compute the size, including separators and the terminator.
     */
    size = 0;
    do {
        size += node->size + 1;
    } while ((node = node->next) != NULL);

    /* Allocate memory for the result.
     */
    str = (char *)VRPSX_CALL1(allocator, allocate_memory, size);
    if (! str) {
        return ENOMEM;
    }

    /* Allocation succeeded, so compose the path list.
     */
    dp = str;
    for (node = list->head; node; node = node->next) {
        MEMCPY_END(dp, node->path, node->size);
        *dp++ = separator;
    }
    dp[-1] = '\0';

    assert((size_t)(dp - str) == size);

    *out = str;
    return 0;
}

/* Given the address of a scanning pointer, extracts the next pathname,
 * or NULL if there are no more pathnames.  Modifies the string as we go.
 *
 * We cannot use strtok_r here because it would elide empty elements of
 * the path list, whereas we need this routine to return empty pathnames,
 * because in a Cygwin path list those mean the current directory.
 */
static char *path_split(
    char **scan,
    char separator)     /* Separator character--must not be NUL. */
{
    char *begin, *end;

    begin = *scan;

    if (! *begin) {
        return NULL;
    }

    end = strchr(begin, separator);
    if (end) {
        *scan = end + 1;
        *end = '\0';
    } else {
        *scan = begin + strlen(begin);
    }

    return begin;
}

#endif /* CYGWIN_VERSION_DLL_MAJOR == 1005 */

static vrpsx_errno_t direct_embedding_internalize_execution_path(
    vrpsx_embedding_t *embedding,
    char const *outer_path,
    vrpsx_allocator_t *allocator,
    char **inner_path)
{
#if defined(__CYGWIN__)

#if CYGWIN_VERSION_DLL_MAJOR < 1005

#error Cygwin versions before 1.5 are unsupported

#elif CYGWIN_VERSION_DLL_MAJOR == 1005

    *inner_path = NULL;

    /* Special case: cygwin_win32_to_posix_path_list converts "" to "", which
     * we can indicate without actually needing to allocate a copy of "".
     */
    if (! *outer_path) {
        return 0;
    }

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    /* At least in Cygwin 1.5.25-15, failure of cygwin_win32_to_posix_path_list
     * often results in an incorrect errno value of -1, due to a mistake in how
     * that function propagates errors through its subroutines.  As yet we have
     * not taken the time to fully optimize the following alternative:
     */

    /* General case: split a copy of the path list string at ';' characters,
     * and convert the resulting pathname elements individually.
     */
    size_t outer_path_size = strlen(outer_path);
    char *outer, *scan = (char *)alloca(outer_path_size + 1);
    memcpy(scan, outer_path, outer_path_size + 1);

    path_list_t list;
    path_list_construct(&list);

    char inner[PATH_MAX];
    vrpsx_errno_t ern = 0;

    while (NULL != (outer = path_split(&scan, ';'))) {
        /* Empty elements of a Windows path list have no effect.
         */
        if (! *outer) {
            continue;
        }

        /* Convert an individual pathname element.
         */
        if (cygwin_conv_to_posix_path(outer, inner)) {
            ern = errno;
            goto done;
        }

        /* Keep track of the result of that conversion.
         */
        ern = path_list_append(&list, inner);
        if (ern) {
            goto done;
        }
    }

    /* Compose the ':'-separated path list from the individual results.
     */
    ern = path_list_concat(&list, ':', allocator, inner_path);

done:
    path_list_clear(&list);
    return ern;

#elif CYGWIN_VERSION_DLL_MAJOR == 1006

#error Cygwin version 1.6 unsupported

#elif CYGWIN_VERSION_DLL_MAJOR >= 1007

    char *allocation;
    ssize_t ssize;
    size_t size;
    vrpsx_errno_t ern;

    int const conv = CCP_WIN_A_TO_POSIX | CCP_RELATIVE;

    *inner_path = NULL;

    /* Special case:
     * cygwin_conv_path_list(CCP_WIN_A_TO_POSIX | CCP_RELATIVE, ...)
     * converts "" to "", which we can indicate without actually
     * needing to allocate a copy of "".
     */
    if (! *outer_path) {
        return 0;
    }

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    for (;;) {
        ssize = cygwin_conv_path_list(conv, outer_path, NULL, 0);
        if (ssize == -1) {
            return errno;
        }
        size = (size_t)ssize;

        allocation = (char *) VRPSX_CALL1(allocator, allocate_memory, size);
        if (! allocation) {
            return ENOMEM;
        }

        if (! cygwin_conv_path_list(conv, outer_path, allocation, size)) {
            break;
        }
        ern = errno;

        /* Just in case failure of cygwin_conv_path_list leaves a non-NUL char
         * at the end of the block, overwrite the end of the block with NUL
         * in order to meet the requirement that we write only NUL there.
         */
        if (size) {
            allocation[size - 1] = '\0';
        }

        VRPSX_CALL1(allocator, free_memory, allocation);

        if (ern != ENOSPC) {
            return ern;
        }

        /* The result size must have changed between we asked for the length
         * and when we then tried to actually perform the conversion, due to
         * some change in the filesystem.  Try again...
         */
    }

    /* Cygwin 1.7.9 appears to routinely overestimate the result size
     * of a path list conversion.  But even if that were not the case:
     *
     * If the filesystem changed between when we asked for the size
     * and when we performed the conversion, then we might have excess
     * capacity in the allocation.  Part of our singular allocation
     * guarantee is to avoid that, so here we reallocate if necessary.
     */
    *inner_path = allocation;
    return shrink_singular_string(allocator, inner_path, &size);

#endif /* CYGWIN_VERSION_DLL_MAJOR */

#else /* not Cygwin */

    /* Tell the caller that no conversion is required.
     */
    *inner_path = NULL;
    return 0;

#endif /* Cygwin? */
}

static vrpsx_errno_t trace_embedding_internalize_execution_path(
    vrpsx_embedding_t *embedding,
    char const *outer_path,
    vrpsx_allocator_t *allocator,
    char **inner_path)
{
    static char const me[] = "internalize_execution_path";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_string(&top, outer_path);
    trace_address(&top, allocator);
    trace_address(&top, inner_path);
    trace_end(&top);

    vrpsx_errno_t ern = direct_embedding_internalize_execution_path(
        embedding,
        outer_path,
        allocator,
        inner_path);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && inner_path) {
        trace_string(&top, *inner_path);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_errno_t direct_embedding_externalize_execution_path(
    vrpsx_embedding_t *embedding,
    char const *inner_path,
    vrpsx_allocator_t *allocator,
    char **outer_path)
{
#if defined(__CYGWIN__)

#if CYGWIN_VERSION_DLL_MAJOR < 1005

#error Cygwin versions before 1.5 are unsupported

#elif CYGWIN_VERSION_DLL_MAJOR == 1005

    *outer_path = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    /* At least in Cygwin 1.5.25-15, failure of cygwin_posix_to_win32_path_list
     * often results in an incorrect errno value of -1, due to a mistake in how
     * that function propagates errors through its subroutines.  As yet we have
     * not taken the time to fully optimize the following alternative:
     */

    /* Special case: cygwin_posix_to_win32_path_list converts "" to ".".
     */
    if (! *inner_path) {
        char *dot = (char *)VRPSX_CALL1(allocator, allocate_memory, 2);
        *outer_path = dot;
        if (dot) {
            dot[0] = '.';
            dot[1] = '\0';
            return 0;
        } else {
            return ENOMEM;
        }
    }

    /* General case: split a copy of the path list string at ':' characters,
     * and convert the resulting pathname elements individually.
     */
    size_t inner_path_size = strlen(inner_path);
    char *inner, *scan = (char *)alloca(inner_path_size + 1);
    memcpy(scan, inner_path, inner_path_size + 1);

    path_list_t list;
    path_list_construct(&list);

    char outer[PATH_MAX];
    vrpsx_errno_t ern = 0;

    while (NULL != (inner = path_split(&scan, ':'))) {
        /* Empty elements of a Cygwin path list mean the current directory.
         */
        if (! *inner) {
            inner = ".";
        }

        /* Convert an individual pathname element.
         */
        if (cygwin_conv_to_win32_path(inner, outer)) {
            ern = errno;
            goto done;
        }

        /* Keep track of the result of that conversion.
         */
        ern = path_list_append(&list, outer);
        if (ern) {
            goto done;
        }
    }

    /* Compose the ';'-separated path list from the individual results.
     */
    ern = path_list_concat(&list, ';', allocator, outer_path);

done:
    path_list_clear(&list);
    return ern;

#elif CYGWIN_VERSION_DLL_MAJOR == 1006

#error Cygwin version 1.6 unsupported

#elif CYGWIN_VERSION_DLL_MAJOR >= 1007

    char *allocation;
    ssize_t ssize;
    size_t size;
    vrpsx_errno_t ern;

    /* We would like to avoid CCP_POSIX_TO_WIN_A here for the same
     * reasons that we do in direct_embedding_externalize_path.
     *
     * But at least on Cygwin 1.7.9, use of CCP_POSIX_TO_WIN_W
     * when converting path lists results in this fatal error:
     *
     *     *** fatal error - wide char path lists not yet supported
     *
     * At least for now we just use the narrow conversion instead.
     */
    int const conv = CCP_POSIX_TO_WIN_A | CCP_RELATIVE;

    *outer_path = NULL;

    if (! allocator) {
        allocator = &allocator_singleton;
    }

    for (;;) {
        ssize = cygwin_conv_path_list(conv, inner_path, NULL, 0);
        if (ssize == -1) {
            return errno;
        }
        size = (size_t)ssize;

        allocation = (char *) VRPSX_CALL1(allocator, allocate_memory, size);
        if (! allocation) {
            return ENOMEM;
        }

        if (! cygwin_conv_path_list(conv, inner_path, allocation, size)) {
            break;
        }
        ern = errno;

        /* Just in case failure of cygwin_conv_path_list leaves a non-NUL char
         * at the end of the block, overwrite the end of the block with NUL
         * in order to meet the requirement that we write only NUL there.
         */
        if (size) {
            allocation[size - 1] = '\0';
        }

        VRPSX_CALL1(allocator, free_memory, allocation);

        if (ern != ENOSPC) {
            return ern;
        }

        /* The result size must have changed between we asked for the length
         * and when we then tried to actually perform the conversion, due to
         * some change in the filesystem.  Try again...
         */
    }

    /* Cygwin 1.7.9 appears to routinely overestimate the result size
     * of a path list conversion.  But even if that were not the case:
     *
     * If the filesystem changed between when we asked for the size
     * and when we performed the conversion, then we might have excess
     * capacity in the allocation.  Part of our singular allocation
     * guarantee is to avoid that, so here we reallocate if necessary.
     */
    *outer_path = allocation;
    return shrink_singular_string(allocator, outer_path, &size);

#endif /* CYGWIN_VERSION_DLL_MAJOR */

#else /* not Cygwin */

    /* Tell the caller that no conversion is required.
     */
    *outer_path = NULL;
    return 0;

#endif /* Cygwin? */
}

static vrpsx_errno_t trace_embedding_externalize_execution_path(
    vrpsx_embedding_t *embedding,
    char const *inner_path,
    vrpsx_allocator_t *allocator,
    char **outer_path)
{
    static char const me[] = "externalize_execution_path";
    trace_stack_t top;
    trace_enter_begin(&top, me);
    trace_string(&top, inner_path);
    trace_address(&top, allocator);
    trace_address(&top, outer_path);
    trace_end(&top);

    vrpsx_errno_t ern = direct_embedding_externalize_execution_path(
        embedding,
        inner_path,
        allocator,
        outer_path);

    trace_leave_begin_ern(&top, me, ern);
    if (! ern && outer_path) {
        trace_string(&top, *outer_path);
    }
    trace_end(&top);

    return ern;
}

static vrpsx_system_operations_end_t const system_operations_end;

static vrpsx_system_errno_end_t const system_errno_end;

static vrpsx_system_signo_end_t const system_signo_end;

static vrpsx_system_signal_end_t const system_signal_end;

static vrpsx_system_thread_end_t const system_thread_end;

static vrpsx_system_fcntl_end_t const system_fcntl_end;

static vrpsx_system_stat_end_t const system_stat_end;

static vrpsx_system_time_end_t const system_time_end;

static vrpsx_system_limits_end_t const system_limits_end;

static vrpsx_system_stdlib_end_t const system_stdlib_end;

static vrpsx_system_sysconf_end_t const system_sysconf_end;

static vrpsx_system_pathconf_end_t const system_pathconf_end;

static vrpsx_system_end_t const system_end;

static vrpsx_embedding_end_t const embedding_end;

/* On platforms that do not provide a monotonic clock we still need
 * a placeholder in the interface for the identifier for such a clock.
 * And if some client of the interface neglects to check sysconf
 * before trying to use that identifier, then we want some value that
 * at least approximates what they want.  Hence the following fallback:
 */
#ifndef CLOCK_MONOTONIC
#define CLOCK_MONOTONIC CLOCK_REALTIME
#endif

/* The concept expressed by our "enomedium" symbol concerns
 * widely available hardware, and seems generally applicable.
 *
 * Some future version POSIX might standardize a symbol for it,
 * but meanwhile we can map platform-specific symbols to it.
 * Both Linux and Cygwin define ENOMEDIUM for this concept.
 *
 * On platforms where we have not yet found a number for this
 * concept, we use a nonzero number distinct from actual error
 * numbers, as specified by the interface we are implementing.
 */
#ifndef ENOMEDIUM
#define ENOMEDIUM (-1)
#endif

#define SYSTEM_VTBL_INITIALIZATION(PREFIX)                              \
                                                                        \
static vrpsx_system_vtbl_t const VRPSX_CAT2(PREFIX, _system_vtbl) = {   \
    vrpsx_system_IDENTIFIER,                                            \
    &direct_system_replace_interface,                                   \
    &identified_end,                                                    \
    &direct_system_release,                                             \
    &direct_system_acquire,                                             \
    &counted_identified_end,                                            \
    &direct_system_error_acquire,                                       \
    &direct_system_free_memory,                                         \
    &direct_system_allocate_memory,                                     \
    &direct_system_exit,                                                \
    &direct_system_uname,                                               \
    &direct_system_sysconf,                                             \
    &direct_system_pathconf,                                            \
    &direct_system_fpathconf,                                           \
    &direct_system_getcwd,                                              \
    &direct_system_get_timezone,                                        \
    &direct_system_get_tzname,                                          \
    &direct_system_gmtime,                                              \
    &direct_system_localtime,                                           \
    &direct_system_mktime,                                              \
    &direct_system_time,                                                \
    &direct_system_clock_gettime,                                       \
    &direct_system_nanosleep,                                           \
    &direct_system_getlogin,                                            \
    &direct_system_getpwnam,                                            \
    &direct_system_getpid,                                              \
    &direct_system_getpgrp,                                             \
    &direct_system_getpgid,                                             \
    &direct_system_kill,                                                \
    &VRPSX_CAT2(PREFIX, _system_raise),                                 \
    &direct_system_strsignal,                                           \
    &direct_system_sigset_init,                                         \
    &direct_system_sigaction_init,                                      \
    &direct_system_sigaction,                                           \
    &VRPSX_CAT2(PREFIX, _system_sigwait),                               \
    &VRPSX_CAT2(PREFIX, _system_thread_sigmask),                        \
    &direct_system_thread_attr_create,                                  \
    &VRPSX_CAT2(PREFIX, _system_thread_create),                         \
    &direct_system_thread_detach,                                       \
    &VRPSX_CAT2(PREFIX, _system_thread_join),                           \
    &VRPSX_CAT2(PREFIX, _system_thread_kill),                           \
    &direct_system_thread_equal,                                        \
    &direct_system_thread_self,                                         \
    &direct_system_thread_key_create,                                   \
    &direct_system_mutex_attr_create,                                   \
    &direct_system_mutex_create,                                        \
    &direct_system_rwlock_attr_create,                                  \
    &direct_system_rwlock_create,                                       \
    &direct_system_cond_attr_create,                                    \
    &direct_system_cond_create,                                         \
    &direct_system_open,                                                \
    &direct_system_close,                                               \
    &direct_system_read,                                                \
    &direct_system_write,                                               \
    &direct_system_pread,                                               \
    &direct_system_pwrite,                                              \
    &direct_system_unlink,                                              \
    &direct_system_link,                                                \
    &direct_system_symlink,                                             \
    &direct_system_readlink,                                            \
    &direct_system_chmod,                                               \
    &direct_system_fchmod,                                              \
    &direct_system_stat,                                                \
    &direct_system_lstat,                                               \
    &direct_system_fstat,                                               \
    &direct_system_mkdir,                                               \
    &direct_system_rmdir,                                               \
    &direct_system_chdir,                                               \
    &direct_system_opendir,                                             \
    &direct_system_fdopendir,                                           \
    &VRPSX_CAT2(PREFIX, _system_get_mounts),                            \
    &direct_system_mount,                                               \
    &direct_system_umount,                                              \
    &system_operations_end,                                             \
    EXIT_FAILURE,                                                       \
    EXIT_SUCCESS,                                                       \
    &system_stdlib_end,                                                 \
    E2BIG,                                                              \
    EACCES,                                                             \
    EADDRINUSE,                                                         \
    EADDRNOTAVAIL,                                                      \
    EAFNOSUPPORT,                                                       \
    EAGAIN,                                                             \
    EALREADY,                                                           \
    EBADF,                                                              \
    EBADMSG,                                                            \
    EBUSY,                                                              \
    ECANCELED,                                                          \
    ECHILD,                                                             \
    ECONNABORTED,                                                       \
    ECONNREFUSED,                                                       \
    ECONNRESET,                                                         \
    EDEADLK,                                                            \
    EDESTADDRREQ,                                                       \
    EDOM,                                                               \
    EDQUOT,                                                             \
    EEXIST,                                                             \
    EFAULT,                                                             \
    EFBIG,                                                              \
    EHOSTUNREACH,                                                       \
    EIDRM,                                                              \
    EILSEQ,                                                             \
    EINPROGRESS,                                                        \
    EINTR,                                                              \
    EINVAL,                                                             \
    EIO,                                                                \
    EISCONN,                                                            \
    EISDIR,                                                             \
    ELOOP,                                                              \
    EMFILE,                                                             \
    EMLINK,                                                             \
    EMSGSIZE,                                                           \
    EMULTIHOP,                                                          \
    ENAMETOOLONG,                                                       \
    ENETDOWN,                                                           \
    ENETUNREACH,                                                        \
    ENFILE,                                                             \
    ENOBUFS,                                                            \
    ENODATA,                                                            \
    ENODEV,                                                             \
    ENOENT,                                                             \
    ENOEXEC,                                                            \
    ENOLCK,                                                             \
    ENOLINK,                                                            \
    ENOMEDIUM,                                                          \
    ENOMEM,                                                             \
    ENOMSG,                                                             \
    ENOPROTOOPT,                                                        \
    ENOSPC,                                                             \
    ENOSR,                                                              \
    ENOSTR,                                                             \
    ENOSYS,                                                             \
    ENOTCONN,                                                           \
    ENOTDIR,                                                            \
    ENOTEMPTY,                                                          \
    ENOTSOCK,                                                           \
    ENOTSUP,                                                            \
    ENOTTY,                                                             \
    ENXIO,                                                              \
    EOPNOTSUPP,                                                         \
    EOVERFLOW,                                                          \
    EPERM,                                                              \
    EPIPE,                                                              \
    EPROTO,                                                             \
    EPROTONOSUPPORT,                                                    \
    EPROTOTYPE,                                                         \
    ERANGE,                                                             \
    EROFS,                                                              \
    ESPIPE,                                                             \
    ESRCH,                                                              \
    ESTALE,                                                             \
    ETIME,                                                              \
    ETIMEDOUT,                                                          \
    ETXTBSY,                                                            \
    EWOULDBLOCK,                                                        \
    EXDEV,                                                              \
    &system_errno_end,                                                  \
    SIGABRT,                                                            \
    SIGALRM,                                                            \
    SIGBUS,                                                             \
    SIGCHLD,                                                            \
    SIGCONT,                                                            \
    SIGFPE,                                                             \
    SIGHUP,                                                             \
    SIGILL,                                                             \
    SIGINT,                                                             \
    SIGKILL,                                                            \
    SIGPIPE,                                                            \
    SIGPOLL,                                                            \
    SIGPROF,                                                            \
    SIGQUIT,                                                            \
    SIGSEGV,                                                            \
    SIGSTOP,                                                            \
    SIGSYS,                                                             \
    SIGTERM,                                                            \
    SIGTRAP,                                                            \
    SIGTSTP,                                                            \
    SIGTTIN,                                                            \
    SIGTTOU,                                                            \
    SIGURG,                                                             \
    SIGUSR1,                                                            \
    SIGUSR2,                                                            \
    SIGVTALRM,                                                          \
    SIGXCPU,                                                            \
    SIGXFSZ,                                                            \
    &system_signo_end,                                                  \
    SIG_DFL,                                                            \
    SIG_ERR,                                                            \
    SIG_IGN,                                                            \
    sizeof(direct_bundled_sigset_t),                                    \
    DIRECT_BUNDLED_SIGSET_ALIGNOF,                                      \
    sizeof(direct_sigaction_t),                                         \
    DIRECT_SIGACTION_ALIGNOF,                                           \
    SA_NOCLDSTOP,                                                       \
    SA_NOCLDWAIT,                                                       \
    SA_NODEFER,                                                         \
    SA_ONSTACK,                                                         \
    SA_RESETHAND,                                                       \
    SA_RESTART,                                                         \
    SA_SIGINFO,                                                         \
    SIG_BLOCK,                                                          \
    SIG_SETMASK,                                                        \
    SIG_UNBLOCK,                                                        \
    &system_signal_end,                                                 \
    PTHREAD_CREATE_DETACHED,                                            \
    PTHREAD_CREATE_JOINABLE,                                            \
    PTHREAD_MUTEX_DEFAULT,                                              \
    PTHREAD_MUTEX_ERRORCHECK,                                           \
    PTHREAD_MUTEX_NORMAL,                                               \
    PTHREAD_MUTEX_RECURSIVE,                                            \
    PTHREAD_PROCESS_PRIVATE,                                            \
    PTHREAD_PROCESS_SHARED,                                             \
    PTHREAD_SCOPE_PROCESS,                                              \
    PTHREAD_SCOPE_SYSTEM,                                               \
    &system_thread_end,                                                 \
    F_DUPFD,                                                            \
    F_GETFD,                                                            \
    F_GETFL,                                                            \
    F_GETLK,                                                            \
    F_RDLCK,                                                            \
    F_SETFD,                                                            \
    F_SETFL,                                                            \
    F_SETLK,                                                            \
    F_SETLKW,                                                           \
    F_UNLCK,                                                            \
    F_WRLCK,                                                            \
    FD_CLOEXEC,                                                         \
    O_ACCMODE,                                                          \
    O_APPEND,                                                           \
    O_CREAT,                                                            \
    O_DSYNC,                                                            \
    O_EXCL,                                                             \
    O_NOCTTY,                                                           \
    O_NONBLOCK,                                                         \
    O_RDONLY,                                                           \
    O_RDWR,                                                             \
    O_RSYNC,                                                            \
    O_SYNC,                                                             \
    O_TRUNC,                                                            \
    O_WRONLY,                                                           \
    &system_fcntl_end,                                                  \
    S_IFMT,                                                             \
    S_IFBLK,                                                            \
    S_IFCHR,                                                            \
    S_IFIFO,                                                            \
    S_IFREG,                                                            \
    S_IFDIR,                                                            \
    S_IFLNK,                                                            \
    S_IFSOCK,                                                           \
    &system_stat_end,                                                   \
    CLOCK_MONOTONIC,                                                    \
    CLOCK_REALTIME,                                                     \
    &system_time_end,                                                   \
    NAME_MAX,                                                           \
    PATH_MAX,                                                           \
    &system_limits_end,                                                 \
    _SC_2_C_BIND,                                                       \
    _SC_2_C_DEV,                                                        \
    _SC_2_CHAR_TERM,                                                    \
    _SC_2_FORT_DEV,                                                     \
    _SC_2_FORT_RUN,                                                     \
    _SC_2_LOCALEDEF,                                                    \
    _SC_2_PBS,                                                          \
    _SC_2_PBS_ACCOUNTING,                                               \
    _SC_2_PBS_CHECKPOINT,                                               \
    _SC_2_PBS_LOCATE,                                                   \
    _SC_2_PBS_MESSAGE,                                                  \
    _SC_2_PBS_TRACK,                                                    \
    _SC_2_SW_DEV,                                                       \
    _SC_2_UPE,                                                          \
    _SC_2_VERSION,                                                      \
    _SC_ADVISORY_INFO,                                                  \
    _SC_AIO_LISTIO_MAX,                                                 \
    _SC_AIO_MAX,                                                        \
    _SC_AIO_PRIO_DELTA_MAX,                                             \
    _SC_ARG_MAX,                                                        \
    _SC_ASYNCHRONOUS_IO,                                                \
    _SC_ATEXIT_MAX,                                                     \
    _SC_BARRIERS,                                                       \
    _SC_BC_BASE_MAX,                                                    \
    _SC_BC_DIM_MAX,                                                     \
    _SC_BC_SCALE_MAX,                                                   \
    _SC_BC_STRING_MAX,                                                  \
    _SC_CHILD_MAX,                                                      \
    _SC_CLK_TCK,                                                        \
    _SC_CLOCK_SELECTION,                                                \
    _SC_COLL_WEIGHTS_MAX,                                               \
    _SC_CPUTIME,                                                        \
    _SC_DELAYTIMER_MAX,                                                 \
    _SC_EXPR_NEST_MAX,                                                  \
    _SC_FSYNC,                                                          \
    _SC_GETGR_R_SIZE_MAX,                                               \
    _SC_GETPW_R_SIZE_MAX,                                               \
    _SC_HOST_NAME_MAX,                                                  \
    _SC_IOV_MAX,                                                        \
    _SC_IPV6,                                                           \
    _SC_JOB_CONTROL,                                                    \
    _SC_LINE_MAX,                                                       \
    _SC_LOGIN_NAME_MAX,                                                 \
    _SC_MAPPED_FILES,                                                   \
    _SC_MEMLOCK,                                                        \
    _SC_MEMLOCK_RANGE,                                                  \
    _SC_MEMORY_PROTECTION,                                              \
    _SC_MESSAGE_PASSING,                                                \
    _SC_MONOTONIC_CLOCK,                                                \
    _SC_MQ_OPEN_MAX,                                                    \
    _SC_MQ_PRIO_MAX,                                                    \
    _SC_NGROUPS_MAX,                                                    \
    _SC_OPEN_MAX,                                                       \
    _SC_PAGE_SIZE,                                                      \
    _SC_PAGESIZE,                                                       \
    _SC_PRIORITIZED_IO,                                                 \
    _SC_PRIORITY_SCHEDULING,                                            \
    _SC_RAW_SOCKETS,                                                    \
    _SC_RE_DUP_MAX,                                                     \
    _SC_READER_WRITER_LOCKS,                                            \
    _SC_REALTIME_SIGNALS,                                               \
    _SC_REGEXP,                                                         \
    _SC_RTSIG_MAX,                                                      \
    _SC_SAVED_IDS,                                                      \
    _SC_SEM_NSEMS_MAX,                                                  \
    _SC_SEM_VALUE_MAX,                                                  \
    _SC_SEMAPHORES,                                                     \
    _SC_SHARED_MEMORY_OBJECTS,                                          \
    _SC_SHELL,                                                          \
    _SC_SIGQUEUE_MAX,                                                   \
    _SC_SPAWN,                                                          \
    _SC_SPIN_LOCKS,                                                     \
    _SC_SPORADIC_SERVER,                                                \
    _SC_SS_REPL_MAX,                                                    \
    _SC_STREAM_MAX,                                                     \
    _SC_SYMLOOP_MAX,                                                    \
    _SC_SYNCHRONIZED_IO,                                                \
    _SC_THREAD_ATTR_STACKADDR,                                          \
    _SC_THREAD_ATTR_STACKSIZE,                                          \
    _SC_THREAD_CPUTIME,                                                 \
    _SC_THREAD_DESTRUCTOR_ITERATIONS,                                   \
    _SC_THREAD_KEYS_MAX,                                                \
    _SC_THREAD_PRIO_INHERIT,                                            \
    _SC_THREAD_PRIO_PROTECT,                                            \
    _SC_THREAD_PRIORITY_SCHEDULING,                                     \
    _SC_THREAD_PROCESS_SHARED,                                          \
    _SC_THREAD_ROBUST_PRIO_INHERIT,                                     \
    _SC_THREAD_ROBUST_PRIO_PROTECT,                                     \
    _SC_THREAD_SAFE_FUNCTIONS,                                          \
    _SC_THREAD_SPORADIC_SERVER,                                         \
    _SC_THREAD_STACK_MIN,                                               \
    _SC_THREAD_THREADS_MAX,                                             \
    _SC_THREADS,                                                        \
    _SC_TIMEOUTS,                                                       \
    _SC_TIMER_MAX,                                                      \
    _SC_TIMERS,                                                         \
    _SC_TRACE,                                                          \
    _SC_TRACE_EVENT_FILTER,                                             \
    _SC_TRACE_EVENT_NAME_MAX,                                           \
    _SC_TRACE_INHERIT,                                                  \
    _SC_TRACE_LOG,                                                      \
    _SC_TRACE_NAME_MAX,                                                 \
    _SC_TRACE_SYS_MAX,                                                  \
    _SC_TRACE_USER_EVENT_MAX,                                           \
    _SC_TTY_NAME_MAX,                                                   \
    _SC_TYPED_MEMORY_OBJECTS,                                           \
    _SC_TZNAME_MAX,                                                     \
    _SC_V6_ILP32_OFF32,                                                 \
    _SC_V6_ILP32_OFFBIG,                                                \
    _SC_V6_LP64_OFF64,                                                  \
    _SC_V6_LPBIG_OFFBIG,                                                \
    _SC_V7_ILP32_OFF32,                                                 \
    _SC_V7_ILP32_OFFBIG,                                                \
    _SC_V7_LP64_OFF64,                                                  \
    _SC_V7_LPBIG_OFFBIG,                                                \
    _SC_VERSION,                                                        \
    _SC_XOPEN_CRYPT,                                                    \
    _SC_XOPEN_ENH_I18N,                                                 \
    _SC_XOPEN_REALTIME,                                                 \
    _SC_XOPEN_REALTIME_THREADS,                                         \
    _SC_XOPEN_SHM,                                                      \
    _SC_XOPEN_STREAMS,                                                  \
    _SC_XOPEN_UNIX,                                                     \
    _SC_XOPEN_VERSION,                                                  \
    &system_sysconf_end,                                                \
    _PC_2_SYMLINKS,                                                     \
    _PC_ALLOC_SIZE_MIN,                                                 \
    _PC_ASYNC_IO,                                                       \
    _PC_CHOWN_RESTRICTED,                                               \
    _PC_FILESIZEBITS,                                                   \
    _PC_LINK_MAX,                                                       \
    _PC_MAX_CANON,                                                      \
    _PC_MAX_INPUT,                                                      \
    _PC_NAME_MAX,                                                       \
    _PC_NO_TRUNC,                                                       \
    _PC_PATH_MAX,                                                       \
    _PC_PIPE_BUF,                                                       \
    _PC_PRIO_IO,                                                        \
    _PC_REC_INCR_XFER_SIZE,                                             \
    _PC_REC_MAX_XFER_SIZE,                                              \
    _PC_REC_MIN_XFER_SIZE,                                              \
    _PC_REC_XFER_ALIGN,                                                 \
    _PC_SYMLINK_MAX,                                                    \
    _PC_SYNC_IO,                                                        \
    _PC_TIMESTAMP_RESOLUTION,                                           \
    _PC_VDISABLE,                                                       \
    &system_pathconf_end,                                               \
    &system_end                                                         \
}

SYSTEM_VTBL_INITIALIZATION(direct);
SYSTEM_VTBL_INITIALIZATION(trace);

#define EMBEDDING_VTBL_INITIALIZATION(PREFIX)                           \
                                                                        \
static vrpsx_embedding_vtbl_t const VRPSX_CAT2(PREFIX, _embedding_vtbl) = { \
    vrpsx_embedding_IDENTIFIER,                                         \
    &direct_embedding_replace_interface,                                \
    &identified_end,                                                    \
    &direct_embedding_release,                                          \
    &direct_embedding_acquire_op,                                       \
    &counted_identified_end,                                            \
    &VRPSX_CAT2(PREFIX, _embedding_internalize_pid),                    \
    &VRPSX_CAT2(PREFIX, _embedding_externalize_pid),                    \
    &VRPSX_CAT2(PREFIX, _embedding_internalize_path),                   \
    &VRPSX_CAT2(PREFIX, _embedding_externalize_path),                   \
    &VRPSX_CAT2(PREFIX, _embedding_internalize_execution_path),         \
    &VRPSX_CAT2(PREFIX, _embedding_externalize_execution_path),         \
    &embedding_end                                                      \
}

EMBEDDING_VTBL_INITIALIZATION(direct);
EMBEDDING_VTBL_INITIALIZATION(trace);
